/****************************************************************************
** COPYRIGHT (C):    1994 Cay S. Horstmann. All Rights Reserved.
** PROJECT:          ChiLib
** FILE:             chiargs.cpp
** PURPOSE:          Command line argument parsing
** VERSION           1.0
** PROGRAMMER:       Cay Horstmann (CSH)
** START DATE:       5-1-94 (CSH)
** UPDATE HISTORY:
****************************************************************************/

/* IMPORT ******************************************************************/

#ifndef EXPORT
#include "chisetup.h"
#endif

#include <stdlib.h>

EXPORT
#include "chistrng.h"

#include "chiargs.h"

/* TYPES *******************************************************************/

#ifndef EXPORT

class Chi_Args
/* PURPOSE:    Argument parsing
   REMARKS:    the arg and options functions are not const since in some
               command line parsing situations the parsing order has side
               effects. At this point, the class does not support those
               situations, but we want to be prepared for it.
*/
{
public:
   Chi_Args(int, char**);
   Chi_String prog_name() const;
   Chi_String prog_path() const;
   Chi_String arg(int i = - 1);
   Bool option(char c, Chi_String& x, int i = -1);
   Bool int_option(char c, int& x, int i = -1);
   Bool bool_option(char c, Bool& x, int i = -1);

private:
   int _argc;
   char** _argv;
};

#endif

/* FUNCTIONS ***************************************************************/

static Bool atob(const char s[])
/* PURPOSE:    translate ASCII to Bool
   RECEIVES:   s - a string
   RETURNS:    TRUE iff s is empty, "1" or "+"
*/
{  if (s[0] == 0) return TRUE;
   if (s[1] == 0 && (s[0] == '1' || s[0] == '+')) return TRUE;
   return FALSE;
}

/*-------------------------------------------------------------------------*/

Chi_Args::Chi_Args(int argc, char** argv)
/* RECEIVES:   argc, argv - the arguments of main
*/
:  _argc(argc),
   _argv(argv)
{}

/*...........................................................................*/

Chi_String Chi_Args::prog_name() const
/* RETURNS:    the name of the program with drive letter, path name and
               extension stripped off.
*/
{  Chi_String prog_path = _argv[0];
   int n1 = prog_path.find_last_of("\\/:");
   if (n1 == Chi_String::NPOS) n1 = 0; else n1++;
   int n2 = prog_path.find('.', n1);
   if (n2 != Chi_String::NPOS) n2 -= n1;
   return prog_path.substr(n1, n2);
}

/*...........................................................................*/

Chi_String Chi_Args::prog_path() const
/* RETURNS:    the full path of the program
*/
{  return _argv[0];
}

/*...........................................................................*/

Chi_String Chi_Args::arg(int i)
/* RECEIVES:   i - the arg number (-1 = last argument)
   RETURNS:    the ith argument
   REMARKS:    options (starting with -) are skipped
*/
{  int a = 0;
   Chi_String r;
   for (int c = 1; c < _argc; c++)
   {  if (_argv[c][0] != '-')
      {  a++;
         if (a == i) return _argv[c];
         else if (i == -1) r = _argv[c];
      }
   }
   return r;
}

/*...........................................................................*/

Bool Chi_Args::option(char ch, Chi_String& x, int i)
/* PURPOSE:    get a string option
   RECEIVES:   ch - the option character
               x (OUT) - the string to be set to the option value
               i - the option number (-1 = last option with this char)
   RETURNS:    TRUE iff option found
   REMARKS:    x not set if none found
*/
{  int a = 0;
   Bool found = FALSE;
   for (int c = 1; c < _argc; c++)
   {  if (_argv[c][0] == '-' && _argv[c][1] == ch)
      {  a++;
         if (a == i) { x = _argv[c] + 2; return TRUE; }
         else if (i == -1) { x = _argv[c] + 2; found = TRUE; }
      }
   }
   return found;
}

/*...........................................................................*/

Bool Chi_Args::int_option(char ch, int& x, int i)
/* PURPOSE:    get an integer option
   RECEIVES:   ch - the option character
               x (OUT) - the integer to be set to the option value
               i - the option number (-1 = last option with this char)
   RETURNS:    TRUE iff option found
   REMARKS:    x not set if none found
*/
{  int a = 0;
   Bool found = FALSE;
   for (int c = 1; c < _argc; c++)
   {  if (_argv[c][0] == '-' && _argv[c][1] == ch)
      {  a++;
         if (a == i) { x = atoi(_argv[c] + 2); return TRUE; }
         else if (i == -1) { x = atoi(_argv[c] + 2); found = TRUE; }
      }
   }
   return found;
}


/*...........................................................................*/

Bool Chi_Args::bool_option(char ch, Bool& x, int i)
/* PURPOSE:    get a Boolean option
   RECEIVES:   ch - the option character
               x (OUT) - the Boolean to be set to the option value
               i - the option number (-1 = last option with this char)
   RETURNS:    TRUE iff option found
   REMARKS:    x not set if none found
*/
{  int a = 0;
   Bool found = FALSE;
   for (int c = 1; c < _argc; c++)
   {  if (_argv[c][0] == '-' && _argv[c][1] == ch)
      {  a++;
         if (a == i) { x = atob(_argv[c] + 2); return TRUE; }
         else if (i == -1) { x = atob(_argv[c] + 2); found = TRUE; }
      }
   }
   return found;
}

/* TEST ********************************************************************/

#ifdef CHIARGS_TEST

#include <iostream.h>
#ifdef CHI_GCC
#include <unistd.h>
#else
#include <process.h>
#endif

int main(int argc, char** argv)
{  if (argc == 1)
   {  cout << "No arguments supplied. Calling itself with arguments" << endl
         << "\t1990 -e1024 -GBGI -v -GWIN input.dat" << endl;
      execl(argv[0], argv[0], "1990", "-e1024", "-GBGI", "-v", "-GWIN", "input.dat", NULL);
   }
   Chi_Args args(argc, argv);
   cout << "Program name: " << args.prog_name() << endl;
   cout << "Program path: " << args.prog_path() << endl;
   Chi_String s;
   int i = -1;
   cout << "Non-option arguments:" << endl;
   while ((s = args.arg(i)).length() > 0)
   {  cout << "arg " << i << " " << s << endl;
      i++;
      if (i == 0) i++;
   }
   cout << "String option arguments:" << endl;
   char c;
   for (c = 'A'; c <= 'z'; c++)
   {  int i = -1;
      Chi_String t;
      while (args.option(c, t, i))
      {  cout << "opt " << c << " " << i << " " << t << endl;
         i++;
         if (i == 0) i++;
      }
   }

   cout << "Integer option arguments:" << endl;
   for (c = 'A'; c <= 'z'; c++)
   {  int i = -1;
      int t;
      while (args.int_option(c, t, i))
      {  cout << "opt " << c << " " << i << " " << t << endl;
         i++;
         if (i == 0) i++;
      }
   }

   cout << "Boolean option arguments:" << endl;
   for (c = 'A'; c <= 'z'; c++)
   {  int i = -1;
      int t;
      while (args.bool_option(c, t, i))
      {  cout << "opt " << c << " " << i << " " << t << endl;
         i++;
         if (i == 0) i++;
      }
   }

   return 0;
}

#endif

/***************************************************************************/
