/****************************************************************************
** COPYRIGHT (C):    1993 Cay S. Horstmann. All Rights Reserved.
** PROJECT:          ChiLib
** FILE:             chiarray.cpp
** PURPOSE:          Smart arrays
** VERSION           1.1
** LANGUAGE:         Borland C++ 3.1
** TARGET:           IBM PC
** PROGRAMMERS:      Cay S. Horstmann (CSH)
** START DATE:       June 4, 1993
** UPDATE HISTORY:

10/3/95 CSH
Ensure that array memory block is aligned with double--Sparc didn't like
int aligned boundaries for Chi_NumArray<double>

11/22/98 CSH
Made copy constructor and operator= inline for g++

****************************************************************************/

/* IMPORT ******************************************************************/

#ifndef EXPORT
#include "chisetup.h"
#endif

EXPORT
#include <stddef.h>
EXPORT
#include <stdlib.h>
#include <iostream.h>
EXPORT
#include <string.h>

EXPORT
#include "chierror.h"
#include "chiarray.h"

/* TYPES *******************************************************************/

#ifndef EXPORT

template<class X> class Chi_Array
/* PURPOSE:    smart array class template
   RECEIVES:   X - any type with copy construction and assignment
*/
{
public:
   Chi_Array();
   Chi_Array(int lo, int hi);

   void grow(int from, int to);
   void grow(int to);
   void shrink(int from, int to);

   void set(int i, const X& t);  // set ith entry to t; grow if necessary
   void insert(int i, X t); // insert t to ith position, shift others up
   void append(X t);
   void insert_range(int from, int n);
   X remove(int);
   void remove_range(int from, int n);

   void qsort(int (*)(const X&, const X&));
   void qsort(int (*)(const X&, const X&), int from, int n);

   void empty() { detach(); } // remove all elements

   int length() const { return _high - _low + 1; }
   int low() const { return _low; }
   int high() const { return _high; }

   X const& get(int i) const
   {  if (_low <= i && i <= _high)
         return _elements[i - _low];
      else
         return defval();
   }

   X& operator[](int i)
   {  CHI_ASSERT_BOUNDS(i, _low, _high);
      unique();
      return _elements[i - _low];
   }

   X const& operator[](int i) const
   {  CHI_ASSERT_BOUNDS(i, _low, _high);
      return _elements[i - _low];
   }

   Chi_Array(const Chi_Array& b) 
   : _elements(b._elements),
     _low(b._low),
     _high(b._high)
   {  if (_elements) Chi_ArrayHeader::header(_elements)->incref(); }

   const Chi_Array& operator=(const Chi_Array& b)
   {  if (b._elements) Chi_ArrayHeader::header(b._elements)->incref();
      detach();
      _elements = b._elements;
      _low = b._low;
      _high = b._high;
      return *this;
   }

   ~Chi_Array() { detach(); }

private:
   void init(int lo, int hi);
   void ins(int ins1, int keep1, int ins2, int keep2);
   void rem(int rem1, int keep1, int rem2, int keep2);
   static void construct_block(X* p, int size);
   static void copy_block(X* p, const X* from, int size);
   static void destroy_block(X* p, int size);
   static void move_block(X* tt, const X* ff, int size);
   static const X& defval();

   void detach();
   void unique();

   X* _elements;
   int _low;
   int _high;
};

/*-------------------------------------------------------------------------*/

class Chi_PtrArrayBase
/* PURPOSE:    base class for PtrArray template
*/
{
public:
   Chi_PtrArrayBase();
   Chi_PtrArrayBase(int lo, int hi);

   void grow(int from, int to);
   void grow(int to);
   void shrink(int from, int to);

   void insert_range(int from, int n);
   void* remove(int);
   void remove_range(int from, int n);

   void empty() { detach(); } // remove all elements
   int length() const { return _high - _low + 1; }
   int low() const { return _low; }
   int high() const { return _high; }

   void* get(int i) const
   {  if (_low <= i && i <= _high)
         return _elements[i - _low];
      else
         return 0;
   }

   void*& operator[](int i)
   {  CHI_ASSERT_BOUNDS(i, _low, _high);
      unique();
      return _elements[i - _low];
   }

   void* operator[](int i) const
   {  CHI_ASSERT_BOUNDS(i, _low, _high);
      return _elements[i - _low];
   }

   Chi_PtrArrayBase(const Chi_PtrArrayBase& b);
   const Chi_PtrArrayBase& operator=(const Chi_PtrArrayBase& b);
   ~Chi_PtrArrayBase() { detach(); }

protected:
   void set(int i, void* t);  // set ith entry to t; grow if necessary
   void insert(int i, void* t); // insert t to ith position, shift others up
   void append(void* t);
   void qsort(int (*)(const void*, const void*));
   void qsort(int (*)(const void*, const void*), int from, int n);

private:
   void init(int lo, int hi);
   void ins(int ins1, int keep1, int ins2, int keep2);
   void rem(int rem1, int keep1, int rem2, int keep2);
   static void construct_block(void** p, int size);
   static void copy_block(void** p, void* const* from, int size);
   static void move_block(void** tt, void* const* ff, int size);

   void detach();
   void unique();

   void** _elements;
   int _low;
   int _high;
};

/*-------------------------------------------------------------------------*/

template<class X> class Chi_PtrArray : public Chi_PtrArrayBase
/* PURPOSE:    template for smart arrays of pointers
   RECEIVES:   X - any type
   REMARKS:    Makes an array of X*
*/
{
public:
   Chi_PtrArray() {}
   Chi_PtrArray(int lo, int hi) : Chi_PtrArrayBase(lo, hi) {}

   void set(int i, X* t) { Chi_PtrArrayBase::set(i, t); }
   void insert(int i, X* t) { Chi_PtrArrayBase::insert(i, t); }
   void append(X* t) { Chi_PtrArrayBase::append(t); }
   X* remove(int i) { return (X*) Chi_PtrArrayBase::remove(i); }
   void qsort(int (*c)(const X*, const X*))
   { Chi_PtrArrayBase::qsort((int (*)(const void*, const void*))c); }
   void qsort(int (*c)(const X*, const X*), int from, int n)
   { Chi_PtrArrayBase::qsort((int (*)(const void*, const void*))c, from, n); }
   X* get(int i) const { return (X*) Chi_PtrArrayBase::get(i); }
   X*& operator[](int i) { return (X*&) Chi_PtrArrayBase::operator[](i); }
   X* operator[](int i) const { return (X*) Chi_PtrArrayBase::operator[](i); }
};

/*-------------------------------------------------------------------------*/

class Chi_NumArrayBase
/* PURPOSE:    base class for NumArray template
*/
{
public:
   Chi_NumArrayBase(int e);
   Chi_NumArrayBase(int lo, int hi, int e);

   void grow(int from, int to);
   void grow(int to);
   void shrink(int from, int to);

   void insert_range(int from, int n);
   char* remove(int, char* temp);
   void remove_range(int from, int n);

   void empty() { detach(); } // remove all elements
   int length() const { return _high - _low + 1; }
   int low() const { return _low; }
   int high() const { return _high; }

   char* get(int i, char* temp) const
   {  if (_low <= i && i <= _high)
         return _elements + (i - _low) * _elem_size;
      else
      {  construct_block(temp, _elem_size);
         return temp;
      }
   }

   char* at(int i)
   {  CHI_ASSERT_BOUNDS(i, _low, _high);
      unique();
      return _elements + (i - _low) * _elem_size;
   }

   char* at(int i) const
   {  CHI_ASSERT_BOUNDS(i, _low, _high);
      return _elements + (i - _low) * _elem_size;
   }

   Chi_NumArrayBase(const Chi_NumArrayBase& b);
   const Chi_NumArrayBase& operator=(const Chi_NumArrayBase& b);
   ~Chi_NumArrayBase() { detach(); }

protected:
   void set(int i, char* t);  // set ith entry to t; grow if necessary
   void insert(int i, char* t); // insert t to ith position, shift others up
   void append(char* t);
   void qsort(int (*)(const char&, const char&));
   void qsort(int (*)(const char&, const char&), int from, int n);

private:
   void init(int lo, int hi);
   void ins(int ins1, int keep1, int ins2, int keep2);
   void rem(int rem1, int keep1, int rem2, int keep2);
   static void construct_block(char* p, int size);
   static void copy_block(char* p, const char* from, int size);
   static void move_block(char* tt, const char* ff, int size);
   static void swap_block(char* p, char* q, int size);

   void detach();
   void unique();

   char* _elements;
   int _low;
   int _high;
   int _elem_size;
};

/*-------------------------------------------------------------------------*/

template<class X> class Chi_NumArray : public Chi_NumArrayBase
/* PURPOSE:    smart array class template
   RECEIVES:   X - any type that can be bitwise 0 initialized and bitwise copied
   REMARKS:    Use this template for more efficient arrays of numeric types
               (int, double, ...)
*/
{
public:
   Chi_NumArray() : Chi_NumArrayBase(sizeof(X)) {}
   Chi_NumArray(int lo, int hi) : Chi_NumArrayBase(lo, hi, sizeof(X)) {}

   void set(int i, const X& t) { Chi_NumArrayBase::set(i, (char*)&t); }
   void insert(int i, const X& t) { Chi_NumArrayBase::insert(i, (char*)&t); }
   void append(const X& t) { Chi_NumArrayBase::append((char*)&t); }
   X remove(int i) { return *(X*) Chi_NumArrayBase::remove(i, (char*)&tempval()); }
   void qsort(int (*c)(const X&, const X&))
   { Chi_NumArrayBase::qsort((int (*)(const char&, const char&))c); }
   void qsort(int (*c)(const X&, const X&), int from, int n)
   { Chi_NumArrayBase::qsort((int (*)(const char&, const char&))c, from, n); }
   X get(int i) const { return *(X*) Chi_NumArrayBase::get(i, (char*)&tempval()); }
   X& operator[](int i) { return *(X*) Chi_NumArrayBase::at(i); }
   X operator[](int i) const { return *(X*) Chi_NumArrayBase::at(i); }

   static X& tempval();
};

/*-------------------------------------------------------------------------*/

class Chi_ArrayHeader
/* PURPOSE:    The (invisible) header stored with each smart array
               It contains the reference count, current block size and
               growth amount.
   REMARKS:    This is stored below the array itself and hidden on purpose
               so the array values can be inspected easily in the debugger
*/
{
public:
   enum { GROWBY = 20, };

   int unique() { return _rc == 1; }
   void incref() { ++_rc; }
   int decref() { return --_rc; }

   static Chi_ArrayHeader* alloc(int xsize, int size, int growby );

   static Chi_ArrayHeader* header(void*);

   void free();

   size_t block_size() const { return _block_size; }
   size_t growby() const { return _growby; }
   void* start() const;

private:
   unsigned _rc;                       // reference count
   size_t _block_size;
   int _growby;
};

typedef double Chi_AlignType; 

union Chi_AlignedArrayHeader
{  Chi_ArrayHeader _header;
   Chi_AlignType _pad[(sizeof(Chi_ArrayHeader) - 1) / sizeof(Chi_AlignType) + 1];
};

#endif

/* GLOBALS *****************************************************************/

EXPORT
#if defined(CHI_BC30) || defined(CHI_BC31) || defined(CHI_BC40) || defined(CHI_BC45)
EXPORT void* chiarray_use_me;
   // To avoid a spurious "argument not used" message, assign to this pointer
EXPORT
#endif

/* FUNCTIONS ***************************************************************/

EXPORT void* operator new(size_t, void* p, Chi_ArrayHeader*)
/* PURPOSE:    placement version of operator new
   RECEIVES:   p - the address at which to run the constructor
   RETURNS:    p
   REMARKS:    We add a dummy Chi_ArrayHeader argument because placement isn't
               yet in a standard header
*/
{  return p;
}

/*-------------------------------------------------------------------------*/

#ifndef EXPORT

template< class X >
Chi_Array<X>::Chi_Array()
:  _elements(0),
   _low(0),
   _high(-1)
{}

/*.........................................................................*/

template< class X >
Chi_Array<X>::Chi_Array(int lo, int hi)
/* RECEIVES:   lo, hi - low and high bounds of the array
*/
{  init(lo, hi);
}

/*.........................................................................*/

template< class X >
void Chi_Array<X>::detach()
/* PURPOSE:    decrement reference count and free associated object if
               necessary
*/
{  if (_elements && Chi_ArrayHeader::header(_elements)->decref() == 0)
   {  Chi_ArrayHeader* ph = Chi_ArrayHeader::header(_elements);
      destroy_block(_elements, ph->block_size());
      ph->free();
   }
   _elements = 0;
   _low = 0;
   _high = -1;
}

/*.........................................................................*/

template< class X >
void Chi_Array<X>::unique()
/* PURPOSE:    clone array rep if ref count > 1
*/
{  if (_elements && !Chi_ArrayHeader::header(_elements)->unique())
   {  Chi_ArrayHeader* ph = Chi_ArrayHeader::header(_elements);
      ph->decref();
      Chi_ArrayHeader* nh = Chi_ArrayHeader::alloc(sizeof(X), ph->block_size(), ph->growby());
      if (nh != 0)
      {  // initialize all array entries
         X* pp = _elements;
         _elements = (X*)nh->start();
         copy_block(_elements, pp, ph->block_size());
         construct_block(_elements + ph->block_size(), nh->block_size() - ph->block_size());
      }
   }
}

/*.........................................................................*/

template< class X >
void Chi_Array<X>::grow(int newindex)
/* PURPOSE:    grow array to contain an index
   RECEIVES:   newindex - the index to grow to
   REMARKS:    if newindex already within bounds, nothing happens
*/
{  if (_elements)
   {  if (newindex < _low)
         ins(_low - newindex, _high - _low + 1, 0, 0);
      else if (newindex > _high)
         ins(0, _high - _low + 1, newindex - _high, 0);
   }
   else
      init(newindex, newindex);
}

/*.........................................................................*/

template< class X >
void Chi_Array<X>::grow(int newlo, int newhi)
/* PURPOSE:    grow array to contain an index range
   RECEIVES:   newlo ... newhi - the index range to grow to
*/
{  if (_elements)
   {  if (_low < newlo) newlo = _low;
      if (_high > newhi) newhi = _high;
      ins(_low - newlo, _high - _low + 1, newhi - _high, 0);
   }
   else
      init(newlo, newhi);
}

/*.........................................................................*/

template< class X >
void Chi_Array<X>::shrink(int newlo, int newhi)
/* PURPOSE:    shrink array to contain no more than an index range
   RECEIVES:   newlo ... newhi - the index range to shrink into
*/
{  if (_elements)
   {  if (newlo > newhi) detach();
      else
      {  if (newlo < _low) newlo = _low;
         if (newhi > _high) newhi = _high;
         rem(newlo - _low, newhi - newlo + 1, _high - newhi, 0);
      }
   }
}

/*.........................................................................*/

template< class X >
void Chi_Array<X>::set(int i, const X& t)
/* PURPOSE:    Set an array element to a value
   RECEIVES:   i - index of element to set
               t - the new element
   REMARKS:    grows the array if necessary
               This correctly handles the case in which newElement is an
               element of the array, e.g. a.set(i+1, a[i]) will work
               even if the array has to grow as a result
*/
{  if (_low <= i && i <= _high)
   {  unique();
      _elements[i - _low] = t;
   }
   else
   {  X ttemp = t;
      grow(i);
      _elements[i - _low] = ttemp;
   }
}

/*.........................................................................*/

template< class X >
void Chi_Array<X>::insert(int i, X t)
/* PURPOSE:    insert an element, shift others up
   RECEIVES:   i - the place to insert at
               t - the element to insert
   REMARKS:    t is NOT passed by const reference since it might come elsewhere
               from the same array
*/
{  if (_low <= i && i <= _high)
   {  insert_range(i, 1);
      _elements[i - _low] = t;
   }
   else set(i, t);
}

/*.........................................................................*/

template< class X >
void Chi_Array<X>::append(X t)
/* PURPOSE:    append an element at the end of the array
   RECEIVES:   t - the element to insert
   REMARKS:    if the array is empty, the element is placed at index 0
*/
{  grow(_high + 1);
   _elements[_high - _low] = t;
}

/*.........................................................................*/

template< class X >
void Chi_Array<X>::insert_range(int from, int n)
/* PURPOSE:    insert a range of default elements into an array, shifting the
               others up
   RECEIVES:   from - the index to start insertion
               n - the number of elements to insert
*/
{  if (n <= 0) return;
   if (_elements) 
   {  if (from > _high) // CSH 4-7-95
         ins(0, _high - _low + 1, from + n - 1 - _high, 0);
      else if (from < _low)
      {  if (from + n <= _low)
            ins(_low - from, _high - _low + 1, 0, 0);
         else
         {  int shift = from + n - _low;
            ins(n, _high - _low + 1, 0, 0);
            _high += shift;
            _low += shift;
         }
      }
      else
         ins(0, from - _low, n, _high - from + 1);
   }
   else
      init(from, from + n - 1);
}

/*.........................................................................*/

template< class X >
X Chi_Array<X>::remove(int at)
/* PURPOSE:    remove an element and shift the others down
   RECEIVES:   n - the position at which to remove the element
   RETURNS:    the removed element
*/
{  X r = get(at);
   remove_range(at, 1);
   return r;
}

/*.........................................................................*/

template< class X >
void Chi_Array<X>::remove_range(int from, int n)
/* PURPOSE:    remove a range of elements and shift the remainder down
   RECEIVES:   from - the index to start removal
               n - the number of elements to remove
*/
{  if (_elements)
   {  if (from < _low) { n -= _low - from; from = _low; }
      if (n <= 0) return;
      if (from > _high) return;
      if (n > _high - from + 1) n = _high - from + 1;
      rem(0, from - _low, n, _high - from + 1 - n);
   }
}

/*.........................................................................*/

template<class X>
void Chi_Array<X>::qsort(int (*comp)(const X&, const X&))
/* PURPOSE:    sort the array
   RECEIVES:   comp - pointer to comparison function
*/
{  qsort(comp, _low, _high - _low + 1);
}

/*.........................................................................*/

template<class X>
void Chi_Array<X>::qsort(int (*comp)(const X&, const X&), int from, int n)
/* PURPOSE:    sort a range of the array
   RECEIVES:   comp - pointer to comparison function
               from - the index at which to start sorting
               n - the number of elements to sort
   REMARKS:    Taken from Bentley, Programming Pearls, p. 112
*/
{  if (from < _low) from = _low;
   int to = from + n - 1;
   if (to > _high) to = _high;
   if (from < to)
   {  unique();
      int i = rand() % (to - from + 1) + from;
         // a random integer between from and to
      chi_swap(self[from], self[i]);
      int mid = from;
      for (i = from + 1; i <= to; i++)
      /* INVARIANT:  self[from + 1...mid] < self[from] &&
                     self[mid+1...i-1] >= self[from]
      */
      {  if ((*comp)(self[i], self[from]) < 0)
         {  mid++;
            chi_swap(self[mid], self[i]);
         }
      }
      chi_swap(self[from], self[mid]);
      /* self[from...mid-1] < self[m] <= self[m+1...to] */
      qsort(comp, from, mid - from);
      qsort(comp, mid+1, to - mid);
   }
}

/*.........................................................................*/

template< class X >
void Chi_Array<X>::init(int lo, int hi)

/* PURPOSE:    Allocate array storage from the heap
               and initialize it with the default constructor
   RECEIVES:   lo, hi - array range
*/
{  // allocate storage

   if (lo > hi)
   {  _elements = 0;
      _low = 0;
      _high = -1;
      return;
   }

   _low = lo;
   _high = hi;
   Chi_ArrayHeader* ph = Chi_ArrayHeader::alloc(sizeof(X), hi - lo + 1, Chi_ArrayHeader::GROWBY);
   if (ph == 0) return;

   // initialize all array entries

   _elements = (X*)ph->start();
   construct_block(_elements, ph->block_size());
}

/*.........................................................................*/

template< class X >
void Chi_Array<X>::construct_block(X* p, int size)
/* PURPOSE:    run the default constructor on a block of elements
   RECEIVES:   p - start of block
               size - number of elements
*/
{  for (int i = 0; i < size; i++)
      new(p + i, (Chi_ArrayHeader*)0) X;
}

/*.........................................................................*/

template< class X >
void Chi_Array<X>::copy_block(X* p, const X* from, int size)
/* PURPOSE:    run the copy constructor to make a copy of a block of elements
   RECEIVES:   p - start of target block
               from - start of source block
               size - number of elements
*/
{  for (int i = 0; i < size; i++)
      new(p + i, (Chi_ArrayHeader*)0) X(from[i]);
}

/*.........................................................................*/

template< class X >
void Chi_Array<X>::destroy_block(X* p, int size)
/* PURPOSE:    run the destructor on a block of elements
   RECEIVES:   p - start of block
               size - number of elements
   REMARKS:    it is legal to run the "destructor" on a non-class type
               (e.g. X = int) but some compilers do not support this.
               In that case, use PtrArray/NumArray.

               Borland 3.1 and 4.0 give a warning "p not used" if X has no
               destructor. For that reason, we make a spurious usage of p.
*/
{  for (int i = size - 1; i >= 0; i--)
#ifdef CHI_WC10
      (p + i)->~X();
#elif defined(CHI_BC30) || defined(CHI_MC21)
      ;
#else
      (p + i)->X::~X();
#endif
#if defined(CHI_BC30) || defined(CHI_BC31) || defined(CHI_BC40) || defined(CHI_BC45)
   chiarray_use_me = p;
#endif
}

/*.........................................................................*/

template< class X >
void Chi_Array<X>::move_block(X* tt, const X* ff, int size)
/* PURPOSE:    run the copy constructor and destructor to move a block of elements
   RECEIVES:   tt - start of target block
               from - start of source block
               size - number of elements
   REMARKS:    source and target may overlap
*/
{  if (size == 0) return;
   if (tt == ff) return;
   if (tt > ff) // move from high to low address to avoid overwrite
   {  tt += size;
      ff += size;
      for (int i = size; i > 0; i--)
      {  tt--;
         ff--;
         new(tt, (Chi_ArrayHeader*)0) X(*ff);
#ifdef CHI_WC10
         ff->~X();
#elif defined(CHI_BC30) || defined(CHI_MC21)
         ;
#else
         ff->X::~X();
#endif
      }
   }
   else
   {  for (int i = 0; i < size; i++)
      {  new(tt, (Chi_ArrayHeader*)0) X(*ff);
#ifdef CHI_WC10
         ff->~X();
#elif defined(CHI_BC30) || defined(CHI_MC21)
         ;
#else
         ff->X::~X();
#endif
         tt++;
         ff++;
      }
   }
}

/*.........................................................................*/

template< class X >
void Chi_Array<X>::ins(int ins1, int keep1, int ins2, int keep2)
/* PURPOSE:    Grow the array
   RECEIVES:   ins1 - the #new entries to insert at the bottom of the array
               keep1 - the #entries to keep after that
               ins2 - the #new entries to insert after that
               keep2 - the #entries to keep after that
   RETURNS:    the new address
   REMARKS:    For grow, keep2 is 0
               For insert, ins1 is 0
   PRECOND:    keep1 + keep2 == _high - _low + 1, keep1, keep2 >= 0
*/
{  _low -= ins1; // CSH 4-7-95
   _high += ins2;

   int new_size = ins1 + keep1 + ins2 + keep2;
   Chi_ArrayHeader* ph = Chi_ArrayHeader::header(_elements);
   if (new_size <= ph->block_size() && ins1 == 0 && keep2 == 0) return;
   Bool uniq = ph->unique();
         // TRUE iff pp is the only reference to that information

   if (uniq && new_size <= ph->block_size()) // shift entries
   {  if (ins1 + ins2 > 0)
      {  // start on top since shifting up
         destroy_block(_elements + keep1 + keep2, ins1 + ins2);
         move_block(_elements + ins1 + keep1 + ins2, _elements + keep1, keep2);
         if (ins2 > 0)
            construct_block(_elements + ins1 + keep1, ins2);
         if (ins1 > 0)
         {  move_block(_elements + ins1, _elements, keep1);
            construct_block(_elements, ins1);
         }
      }
   }
   else // allocate new block
   {  ph->decref();
      Chi_ArrayHeader* nh = Chi_ArrayHeader::alloc(sizeof(X), new_size, ph->growby());
      if (nh == 0) return;

      X* pp = _elements;
      _elements = (X*)nh->start();

      // initialize all new array entries

      if (ins1) construct_block(_elements, ins1);
      copy_block(_elements + ins1, pp, keep1);
      construct_block(_elements + ins1 + keep1, ins2);
      if (keep2)  copy_block(_elements + ins1 + keep1 + ins2, pp + keep1, keep2);
      construct_block(_elements + new_size, nh->block_size() - new_size);

      if (uniq)
      {  destroy_block(pp, ph->block_size());
         ph->free();
      }
   }
}

/*.........................................................................*/

template< class X >
void Chi_Array<X>::rem(int rem1, int keep1, int rem2, int keep2)
/* PURPOSE:    remove elements from the array
   RECEIVES:   rem1 - #elements to remove at bottom
               keep1 - #elements to keep after that
               rem2 - #elements to remove after that
               keep2 - #elements to keep after that
*/
{  _low += rem1; // CSH 4-7-95
   _high -= rem2;
   int new_size = keep1 + keep2;
   if (new_size == 0) { detach(); return; }
   Chi_ArrayHeader* ph = Chi_ArrayHeader::header(_elements);
   if (rem1 == 0 && rem2 == 0 && new_size > ph->block_size() - ph->growby() )
      return;

   Bool uniq = ph->unique();
      // TRUE iff pp is the only reference to that information

   if (uniq && new_size > ph->block_size() - ph->growby()) // shift entries
   {  if (rem1 > 0)
      {  destroy_block(_elements, rem1);
         move_block(_elements, _elements + rem1, keep1);
      }
      if (rem2 > 0)
         destroy_block(_elements + rem1 + keep1, rem2);
      move_block(_elements + keep1, _elements + rem1 + keep1 + rem2, keep2);
      construct_block(_elements + keep1 + keep2, rem1 + rem2);
   }
   else // allocate new block
   {  ph->decref();
      Chi_ArrayHeader* nh = Chi_ArrayHeader::alloc(sizeof(X), new_size, ph->growby());
      if (nh == 0) return;

      X* pp = _elements;
      _elements = (X*)nh->start();

      copy_block(_elements, pp + rem1, keep1);
      copy_block(_elements + keep1, pp + rem1 + keep1 + rem2, keep2);
      construct_block(_elements + new_size, nh->block_size() - new_size);
      if (uniq)
      {  destroy_block(pp, ph->block_size());
         ph->free();
      }
   }
}

//............................................................................

template< class X >
const X& Chi_Array<X>::defval()
/* PURPOSE:    return a default value of type X
*/
{  static X d;
   return d;
}

#endif

/*-------------------------------------------------------------------------*/

Chi_PtrArrayBase::Chi_PtrArrayBase()
:  _elements(0),
   _low(0),
   _high(-1)
{}

/*.........................................................................*/

Chi_PtrArrayBase::Chi_PtrArrayBase(int lo, int hi)
/* RECEIVES:   lo, hi - low and high bounds
*/
{  init(lo, hi);
}

/*.........................................................................*/


Chi_PtrArrayBase::Chi_PtrArrayBase(const Chi_PtrArrayBase& b)
:  _elements(b._elements),
   _low(b._low),
   _high(b._high)
{  if (_elements) Chi_ArrayHeader::header(_elements)->incref(); }

/*.........................................................................*/


const Chi_PtrArrayBase& Chi_PtrArrayBase::operator=(const Chi_PtrArrayBase& b)
{  if (b._elements) Chi_ArrayHeader::header(b._elements)->incref();
   detach();
   _elements = b._elements;
   _low = b._low;
   _high = b._high;
   return *this;
}

/*.........................................................................*/


void Chi_PtrArrayBase::detach()
/* PURPOSE:    decrement reference count and free associated object if
               necessary
*/
{  if (_elements && Chi_ArrayHeader::header(_elements)->decref() == 0)
   {  Chi_ArrayHeader* ph = Chi_ArrayHeader::header(_elements);
      ph->free();
   }
   _elements = 0;
   _low = 0;
   _high = -1;
}

/*.........................................................................*/


void Chi_PtrArrayBase::unique()
/* PURPOSE:    clone array rep if ref count > 1
*/
{  if (_elements && !Chi_ArrayHeader::header(_elements)->unique())
   {  Chi_ArrayHeader* ph = Chi_ArrayHeader::header(_elements);
      ph->decref();
      Chi_ArrayHeader* nh = Chi_ArrayHeader::alloc(sizeof(void*), ph->block_size(), ph->growby());
      if (nh != 0)
      {  // initialize all array entries
         void** pp = _elements;
         _elements = (void**)nh->start();
         copy_block(_elements, pp, ph->block_size());
         construct_block(_elements + ph->block_size(), nh->block_size() - ph->block_size());
      }
   }
}

/*.........................................................................*/


void Chi_PtrArrayBase::grow(int newindex)
/* PURPOSE:    grow array to contain an index
   RECEIVES:   newindex - the index to grow to
   REMARKS:    if newindex already within bounds, nothing happens
*/
{  if (_elements)
   {  if (newindex < _low)
         ins(_low - newindex, _high - _low + 1, 0, 0);
      else if (newindex > _high)
         ins(0, _high - _low + 1, newindex - _high, 0);
   }
   else
      init(newindex, newindex);
}

/*.........................................................................*/


void Chi_PtrArrayBase::grow(int newlo, int newhi)
/* PURPOSE:    grow array to contain an index range
   RECEIVES:   newlo ... newhi - the index range to grow to
*/
{  if (_elements)
   {  if (_low < newlo) newlo = _low;
      if (_high > newhi) newhi = _high;
      ins(_low - newlo, _high - _low + 1, newhi - _high, 0);
   }
   else
      init(newlo, newhi);
}

/*.........................................................................*/


void Chi_PtrArrayBase::shrink(int newlo, int newhi)
/* PURPOSE:    shrink array to contain no more than an index range
   RECEIVES:   newlo ... newhi - the index range to shrink into
*/
{  if (_elements)
   {  if (newlo > newhi) detach();
      else
      {  if (newlo < _low) newlo = _low;
         if (newhi > _high) newhi = _high;
         rem(newlo - _low, newhi - newlo + 1, _high - newhi, 0);
      }
   }
}

/*.........................................................................*/


void Chi_PtrArrayBase::set(int i, void* t)
/* PURPOSE:    Set an array element to a value
   RECEIVES:   i - index of element to set
               t - the new element
*/
{  if (_low <= i && i <= _high)
   {  unique();
      _elements[i - _low] = t;
   }
   else
   {  grow(i);
      _elements[i - _low] = t;
   }
}

/*.........................................................................*/

void Chi_PtrArrayBase::insert(int i, void* t)
/* PURPOSE:    insert an element, shift others up
   RECEIVES:   i - the place to insert at
               t - the element to insert
*/
{  insert_range(i, 1);
   _elements[i - _low] = t;
}

/*.........................................................................*/


void Chi_PtrArrayBase::append(void* t)
/* PURPOSE:    append an element at the end of the array
   RECEIVES:   t - the element to insert
   REMARKS:    if the array is empty, the element is placed at index 0
*/
{  grow(_high + 1);
   _elements[_high - _low] = t;
}

/*.........................................................................*/

void Chi_PtrArrayBase::insert_range(int from, int n)
/* PURPOSE:    insert a range of default elements into an array, shifting the
               others up
   RECEIVES:   from - the index to start insertion
               n - the number of elements to insert
*/
{  if (n <= 0) return;
   if (_elements) 
   {  if (from > _high) // CSH 4-7-95
         ins(0, _high - _low + 1, from + n - 1 - _high, 0);
      else if (from < _low)
      {  if (from + n <= _low)
            ins(_low - from, _high - _low + 1, 0, 0);
         else
         {  int shift = from + n - _low;
            ins(n, _high - _low + 1, 0, 0);
            _high += shift;
            _low += shift;
         }
      }
      else
         ins(0, from - _low, n, _high - from + 1);
   }
   else
      init(from, from + n - 1);
}

/*.........................................................................*/

void* Chi_PtrArrayBase::remove(int at)
/* PURPOSE:    remove an element and shift the others down
   RECEIVES:   n - the position at which to remove the element
   RETURNS:    the removed element
*/
{  void* r = get(at);
   remove_range(at, 1);
   return r;
}

/*.........................................................................*/

void Chi_PtrArrayBase::remove_range(int from, int n)
/* PURPOSE:    remove a range of elements and shift the remainder down
   RECEIVES:   from - the index to start removal
               n - the number of elements to remove
*/
{  if (_elements)
   {  if (from < _low) { n -= _low - from; from = _low; }
      if (n <= 0) return;
      if (from > _high) return;
      if (n > _high - from + 1) n = _high - from + 1;
      rem(0, from - _low, n, _high - from + 1 - n);
   }
}

/*.........................................................................*/

void Chi_PtrArrayBase::qsort(int (*comp)(const void*, const void*))
/* PURPOSE:    sort the array
   RECEIVES:   comp - pointer to comparison function
*/
{  qsort(comp, _low, _high - _low + 1);
}

/*.........................................................................*/

void Chi_PtrArrayBase::qsort(int (*comp)(const void*, const void*), int from, int n)
/* PURPOSE:    sort a range of the array
   RECEIVES:   comp - pointer to comparison function
               from - the index at which to start sorting
               n - the number of elements to sort
   REMARKS:    Taken from Bentley, Programming Pearls, p. 112
*/
{  if (from < _low) from = _low;
   int to = from + n - 1;
   if (to > _high) to = _high;
   if (from < to)
   {  unique();
      int i = rand() % (to - from + 1) + from;
         // a random integer between from and to
      chi_swap(self[from], self[i]);
      int mid = from;
      for (i = from + 1; i <= to; i++)
      /* INVARIANT:  self[from + 1...mid] < self[from] &&
                     self[mid+1...i-1] >= self[from]
      */
      {  if ((*comp)(self[i], self[from]) < 0)
         {  mid++;
            chi_swap(self[mid], self[i]);
         }
      }
      chi_swap(self[from], self[mid]);
      /* self[from...mid-1] < self[m] <= self[m+1...to] */
      qsort(comp, from, mid - from);
      qsort(comp, mid+1, to - mid);
   }
}

/*.........................................................................*/

void Chi_PtrArrayBase::init(int lo, int hi)
/* PURPOSE:    Allocate array storage from the heap
               and initialize it with the default constructor
   RECEIVES:   lo, hi - array range
*/
{  // allocate storage

   if (lo > hi)
   {  _elements = 0;
      _low = 0;
      _high = -1;
      return;
   }

   _low = lo;
   _high = hi;
   Chi_ArrayHeader* ph = Chi_ArrayHeader::alloc(sizeof(void*), hi - lo + 1, Chi_ArrayHeader::GROWBY);
   if (ph == 0) return;

   // initialize all array entries

   _elements = (void**)ph->start();
   construct_block(_elements, ph->block_size());
}

/*.........................................................................*/

void Chi_PtrArrayBase::construct_block(void** p, int size)
/* PURPOSE:    run the default constructor on a block of elements
   RECEIVES:   p - start of block
               size - number of elements
*/
{  memset(p, 0, size * sizeof(void*));
}

/*.........................................................................*/

void Chi_PtrArrayBase::copy_block(void** p,  void* const* from, int size)
/* PURPOSE:    run the copy constructor to make a copy of a block of elements
   RECEIVES:   p - start of target block
               from - start of source block
               size - number of elements
*/
{  memcpy(p, from, size * sizeof(void*));
}

/*.........................................................................*/

void Chi_PtrArrayBase::move_block(void** tt, void* const* ff, int size)
/* PURPOSE:    run the copy constructor and destructor to move a block of elements
   RECEIVES:   tt - start of target block
               from - start of source block
               size - number of elements
   REMARKS:    source and target may overlap
*/
{  memmove(tt, ff, size * sizeof(void*));
}

/*.........................................................................*/

void Chi_PtrArrayBase::ins(int ins1, int keep1, int ins2, int keep2)
/* PURPOSE:    Grow the array
   RECEIVES:   ins1 - the #new entries to insert at the bottom of the array
               keep1 - the #entries to keep after that
               ins2 - the #new entries to insert after that
               keep2 - the #entries to keep after that
   RETURNS:    the new address
   REMARKS:    For grow, keep2 is 0
               For insert, ins1 is 0
   PRECOND:    keep1 + keep2 == _high - _low + 1, keep1, keep2 >= 0
*/
{  _low -= ins1; // CSH 4-7-95
   _high += ins2;

   int new_size = ins1 + keep1 + ins2 + keep2;
   Chi_ArrayHeader* ph = Chi_ArrayHeader::header(_elements);
   if (new_size <= ph->block_size() && ins1 == 0 && keep2 == 0) return;
   Bool uniq = ph->unique();
         // TRUE iff pp is the only reference to that information

   if (uniq && new_size <= ph->block_size()) // shift entries
   {  if (ins1 + ins2 > 0)
      {  // start on top since shifting up
         move_block(_elements + ins1 + keep1 + ins2, _elements + keep1, keep2);
         if (ins2 > 0)
            construct_block(_elements + ins1 + keep1, ins2);
         if (ins1 > 0)
         {  move_block(_elements + ins1, _elements, keep1);
            construct_block(_elements, ins1);
         }
      }
   }
   else // allocate new block
   {  ph->decref();
      Chi_ArrayHeader* nh = Chi_ArrayHeader::alloc(sizeof(void*), new_size, ph->growby());
      if (nh == 0) return;

      void** pp = _elements;
      _elements = (void**)nh->start();

      // initialize all new array entries

      if (ins1) construct_block(_elements, ins1);
      copy_block(_elements + ins1, pp, keep1);
      construct_block(_elements + ins1 + keep1, ins2);
      if (keep2)  copy_block(_elements + ins1 + keep1 + ins2, pp + keep1, keep2);
      construct_block(_elements + new_size, nh->block_size() - new_size);

      if (uniq)
      {
         ph->free();
      }
   }
}

/*.........................................................................*/

void Chi_PtrArrayBase::rem(int rem1, int keep1, int rem2, int keep2)
/* PURPOSE:    remove elements from the array
   RECEIVES:   rem1 - #elements to remove at bottom
               keep1 - #elements to keep after that
               rem2 - #elements to remove after that
               keep2 - #elements to keep after that
*/
{  _low += rem1; // CSH 4-7-95
   _high -= rem2;
   int new_size = keep1 + keep2;
   if (new_size == 0) { detach(); return; }
   Chi_ArrayHeader* ph = Chi_ArrayHeader::header(_elements);
   if (rem1 == 0 && rem2 == 0 && new_size > ph->block_size() - ph->growby() )
      return;

   Bool uniq = ph->unique();
      // TRUE iff pp is the only reference to that information

   if (uniq && new_size > ph->block_size() - ph->growby()) // shift entries
   {  if (rem1 > 0)
      {
         move_block(_elements, _elements + rem1, keep1);
      }

      move_block(_elements + keep1, _elements + rem1 + keep1 + rem2, keep2);
      construct_block(_elements + keep1 + keep2, rem1 + rem2);
   }
   else // allocate new block
   {  ph->decref();
      Chi_ArrayHeader* nh = Chi_ArrayHeader::alloc(sizeof(void*), new_size, ph->growby());
      if (nh == 0) return;

      void** pp = _elements;
      _elements = (void**)nh->start();

      copy_block(_elements, pp + rem1, keep1);
      copy_block(_elements + keep1, pp + rem1 + keep1 + rem2, keep2);
      construct_block(_elements + new_size, nh->block_size() - new_size);
      if (uniq)
      {  ph->free();
      }
   }
}

/*-------------------------------------------------------------------------*/

Chi_NumArrayBase::Chi_NumArrayBase(int e)
/* RECEIVES:   e - size of each entry
*/
:  _elements(0),
   _low(0),
   _high(-1),
   _elem_size(e)
{}

/*.........................................................................*/

Chi_NumArrayBase::Chi_NumArrayBase(int lo, int hi, int e)
/* RECEIVES:   lo, hi - array bounds
               e - size of each entry
*/
:  _elem_size(e)
{  init(lo, hi);
}

/*.........................................................................*/


Chi_NumArrayBase::Chi_NumArrayBase(const Chi_NumArrayBase& b)
:  _elements(b._elements),
   _low(b._low),
   _high(b._high),
   _elem_size(b._elem_size)
{  if (_elements) Chi_ArrayHeader::header(_elements)->incref();
}

/*.........................................................................*/


const Chi_NumArrayBase& Chi_NumArrayBase::operator=(const Chi_NumArrayBase& b)
{  if (b._elements) Chi_ArrayHeader::header(b._elements)->incref();
   detach();
   _elements = b._elements;
   _low = b._low;
   _high = b._high;
   _elem_size = b._elem_size;
   return *this;
}

/*.........................................................................*/


void Chi_NumArrayBase::detach()
/* PURPOSE:    decrement reference count and free associated object if
               necessary
*/
{  if (_elements && Chi_ArrayHeader::header(_elements)->decref() == 0)
   {  Chi_ArrayHeader* ph = Chi_ArrayHeader::header(_elements);
      ph->free();
   }
   _elements = 0;
   _low = 0;
   _high = -1;
}

/*.........................................................................*/


void Chi_NumArrayBase::unique()
/* PURPOSE:    clone array rep if ref count > 1
*/
{  if (_elements && !Chi_ArrayHeader::header(_elements)->unique())
   {  Chi_ArrayHeader* ph = Chi_ArrayHeader::header(_elements);
      ph->decref();
      Chi_ArrayHeader* nh = Chi_ArrayHeader::alloc(_elem_size, ph->block_size(), ph->growby());
      if (nh != 0)
      {  // initialize all array entries
         char* pp = _elements;
         _elements = (char*)nh->start();
         copy_block(_elements, pp, _elem_size * ph->block_size());
         construct_block(_elements + _elem_size * ph->block_size(), _elem_size * (nh->block_size() - ph->block_size()));
      }
   }
}

/*.........................................................................*/


void Chi_NumArrayBase::grow(int newindex)
/* PURPOSE:    grow array to contain an index
   RECEIVES:   newindex - the index to grow to
   REMARKS:    if newindex already within bounds, nothing happens
*/
{  if (_elements)
   {  if (newindex < _low)
         ins(_low - newindex, _high - _low + 1, 0, 0);
      else if (newindex > _high)
         ins(0, _high - _low + 1, newindex - _high, 0);
   }
   else
      init(newindex, newindex);
}

/*.........................................................................*/


void Chi_NumArrayBase::grow(int newlo, int newhi)
/* PURPOSE:    grow array to contain an index range
   RECEIVES:   newlo ... newhi - the index range to grow to
*/
{  if (_elements)
   {  if (_low < newlo) newlo = _low;
      if (_high > newhi) newhi = _high;
      ins(_low - newlo, _high - _low + 1, newhi - _high, 0);
   }
   else
      init(newlo, newhi);
}

/*.........................................................................*/

void Chi_NumArrayBase::shrink(int newlo, int newhi)
/* PURPOSE:    shrink array to contain no more than an index range
   RECEIVES:   newlo ... newhi - the index range to shrink into
*/
{  if (_elements)
   {  if (newlo > newhi) detach();
      else
      {  if (newlo < _low) newlo = _low;
         if (newhi > _high) newhi = _high;
         rem(newlo - _low, newhi - newlo + 1, _high - newhi, 0);
      }
   }
}

/*.........................................................................*/

void Chi_NumArrayBase::set(int i, char* t)
/* PURPOSE:    Set an array element to a value
   RECEIVES:   i - index of element to set
               t - the new element
*/
{  if (_low <= i && i <= _high)
   {  unique();
      memcpy(_elements + _elem_size * (i - _low), t, _elem_size);
   }
   else if (_elements <= t && t < _elements + _elem_size * (_high - _low + 1))
      // t points to an element in the array -- will relocate under grow
   {  void* t2 = malloc(_elem_size);
      CHI_ASSERT(t2);
      memcpy(t2, t, _elem_size);
      grow(i);
      memcpy(_elements + _elem_size * (i - _low), t2, _elem_size);
      free(t2);
   }
   else
   {  grow(i);
      memcpy(_elements + _elem_size * (i - _low), t, _elem_size);
   }
}

/*.........................................................................*/

void Chi_NumArrayBase::insert(int i, char* t)
/* PURPOSE:    insert an element, shift others up
   RECEIVES:   i - the place to insert at
               t - the element to insert
*/
{  if (_elements <= t && t < _elements + _elem_size * (_high - _low + 1))
      // t points to an element in the array -- will relocate under grow
   {  void* t2 = malloc(_elem_size);
      CHI_ASSERT(t2);
      memcpy(t2, t, _elem_size);
      insert_range(i, 1);
      memcpy(_elements + _elem_size * (i - _low), t2, _elem_size);
      free(t2);
   }
   else
   {  insert_range(i, 1);
      memcpy(_elements + _elem_size * (i - _low), t, _elem_size);
   }
}

/*.........................................................................*/


void Chi_NumArrayBase::append(char* t)
/* PURPOSE:    append an element at the end of the array
   RECEIVES:   t - the element to insert
   REMARKS:    if the array is empty, the element is placed at index 0
*/
{  if (_elements <= t && t < _elements + _elem_size * (_high - _low + 1))
      // t points to an element in the array -- will relocate under grow
   {  void* t2 = malloc(_elem_size);
      CHI_ASSERT(t2);
      memcpy(t2, t, _elem_size);
      grow(_high + 1);
      memcpy(_elements + _elem_size * (_high - _low), t2, _elem_size);
      free(t2);
   }
   else
   {  grow(_high + 1);
      memcpy(_elements + _elem_size * (_high - _low), t, _elem_size);
   }
}

/*.........................................................................*/

void Chi_NumArrayBase::insert_range(int from, int n)
/* PURPOSE:    insert a range of default elements into an array, shifting the
               others up
   RECEIVES:   from - the index to start insertion
               n - the number of elements to insert
*/
{  if (n <= 0) return;
   if (_elements) 
   {  if (from > _high) // CSH 4-7-95
         ins(0, _high - _low + 1, from + n - 1 - _high, 0);
      else if (from < _low)
      {  if (from + n <= _low)
            ins(_low - from, _high - _low + 1, 0, 0);
         else
         {  int shift = from + n - _low;
            ins(n, _high - _low + 1, 0, 0);
            _high += shift;
            _low += shift;
         }
      }
      else
         ins(0, from - _low, n, _high - from + 1);
   }
   else
      init(from, from + n - 1);
}

/*.........................................................................*/

char* Chi_NumArrayBase::remove(int at, char* temp)
/* PURPOSE:    remove an element and shift the others down
   RECEIVES:   n - the position at which to remove the element
   RETURNS:    the removed element
*/
{  char* r = get(at, temp);
   if (r != temp) memcpy(temp, r, _elem_size);
   remove_range(at, 1);
   return temp;
}

/*.........................................................................*/

void Chi_NumArrayBase::remove_range(int from, int n)
/* PURPOSE:    remove a range of elements and shift the remainder down
   RECEIVES:   from - the index to start removal
               n - the number of elements to remove
*/
{  if (_elements)
   {  if (from < _low) { n -= _low - from; from = _low; }
      if (n <= 0) return;
      if (from > _high) return;
      if (n > _high - from + 1) n = _high - from + 1;
      rem(0, from - _low, n, _high - from + 1 - n);
   }
}

/*.........................................................................*/

void Chi_NumArrayBase::qsort(int (*comp)(const char&, const char&))
/* PURPOSE:    sort the array
   RECEIVES:   comp - pointer to comparison function
*/
{  qsort(comp, _low, _high - _low + 1);
}

/*.........................................................................*/

void Chi_NumArrayBase::qsort(int (*comp)(const char&, const char&), int from, int n)
/* PURPOSE:    sort a range of the array
   RECEIVES:   comp - pointer to comparison function
               from - the index at which to start sorting
               n - the number of elements to sort
   REMARKS:    Taken from Bentley, Programming Pearls, p. 112
*/
{  if (from < _low) from = _low;
   int to = from + n - 1;
   if (to > _high) to = _high;
   if (from < to)
   {  unique();
      int i = rand() % (to - from + 1) + from;
         // a random integer between from and to
      swap_block(_elements + _elem_size * from, _elements + _elem_size * i, _elem_size);
      int mid = from;
      for (i = from + 1; i <= to; i++)
      /* INVARIANT:  self[from + 1...mid] < self[from] &&
                     self[mid+1...i-1] >= self[from]
      */
      {  if ((*comp)(_elements[_elem_size * i], _elements[_elem_size * from]) < 0)
         {  mid++;
            swap_block(_elements + _elem_size * mid, _elements + _elem_size * i, _elem_size);
         }
      }
      swap_block(_elements + _elem_size * from, _elements + _elem_size * mid, _elem_size);
      /* self[from...mid-1] < self[m] <= self[m+1...to] */
      qsort(comp, from, mid - from);
      qsort(comp, mid+1, to - mid);
   }
}

/*.........................................................................*/

void Chi_NumArrayBase::init(int lo, int hi)
/* PURPOSE:    Allocate array storage from the heap
               and initialize it with the default constructor
   RECEIVES:   lo, hi - array range
*/
{  // allocate storage
   if (lo > hi)
   {  _elements = 0;
      _low = 0;
      _high = -1;
      return;
   }

   _low = lo;
   _high = hi;
   Chi_ArrayHeader* ph = Chi_ArrayHeader::alloc(_elem_size, hi - lo + 1, Chi_ArrayHeader::GROWBY);
   if (ph == 0) return;

   // initialize all array entries

   _elements = (char*)ph->start();
   construct_block(_elements, _elem_size * ph->block_size());
}

/*.........................................................................*/

void Chi_NumArrayBase::swap_block(char* p, char* q, int size)
{  for (int i = 0; i < size; i++)
   {  char t = *p;
      *p++ = *q;
      *q++ = t;
   }
}

/*.........................................................................*/

void Chi_NumArrayBase::construct_block(char* p, int size)
/* PURPOSE:    run the default constructor on a block of elements
   RECEIVES:   p - start of block
               size - number of elements
*/
{  memset(p, 0, size);
}

/*.........................................................................*/

void Chi_NumArrayBase::copy_block(char* p,  const char* from, int size)
/* PURPOSE:    run the copy constructor to make a copy of a block of elements
   RECEIVES:   p - start of target block
               from - start of source block
               size - number of elements
*/
{  memcpy(p, from, size);
}

/*.........................................................................*/

void Chi_NumArrayBase::move_block(char* tt, const char* ff, int size)
/* PURPOSE:    run the copy constructor and destructor to move a block of elements
   RECEIVES:   tt - start of target block
               from - start of source block
               size - number of elements
   REMARKS:    source and target may overlap
*/
{  memmove(tt, ff, size);
}

/*.........................................................................*/

void Chi_NumArrayBase::ins(int ins1, int keep1, int ins2, int keep2)
/* PURPOSE:    Grow the array
   RECEIVES:   ins1 - the #new entries to insert at the bottom of the array
               keep1 - the #entries to keep after that
               ins2 - the #new entries to insert after that
               keep2 - the #entries to keep after that
   RETURNS:    the new address
   REMARKS:    For grow, keep2 is 0
               For insert, ins1 is 0
   PRECOND:    keep1 + keep2 == _high - _low + 1, keep1, keep2 >= 0
*/
{  _low -= ins1; // CSH 4-7-95
   _high += ins2;

   int new_size = ins1 + keep1 + ins2 + keep2;
   Chi_ArrayHeader* ph = Chi_ArrayHeader::header(_elements);
   if (new_size <= ph->block_size() && ins1 == 0 && keep2 == 0) return;
   Bool uniq = ph->unique();
         // TRUE iff pp is the only reference to that information

   if (uniq && new_size <= ph->block_size()) // shift entries
   {  if (ins1 + ins2 > 0)
      {  // start on top since shifting up
         move_block(_elements + _elem_size * (ins1 + keep1 + ins2),
            _elements + _elem_size * keep1, _elem_size * keep2);
         if (ins2 > 0)
            construct_block(_elements + _elem_size * (ins1 + keep1),
               _elem_size * ins2);
         if (ins1 > 0)
         {  move_block(_elements + _elem_size * ins1, _elements,
               _elem_size * keep1);
            construct_block(_elements, _elem_size * ins1);
         }
      }
   }
   else // allocate new block
   {  ph->decref();
      Chi_ArrayHeader* nh = Chi_ArrayHeader::alloc(_elem_size, new_size, ph->growby());
      if (nh == 0) return;

      char* pp = _elements;
      _elements = (char*)nh->start();

      // initialize all new array entries

      if (ins1) construct_block(_elements, _elem_size * ins1);
      copy_block(_elements + _elem_size * ins1, pp, _elem_size * keep1);
      construct_block(_elements + _elem_size * (ins1 + keep1),
         _elem_size * ins2);
      if (keep2)  copy_block(_elements + _elem_size * (ins1 + keep1 + ins2),
         pp + _elem_size * keep1, _elem_size * keep2);
      construct_block(_elements + _elem_size * new_size,
         _elem_size * (nh->block_size() - new_size));

      if (uniq)
      {
         ph->free();
      }
   }
}

/*.........................................................................*/

void Chi_NumArrayBase::rem(int rem1, int keep1, int rem2, int keep2)
/* PURPOSE:    remove elements from the array
   RECEIVES:   rem1 - #elements to remove at bottom
               keep1 - #elements to keep after that
               rem2 - #elements to remove after that
               keep2 - #elements to keep after that
*/
{  _low += rem1; // CSH 4-7-95
   _high -= rem2;
   int new_size = keep1 + keep2;
   if (new_size == 0) { detach(); return; }
   Chi_ArrayHeader* ph = Chi_ArrayHeader::header(_elements);
   if (rem1 == 0 && rem2 == 0 && new_size > ph->block_size() - ph->growby() )
      return;

   Bool uniq = ph->unique();
      // TRUE iff pp is the only reference to that information

   if (uniq && new_size > ph->block_size() - ph->growby()) // shift entries
   {  if (rem1 > 0)
      {
         move_block(_elements, _elements + _elem_size * rem1, _elem_size * keep1);
      }

      move_block(_elements + _elem_size * keep1,
         _elements + _elem_size * (rem1 + keep1 + rem2),
         _elem_size * keep2);
      construct_block(_elements + _elem_size * (keep1 + keep2),
         _elem_size * (rem1 + rem2));
   }
   else // allocate new block
   {  ph->decref();
      Chi_ArrayHeader* nh = Chi_ArrayHeader::alloc(_elem_size, new_size, ph->growby());
      if (nh == 0) return;

      char* pp = _elements;
      _elements = (char*)nh->start();

      copy_block(_elements, pp + _elem_size * rem1, _elem_size * keep1);
      copy_block(_elements + _elem_size * keep1,
         pp + _elem_size * (rem1 + keep1 + rem2), _elem_size * keep2);
      construct_block(_elements + _elem_size * new_size,
         _elem_size * (nh->block_size() - new_size));
      if (uniq)
      {  ph->free();
      }
   }
}

/*-------------------------------------------------------------------------*/

template< class X >
X& Chi_NumArray<X>::tempval()
/* PURPOSE:    return a temporary storage location of type X
   REMARKS:    This should be a static variable, but Borland 3.1 doesn't
               handle static variables in templates well
*/
{  static X d;
   return d;
}

/*-------------------------------------------------------------------------*/

Chi_ArrayHeader* Chi_ArrayHeader::header(void* p)
/* PURPOSE:    Compute the start of the array header block from a pointer
               to an array in the collectible heap
   PRECOND:    p must point to the first element of an array allocated
               from Chi_ArrayHeader::alloc
*/
{  return (Chi_ArrayHeader*) (p == 0 ? 0 :
      (char*)p - sizeof(Chi_AlignedArrayHeader));
}

/*.........................................................................*/

void* Chi_ArrayHeader::start() const
/* PURPOSE:    Compute the address of the first element in an array allocated
               in the collectible heap
   PRECOND:    this header block must precede an array allocated
               from Chi_ArrayHeader::alloc
*/
{   return (char*)this + sizeof(Chi_AlignedArrayHeader);
}

/*.........................................................................*/

Chi_ArrayHeader* Chi_ArrayHeader::alloc(int xsize, int size, int growby)

/* PURPOSE:    Allocate array storage and set up type ID, lowAvail, highAvail
               Do not initialize array contents
   RECEIVES:   t - type ID of array objects
               lo, hi - array range
   RETURNS:    a pointer to the data area of the allocated storage
               0 if empty range or out of memory
   REMARKS:    The resulting array may be LARGER than size. The caller
               must check the _block_size entry and construct all elements
*/
{  if (size == 0) return 0;

   size = ((size - 1) / growby + 1) * growby;
      // that is ceil((double)size/growby)

   size_t len = xsize * size + sizeof(Chi_AlignedArrayHeader);
   Chi_ArrayHeader* ph = (Chi_ArrayHeader*) ::calloc(len, 1);
   CHI_ASSERT(ph != 0);

   ph->_rc = 1;
   ph->_block_size = size;
   ph->_growby = growby;

   return ph;
}

/*.........................................................................*/

void Chi_ArrayHeader::free()
/* PURPOSE:    return a collectible array to the collectible heap
   RECEIVES:   p - pointer to the data area to free
*/
{  CHI_ASSERT(_rc == 0);
   CHI_ASSERT(_block_size > 0);
   ::free(this);
}

/* TEST ********************************************************************/

#ifdef CHIARRAY_TEST

#include <iostream.h>
#include <assert.h>
#include "chistrng.h"

template< class X > Bool operator==(Chi_Array<X> a, Chi_Array<X> b)
{  if (a.low() != b.low()) return FALSE;
   if (a.high() != b.high()) return FALSE;
   for (int i = a.low(); i <= a.high(); i++)
      if (!(a.get(i) == b.get(i))) return FALSE;
   return TRUE;
}

/*.........................................................................*/

template<class X> Bool comp(Chi_Array<X> a, int afrom, Chi_Array<X> b, int bfrom, int len)
{  for (int i = 0; i < len; i++)
      if (!(a.get(afrom + i) == b.get(bfrom + i))) return FALSE;
   return TRUE;
}

/*.........................................................................*/

template<class X> Bool compdef(Chi_Array<X> a, int afrom, int len)
{  static X def;
   for (int i = 0; i < len; i++)
      if (!(a.get(afrom + i) == def)) return FALSE;
   return TRUE;
}

/*.........................................................................*/

template<class X> Bool test(Chi_Array<X> a)
{  Chi_Array< X > b = a;
   b[1] = a.get(1); // force copy
   assert(comp(a, 1, b, 1, 10));
   // grow one entry, with unique copy
   b.set(11, a.get(2));
   // grow one entry, with shared copy
   Chi_Array< X > c;
   c = b;
   b.set(21, a.get(3));
   assert(b.length() == 21);
   b.grow(30);
   assert(compdef(b, 22, 9));
   // grow entry below, with unique copy
   b.set(-10, a.get(4));
   assert(comp(a, 1, b, 1, 10));
   // grow entry below, with shared copy
   Chi_Array< X > d = b;
   d.set(-11, a.get(5));
   assert(comp(a, 1, d, 1, 10));
   // insert in the middle, with unique copy and room to grow
   d.insert_range(1, 2);
   assert(comp(a, 1, d, 3, 10));
   // insert in the middle, with unique copy and no room to grow
   d.insert_range(1, 8);
   assert(comp(a, 1, d, 11, 10));
   // insert in the middle, with shared copy and room to grow
   c = d;
   d.insert(1, a.get(6));
   assert(comp(a, 1, d, 12, 10));
   // insert in the middle, with shared copy and no room to grow
   c = d;
   d.insert_range(1, 9);
   assert(comp(a, 1, d, 21, 10));
   // remove in the middle, with unique copy and room to shrink
   d.remove_range(1, 2);
   assert(comp(a, 1, d, 19, 10));
   // remove in the middle, with unique copy and no room to shrink
   d.remove_range(1, 8);
   assert(comp(a, 1, d, 11, 10));
   // remove in the middle, with shared copy and room to shrink
   c = d;
   d.remove(1);
   assert(comp(a, 1, d, 10, 10));
   // remove in the middle, with shared copy and no room to shrink
   c = d;
   d.remove_range(1, 9);
   assert(comp(a, 1, d, 1, 10));
   // shrink array below, shared copy
   c = d;
   d.shrink(-5, 30);
   assert(comp(a, 1, d, 1, 10));
   // shrink array below, unique copy
   d.shrink(0, 30);
   assert(comp(a, 1, d, 1, 10));
   // shrink both, unique copy, changing size
   d.shrink(1, 25);
   assert(comp(a, 1, d, 1, 10));
   // shrink top, unique copy, keeping size
   d.shrink(1, 24);
   assert(comp(a, 1, d, 1, 10));
   // shrink both, shared copy, changing size
   c = d;
   d.shrink(1, 10);
   assert(comp(a, 1, d, 1, 10));
   // call relocating set with a reference into own array
   d.set(20, d[5]);
   assert(d[20] == d[5]);
   cout << "test passed" << endl;
   return TRUE;
}

/*.........................................................................*/

template< class X > Bool operator==(Chi_PtrArray<X> a, Chi_PtrArray<X> b)
{  if (a.low() != b.low()) return FALSE;
   if (a.high() != b.high()) return FALSE;
   for (int i = a.low(); i <= a.high(); i++)
      if (!(a.get(i) == b.get(i))) return FALSE;
   return TRUE;
}

/*.........................................................................*/

template<class X> Bool comp(Chi_PtrArray<X> a, int afrom, Chi_PtrArray<X> b, int bfrom, int len)
{  for (int i = 0; i < len; i++)
      if (!(a.get(afrom + i) == b.get(bfrom + i))) return FALSE;
   return TRUE;
}

/*.........................................................................*/

template<class X> Bool compdef(Chi_PtrArray<X> a, int afrom, int len)
{  for (int i = 0; i < len; i++)
      if (!(a.get(afrom + i) == 0)) return FALSE;
   return TRUE;
}

/*.........................................................................*/

template<class X> Bool test(Chi_PtrArray<X> a)
{  Chi_PtrArray< X > b = a;
   b[1] = a.get(1); // force copy
   assert(comp(a, 1, b, 1, 10));
   // grow one entry, with unique copy
   b.set(11, a.get(2));
   // grow one entry, with shared copy
   Chi_PtrArray< X > c;
   c = b;
   b.set(21, a.get(3));
   assert(b.length() == 21);
   b.grow(30);
   assert(compdef(b, 22, 9));
   // grow entry below, with unique copy
   b.set(-10, a.get(4));
   assert(comp(a, 1, b, 1, 10));
   // grow entry below, with shared copy
   Chi_PtrArray< X > d = b;
   d.set(-11, a.get(5));
   assert(comp(a, 1, d, 1, 10));
   // insert in the middle, with unique copy and room to grow
   d.insert_range(1, 2);
   assert(comp(a, 1, d, 3, 10));
   // insert in the middle, with unique copy and no room to grow
   d.insert_range(1, 8);
   assert(comp(a, 1, d, 11, 10));
   // insert in the middle, with shared copy and room to grow
   c = d;
   d.insert(1, a.get(6));
   assert(comp(a, 1, d, 12, 10));
   // insert in the middle, with shared copy and no room to grow
   c = d;
   d.insert_range(1, 9);
   assert(comp(a, 1, d, 21, 10));
   // remove in the middle, with unique copy and room to shrink
   d.remove_range(1, 2);
   assert(comp(a, 1, d, 19, 10));
   // remove in the middle, with unique copy and no room to shrink
   d.remove_range(1, 8);
   assert(comp(a, 1, d, 11, 10));
   // remove in the middle, with shared copy and room to shrink
   c = d;
   d.remove(1);
   assert(comp(a, 1, d, 10, 10));
   // remove in the middle, with shared copy and no room to shrink
   c = d;
   d.remove_range(1, 9);
   assert(comp(a, 1, d, 1, 10));
   // shrink array below, shared copy
   c = d;
   d.shrink(-5, 30);
   assert(comp(a, 1, d, 1, 10));
   // shrink array below, unique copy
   d.shrink(0, 30);
   assert(comp(a, 1, d, 1, 10));
   // shrink both, unique copy, changing size
   d.shrink(1, 25);
   assert(comp(a, 1, d, 1, 10));
   // shrink top, unique copy, keeping size
   d.shrink(1, 24);
   assert(comp(a, 1, d, 1, 10));
   // shrink both, shared copy, changing size
   c = d;
   d.shrink(1, 10);
   assert(comp(a, 1, d, 1, 10));
   // call relocating set with a reference into own array
   d.set(20, d[5]);
   assert(d[20] == d[5]);
   cout << "test passed" << endl;
   return TRUE;
}

/*.........................................................................*/

template< class X > Bool operator==(Chi_NumArray<X> a, Chi_NumArray<X> b)
{  if (a.low() != b.low()) return FALSE;
   if (a.high() != b.high()) return FALSE;
   for (int i = a.low(); i <= a.high(); i++)
      if (!(a.get(i) == b.get(i))) return FALSE;
   return TRUE;
}

/*.........................................................................*/

template<class X> Bool comp(Chi_NumArray<X> a, int afrom, Chi_NumArray<X> b, int bfrom, int len)
{  for (int i = 0; i < len; i++)
      if (!(a.get(afrom + i) == b.get(bfrom + i))) return FALSE;
   return TRUE;
}

/*.........................................................................*/

template<class X> Bool compdef(Chi_NumArray<X> a, int afrom, int len)
{  static X def;
   for (int i = 0; i < len; i++)
      if (!(a.get(afrom + i) == def)) return FALSE;
   return TRUE;
}

/*.........................................................................*/

template<class X> Bool test(Chi_NumArray<X> a)
{  Chi_NumArray< X > b = a;
   b[1] = a.get(1); // force copy
   assert(comp(a, 1, b, 1, 10));
   // grow one entry, with unique copy
   b.set(11, a.get(2));
   // grow one entry, with shared copy
   Chi_NumArray< X > c;
   c = b;
   b.set(21, a.get(3));
   assert(b.length() == 21);
   b.grow(30);
   assert(compdef(b, 22, 9));
   // grow entry below, with unique copy
   b.set(-10, a.get(4));
   assert(comp(a, 1, b, 1, 10));
   // grow entry below, with shared copy
   Chi_NumArray< X > d = b;
   d.set(-11, a.get(5));
   assert(comp(a, 1, d, 1, 10));
   // insert in the middle, with unique copy and room to grow
   d.insert_range(1, 2);
   assert(comp(a, 1, d, 3, 10));
   // insert in the middle, with unique copy and no room to grow
   d.insert_range(1, 8);
   assert(comp(a, 1, d, 11, 10));
   // insert in the middle, with shared copy and room to grow
   c = d;
   d.insert(1, a.get(6));
   assert(comp(a, 1, d, 12, 10));
   // insert in the middle, with shared copy and no room to grow
   c = d;
   d.insert_range(1, 9);
   assert(comp(a, 1, d, 21, 10));
   // remove in the middle, with unique copy and room to shrink
   d.remove_range(1, 2);
   assert(comp(a, 1, d, 19, 10));
   // remove in the middle, with unique copy and no room to shrink
   d.remove_range(1, 8);
   assert(comp(a, 1, d, 11, 10));
   // remove in the middle, with shared copy and room to shrink
   c = d;
   d.remove(1);
   assert(comp(a, 1, d, 10, 10));
   // remove in the middle, with shared copy and no room to shrink
   c = d;
   d.remove_range(1, 9);
   assert(comp(a, 1, d, 1, 10));
   // shrink array below, shared copy
   c = d;
   d.shrink(-5, 30);
   assert(comp(a, 1, d, 1, 10));
   // shrink array below, unique copy
   d.shrink(0, 30);
   assert(comp(a, 1, d, 1, 10));
   // shrink both, unique copy, changing size
   d.shrink(1, 25);
   assert(comp(a, 1, d, 1, 10));
   // shrink top, unique copy, keeping size
   d.shrink(1, 24);
   assert(comp(a, 1, d, 1, 10));
   // shrink both, shared copy, changing size
   c = d;
   d.shrink(1, 10);
   assert(comp(a, 1, d, 1, 10));
   // call relocating set with a reference into own array
   d.set(20, d[5]);
   assert(d[20] == d[5]);
   cout << "test passed" << endl;
   return TRUE;
}

/*.........................................................................*/

int main()
{  Chi_Array<Chi_String> a(1,10);
   int i;
   for (i = 1; i <= 10; i++)
      a[i].from_double(i * i);
   test(a);
   assert(a == a);
   Chi_Array< Chi_Array< Chi_String > > aa(1, 10);
   for (i = 1; i <= 10; i++)
   {  aa[i].grow(1, 10);
      for (int j = 1; j <= 10; j++)
         aa[i][j].from_double(100*i+j);
   }
   assert(aa == aa); // needed for template instantiation

   test(aa);

   char s[] = "This is a test";

   Chi_PtrArray<char> pa(1,10);
   for (i = 1; i <= 10; i++)
      pa[i] = s + i;
   test(pa);
   assert(pa == pa);

   Chi_NumArray<double> na(1,10);
   for (i = 1; i <= 10; i++)
      na[i] = i * i;
   test(na);
   assert(na == na);
   Chi_Array< Chi_NumArray< double > > naa(1, 10);
   for (i = 1; i <= 10; i++)
   {  naa[i].grow(1, 10);
      for (int j = 1; j <= 10; j++)
         naa[i][j] = 100*i+j;
   }
   test(naa);
   assert(naa == naa);

   return 0;
}

#endif

/***************************************************************************/
