/****************************************************************************
** COPYRIGHT (C):    1992 Cay S. Horstmann. All Rights Reserved.
** PROJECT:          ChiLib
** FILE:             chidate.cpp
** PURPOSE:          Date class
** VERSION           1.0
** LANGUAGE:         Borland  C++ 3.0
** TARGET:           IBM PC
** PROGRAMMER:       Cay Horstmann (CSH)
** START DATE:       9-30-93  (CSH)
** UPDATE HISTORY:

2-7-94 CSH Added Date::compare, relational operators

10-3-95 CSH Fixed days_per_month, August has 31 days and September 30

**
****************************************************************************/

#ifndef EXPORT
#include "chisetup.h"
#endif

EXPORT
#include <iostream.h>
#include <math.h>

#include "chierror.h"
#include "chidate.h"

/* TYPES *******************************************************************/

#ifndef EXPORT

class Chi_Date
/* PURPOSE:    Store dates and perform date arithmetic
*/
{
public:
   enum Weekday { MON, TUE, WED, THU, FRI, SAT, SUN };

   Chi_Date();
   Chi_Date(int d, int m, int y); // construct from day, month, year

   void advance(long n); // advance this date by n days

   int day() const;
   int month() const;
   int year() const;
   Weekday weekday() const;
   Chi_Date add_days(long n) const; // the date that is n days away from this date
   long days_between(const Chi_Date& b) const; // number of days between this date and b
   int compare(const Chi_Date& b) const;

   static int compare(const Chi_Date& a, const Chi_Date& b);
   static int hash(const Chi_Date& a);

   Bool operator<(const Chi_Date& b) const;
   Bool operator<=(const Chi_Date& b) const;
   Bool operator==(const Chi_Date& b) const;
   Bool operator!=(const Chi_Date& b) const;
   Bool operator>(const Chi_Date& b) const;
   Bool operator>=(const Chi_Date& b) const;

private:
   Bool is_valid() const;
   static Bool is_leap(int year);

   int _day;
   int _month;
   int _year;

   static int days_per_month[12];
};

#endif

/* GLOBALS *****************************************************************/

const int JULYEAR0 = -4713;
   // Julian day 0 = Jan. 1, -4713

int Chi_Date::days_per_month[12] =
{  31, 28, 31, 30, 31, 30,
   31, 31, 30, 31, 30, 31
};

/* FUNCTIONS ***************************************************************/

static long dat2jul(int d, int m, int y)
/* PURPOSE:    Convert calendar date into Julian day
   RECEIVES:   d, m, y - the day, month and year
   RETURNS:    The Julian day number that begins at noon of the given calendar
               date.
   REMARKS:    Positive year signifies A.D., negative year B.C. Remember that
               the year after 1 B.C. was 1 A.D.

               A convenient reference point is that May 23, 1968 noon is
               Julian day 2440000.

               Julian day 0 is a Monday.

               This algorithm is from Press et al., Numerical Recipes in C,
               2nd ed., Cambridge University Press 1992
*/
{  CHI_ASSERT_PRECOND(y != 0); // there is no year 0
   int jy = y;
   if (y < 0) jy++;
   int jm = m;
   if (m > 2) jm++;
   else
   {  jy--;
      jm += 13;
   }
   long jul = (long) (floor(365.25 * jy) + floor(30.6001*jm) + d + 1720995.0);

   const long IGREG = 15 + 31L*(10+12L*1582);
      // Gregorian Calendar adopted Oct. 15, 1582

   if (d + 31L * (m + 12L * y) >= IGREG)
   // change over to Gregorian calendar
   {  int ja = (int)(0.01 * jy);
      jul += 2 - ja + (int)(0.25 * ja);
   }
   return jul;
}

/*.........................................................................*/

static void jul2dat(long j, int& d, int& m, int& y)
/* PURPOSE:    Converts a Julian day to a calendar date
   RECEIVES:   j - the Julian date
               d, m, y (OUT) - the date on which the specified Julian
               day started at noon.
   REMARKS:    This algorithm is from Press et al., Numerical Recipes in C,
               2nd ed., Cambridge University Press 1992
*/
{  long ja = j;

   const long JGREG = 2299161L;
      // the Julian date of the adoption of the Gregorian calendar

   if (j >= JGREG)
   // cross-over to Gregorian Calendar produces this correction
   {  long jalpha = (long)(((float)(j - 1867216L) - 0.25) / 36524.25);
      ja += 1 + jalpha - (long)(0.25 * jalpha);
   }
   long jb = ja + 1524;
   long jc = (long)(6680.0 + ((float)(jb-2439870L) - 122.1)/365.25);
   long jd = (long)(365 * jc + (0.25 * jc));
   int je = (int)((jb - jd)/30.6001);
   d = (int)(jb - jd - (long)(30.6001 * je));
   m = je - 1;
   if (m > 12) m -= 12;
   y = (int)(jc - 4715);
   if (m > 2) --y;
   if (y <= 0) --y;
}

/*.........................................................................*/

EXPORT istream& operator>>(istream& is, Chi_Date& date)
/* PURPOSE:    Read a date from a stream
   RECEIVES:   is - the stream
               d (OUT) - the date that is read in
   RETURNS:    is
   REMARKS:    This function expects dates in the U.S. format month/day/year
*/
{  int d, m, y;
   char ch;
   is >> m;
   if (!is) return is;
   is >> ch;
   if (!is || ch != '/') { is.clear(is.rdstate() | ios::failbit); return is; }
   is >> d;
   if (!is) return is;
   is >> ch;
   if (!is || ch != '/') { is.clear(is.rdstate() | ios::failbit); return is; }
   is >> y;
   if (!is) return is;
   date = Chi_Date(d, m, y);
   return is;
}

/*.........................................................................*/

EXPORT ostream& operator<<(ostream& os, const Chi_Date& date)
/* PURPOSE:    Write a date to a stream
   RECEIVES:   os - the stream
               date - the date to print
   RETURNS:    os
   REMARKS:    This function prints dates in the U.S. format month/day/year
*/
{  char ch = os.fill('0');
   os.width(2); os << date.month() << "/";
   os.width(2); os << date.day() << "/";
   os.width(4); os << date.year();
   os.fill(ch);
   return os;
}

/*-------------------------------------------------------------------------*/

Chi_Date::Chi_Date()
:  _day(1),
   _month(1),
   _year(JULYEAR0)
/* REMARKS:    makes the date with Julian day number 0
*/
{}

/*.........................................................................*/

Chi_Date::Chi_Date(int d, int m, int y)
/* RECEIVES:   d, m, y - the day, month and year
*/
:  _day(d),
   _month(m),
   _year(y)
{  CHI_ASSERT_PRECOND(is_valid());
}

/*.........................................................................*/

Bool Chi_Date::is_valid() const
/* PURPOSE:    test whether the current day/month/year make a valid date
   RETURNS:    TRUE iff this is a valid date
*/
{  if (_day <= 0) return FALSE;
   if (_month <= 0 || _month > 12) return FALSE;
   if (_year == 0) return FALSE;
   if (_month == 2 && _day == 29) return is_leap(_year);
   return _day <= days_per_month[_month - 1];
}

/*.........................................................................*/

Chi_Date Chi_Date::add_days(long n) const
/* PURPOSE:    Compute a date that is a given number of days away
   RECEIVES:   n - difference in days
   RETURNS:    The date n days away from this date
*/
{  long j = dat2jul(_day, _month, _year);
   j += n;
   int d, m, y;
   jul2dat(j, d, m, y);
   return Chi_Date(d, m, y);
}

/*.........................................................................*/

long Chi_Date::days_between(const Chi_Date& b) const
/* PURPOSE:    Compute the number of days between two dates
   RECEIVES:   b - a date
   RETURNS:    The number of days between this day and b
*/
{  long j = dat2jul(_day, _month, _year);
   long k = dat2jul(b._day, b._month, b._year);
   return j - k;
}

//............................................................................

int Chi_Date::compare(const Chi_Date& b) const
/* PURPOSE:    compares two date
   RECEIVES:   b - date to compare with self
   RETURNS:    < 0 if self before b
               0 if self equals b
               > 0 if self after b
*/
{  int d = _year - b._year;
   if (d != 0) return d;
   d = _month - b._month;
   if (d != 0) return d;
   return _day - b._day;
}
/*.........................................................................*/

int Chi_Date::compare(const Chi_Date& a, const Chi_Date& b)
/* PURPOSE:    compares two date
   RECEIVES:   a, b - dates to compare
   RETURNS:    < 0 if a before b
               0 if a equals b
               > 0 if a after b
*/
{  return a.compare(b);
}

/*.........................................................................*/

int Chi_Date::hash(const Chi_Date& a)
/* PURPOSE:    compute a hash value
   RECEIVES:   a - a date
   RETURNS:    a hash value for a
*/
{  return 13 * a._year + 37 * a._month + a._day;
}

/*.........................................................................*/

void Chi_Date::advance(long n)
/* PURPOSE:    Advance this date by n days
   RECEIVES:   n - difference in days
*/
{  long j = dat2jul(_day, _month, _year);
   j += n;
   jul2dat(j, _day, _month, _year);
}

/*.........................................................................*/

#ifndef EXPORT

inline int Chi_Date::day() const
/* RETURNS:    the day of this date
*/
{  return _day;
}

/*.........................................................................*/

inline int Chi_Date::month() const
/* RETURNS:    the month of this date
*/
{  return _month;
}

/*.........................................................................*/

inline int Chi_Date::year() const
/* RETURNS:    the year of this date
*/
{  return _year;
}

/*.........................................................................*/

inline Bool Chi_Date::operator<(const Chi_Date& b) const
{  return compare(b) < 0;
}

/*.........................................................................*/

inline Bool Chi_Date::operator<=(const Chi_Date& b) const
{  return compare(b) <= 0;
}

/*.........................................................................*/

inline Bool Chi_Date::operator==(const Chi_Date& b) const
{  return compare(b) == 0;
}

/*.........................................................................*/

inline Bool Chi_Date::operator!=(const Chi_Date& b) const
{  return compare(b) != 0;
}

/*.........................................................................*/

inline Bool Chi_Date::operator>(const Chi_Date& b) const
{  return compare(b) > 0;
}

/*.........................................................................*/

inline Bool Chi_Date::operator>=(const Chi_Date& b) const
{  return compare(b) >= 0;
}

#endif

/*.........................................................................*/

Chi_Date::Weekday Chi_Date::weekday() const
/* RETURNS:    the weekday of this date
*/
{  return (Weekday)(dat2jul(_day, _month, _year) % 7);
}

/*.........................................................................*/

Bool Chi_Date::is_leap(int year)
{  if (year % 4 != 0) return FALSE;
   if (year % 100 != 0) return TRUE;
   if (year % 400 != 0) return FALSE;
   return TRUE;
};

/* TEST ********************************************************************/

#ifdef CHIDATE_TEST

#include "chiargs.h"

int main(int argc, char* argv[])
{  Chi_Args args(argc, argv);
   int m;
   int y;
   if (argc == 3)
   {  m = (int) args.arg(1).to_long();
      y = (int) args.arg(2).to_long();
   }
   else
   {  cout << "Usage: " << args.prog_name() << " month year" << endl;
      cout << "Using January 1900." << endl;
      m = 1;
      y = 1900;
   }

   Chi_Date d(1, m, y); // start date of the month

   cout << "Mon Tue Wed Thu Fri Sat Sun" << endl;
   Chi_Date::Weekday w = d.weekday();
   for( int i = 0; i < w; i++ ) cout << "    ";
   while (d.month() == m)
   {  cout.width(3);
      cout << d.day();
      if (w == Chi_Date::SUN) { cout << endl; w = Chi_Date::MON;}
      else { cout << " "; w = (Chi_Date::Weekday)(w + 1); }
      d.advance(1);
   }
   if (w != Chi_Date::MON) cout << endl;
   return 0;
}

#endif
