/****************************************************************************
** COPYRIGHT (C):    1994 Cay S. Horstmann. All Rights Reserved.
** PROJECT:          ChiLib
** FILE:             chierror.cpp
** PURPOSE:          Error handling
** VERSION           1.0
** PROGRAMMER:       Cay Horstmann (CSH)
** START DATE:       4-26-94 (CSH)
** UPDATE HISTORY:
****************************************************************************/

/* IMPORT ******************************************************************/

#ifndef EXPORT
#include "chisetup.h"
#endif

EXPORT
#include <iostream.h>
#include <stdlib.h>

#include "chierror.h"

/* CLASSES *****************************************************************/

#ifndef EXPORT

class Chi_Error
/* PURPOSE:    report an error
*/
{
public:
   Chi_Error(const char* f, int l, const char* expr = 0, const char* msg = 0);
   void abort();
   virtual void message(ostream&) const;

private:
   const char* _msg;
   const char* _expr;
   const char* _file;
   int _line;
};

/*-------------------------------------------------------------------------*/

class Chi_BoundsError : public Chi_Error
/* PURPOSE:    report an array or string bounds error
*/
{
public:
   Chi_BoundsError(const char* f, int l, int i, int lo, int hi);
   virtual void message(ostream&) const;
private:
   int _actual;
   int _low;
   int _high;
};

/*-------------------------------------------------------------------------*/

class Chi_PrecondError : public Chi_Error
/* PURPOSE:    report a precondition failure
*/
{
public:
   Chi_PrecondError(const char* f, int l, const char* expr);
};

/*-------------------------------------------------------------------------*/

class Chi_PostcondError : public Chi_Error
/* PURPOSE:    report a postondition failure
*/
{
public:
   Chi_PostcondError(const char* f, int l, const char* expr);
};

/*-------------------------------------------------------------------------*/

class Chi_InvariantError : public Chi_Error
/* PURPOSE:    report an invariant failure
*/

{
public:
   Chi_InvariantError(const char* f, int l, const char* expr);
};

/*-------------------------------------------------------------------------*/

class Chi_MemError : public Chi_Error
/* PURPOSE:    report memory exhaustion
*/
{
public:
   Chi_MemError(const char* f, int l, const char* expr);
};

#endif

/* FUNCTIONS ***************************************************************/

#ifndef EXPORT

#ifdef CHI_THROW_ERRORS
#define CHI_THROW(E) (throw E)
#else
#define CHI_THROW(E) (E.abort())
#endif

#define CHI_ASSERT(E) ((E) ? (void)0 : CHI_THROW(Chi_Error(__FILE__, __LINE__, #E)))
#define CHI_ASSERT_PRECOND(E) ((E) ? (void)0 : CHI_THROW(Chi_PrecondError(__FILE__, __LINE__, #E)))
#define CHI_ASSERT_POSTCOND(E) ((E) ? (void)0 : CHI_THROW(Chi_PostcondError(__FILE__, __LINE__, #E)))
#define CHI_ASSERT_INVARIANT(E) ((E) ? (void)0 : CHI_THROW(Chi_InvariantError(__FILE__, __LINE__, #E)))
#define CHI_ASSERT_MEM(E) ((E) ? (void)0 : CHI_THROW(Chi_MemError(__FILE__, __LINE__, #E)))
#define CHI_ASSERT_BOUNDS(I, L, H) (chi_check_bounds(__FILE__, __LINE__, I, L, H))

#endif

EXPORT void chi_check_bounds(const char* f, int l, int i, int lo, int hi)
/* PURPOSE:    check for bounds error
   RECEIVES:   f - file name
               l - line number
               i - actual index
               lo, hi - low, high bounds
*/
{  if (lo <= i && i <= hi) return;
   CHI_THROW(Chi_BoundsError(f, l, i, lo, hi));
}

/* OPERATIONS **************************************************************/

Chi_BoundsError::Chi_BoundsError(const char* f, int l, int i, int lo, int hi)
/* RECEIVES:   f - file name
               l - line number
               i - actual index
               lo, hi - low, high bounds
*/
:  Chi_Error(f, l, 0, "Index bounds error"),
   _actual(i),
   _low(lo),
   _high(hi)
{}

/*.........................................................................*/

void Chi_BoundsError::message(ostream& os) const
/* PURPOSE:    report the error to a stream
   RECEIVES:   os - output stream
*/
{  Chi_Error::message(os);
   os << "Index = " << _actual << " Range = " << _low << " ... "
      << _high;
}

/*-------------------------------------------------------------------------*/

Chi_PrecondError::Chi_PrecondError(const char* f, int l, const char* expr)
/* RECEIVES:   f - file name
               l - line number
               expr - expression generating failure
*/
:  Chi_Error(f, l, expr, "Precondition error")
{}

/*-------------------------------------------------------------------------*/

Chi_PostcondError::Chi_PostcondError(const char* f, int l, const char* expr)
/* RECEIVES:   f - file name
               l - line number
               expr - expression generating failure
*/
:  Chi_Error(f, l, expr, "Postcondition error")
{}

/*-------------------------------------------------------------------------*/

Chi_InvariantError::Chi_InvariantError(const char* f, int l, const char* expr)
/* RECEIVES:   f - file name
               l - line number
               expr - expression generating failure
*/
:  Chi_Error(f, l, expr, "Invariant error")
{}

/*-------------------------------------------------------------------------*/

Chi_MemError::Chi_MemError(const char* f, int l, const char* expr)
/* RECEIVES:   f - file name
               l - line number
               expr - expression generating failure
*/
:  Chi_Error(f, l, expr, "Memory allocation error")
{}

/*-------------------------------------------------------------------------*/

Chi_Error::Chi_Error(const char* f, int l, const char* e, const char* m)
/* RECEIVES:   f - file name
               l - line number
               expr - expression generating failure
               m - error message
*/
:  _file(f),
   _line(l),
   _expr(e),
   _msg(m)
{}

/*.........................................................................*/

void Chi_Error::message(ostream& os) const
/* PURPOSE:    report the error to a stream
   RECEIVES:   os - output stream
*/
{  if (_msg) os << _msg;
   else if (_expr) os << "Assertion failure";
   os << "\nFile " << _file << " line " << _line << " ";
   if (_expr) os << _expr;
}
/*.........................................................................*/

void Chi_Error::abort()
/* PURPOSE:    report error and abort program
*/
{  message(cerr);
   cerr << endl;
   // set a breakpoint here and watch the stack
   exit(1);
}

/* TEST ********************************************************************/

#ifdef CHIERROR_TEST

int main()
{  cout << "Testing error messages." << endl;
   Chi_BoundsError bounds(__FILE__, __LINE__, 0, 1, 10);
   bounds.message(cout); cout << endl;
   Chi_PrecondError pre(__FILE__, __LINE__, "i >= 0");
   pre.message(cout); cout << endl;
   Chi_PostcondError post(__FILE__, __LINE__, "r >= 0");
   post.message(cout); cout << endl;
   Chi_InvariantError inv(__FILE__, __LINE__, "1 <= _m && _m <= 12");
   inv.message(cout); cout << endl;
   cout << "The following assertion will fail." << endl;
   CHI_ASSERT(0 == 1);
   return 0;
}

#endif

/***************************************************************************/
