/****************************************************************************
** COPYRIGHT (C):    1994 Cay S. Horstmann. All Rights Reserved.
** PROJECT:          ChiLib
** FILE:             chigcbgi.cpp
** PURPOSE:          Graphics context for BGI graphics
** VERSION           1.0
** LANGUAGE:         Borland  C++ 3.1/4.0
** TARGET:           IBM PC
** PROGRAMMER:       Cay Horstmann (CSH)
** START DATE:       2-19-94  (CSH)
** UPDATE HISTORY:
**
****************************************************************************/

/* IMPORT ******************************************************************/

#define CHI_BGI

#ifndef EXPORT
#include "chisetup.h"
#endif

#include <math.h>
#include <stdlib.h>
#include <assert.h>
#include <iostream.h>

#include <dos.h>
#include <bios.h>
#include <graphics.h>

EXPORT
#include "chistrng.h"

#include "chigcbgi.h"

EXPORT int chi_main(int, char**);

/* CLASSES *****************************************************************/

#ifndef EXPORT

class Chi_GraphicsContext;

class Chi_MouseEvent
/* PURPOSE:    a mouse click or move
*/
{
public:
   enum Type { NOT_INSTALLED, LBUTTONDOWN, MOUSEMOVE, LBUTTONUP };
   Chi_MouseEvent(Type, double, double);

   Type type() const { return _type; }
   double x() const { return _x; }
   double y() const { return _y; }
private:
   Type _type;
   double _x;
   double _y;
};

/*-------------------------------------------------------------------------*/

class Chi_Mouse
/* PURPOSE: Interfaces with mouse to get mouse position and events
*/
{
public:
   Chi_Mouse();

   Chi_MouseEvent read(const Chi_GraphicsContext& gc);

   Bool is_installed() const;

   int get_position(int&, int&) const;
   void show_cursor();
   void hide_cursor();


private:
   enum
   {  MOUSE_INTR = 0x33,
      RETF = 0xCF, // far return instruction
      SHOWCURSOR = 1,
      HIDECURSOR = 2,
      READMOUSE = 3,
   };
   Bool _installed;
};

/*-------------------------------------------------------------------------*/

class Chi_GraphicsContext
/* PURPOSE: Sends graphics primitives to a display window
*/
{
public:
   enum Color
   {  BLACK, BLUE, GREEN, CYAN,
      RED, MAGENTA, BROWN, LIGHTGRAY, /* dark colors */
      DARKGRAY, LIGHTBLUE, LIGHTGREEN, LIGHTCYAN,
      LIGHTRED, LIGHTMAGENTA, YELLOW, WHITE /* light colors */,
      DEFAULT_COLOR = WHITE
   };
   enum PenStyle { NORMAL_PEN, THICK_PEN, DOTTED_PEN, DASHED_PEN };
   enum BrushStyle { EMPTY_BRUSH, SOLID_BRUSH,
      HORIZ_BRUSH, FDIAG_BRUSH, BDIAG_BRUSH, XDIAG_BRUSH };
   enum FontStyle { SYS_FONT, SANS_FONT, ROMAN_FONT };
   enum RopMode { COPY_ROP, XOR_ROP, AND_ROP, NOT_ROP, OR_ROP };

   Chi_GraphicsContext();
   void close(Bool wait = TRUE);

   void point(double x, double y);
   void line(double xfrom, double yfrom, double xto, double yto);
   void arc(double x, double y, double ra, double rb, double x1, double y1,
      double x2, double y2);
   void rectangle(double xleft, double ytop, double xright, double ybot);
   void ellipse(double x, double y, double ra, double rb);
   void piesegment(double x, double y, double ra, double rb, double x1, double y1,
      double x2, double y2);

   void text(Chi_String t, double x, double y);
   void text_extent(Chi_String t, double& x, double& y);

   void set_brush_style(BrushStyle);
   void set_brush_color(Color);
   void set_pen_style(PenStyle);
   void set_pen_color(Color);
   void set_font_size(double s);
   void set_font_style(FontStyle);
   void set_rop_mode(RopMode);

   void set_window_coord(double xmin, double ymin, double xmax, double ymax);
   void set_clip_rectangle(double xmin, double ymin, double xmax, double ymax);
   void clear();
   void mouse_events(int (*h)(const Chi_MouseEvent&, void*), void* p);

private:
   int user_to_disp_x(double x) const;
   int user_to_disp_y(double y) const;
   int user_to_disp_dx(double x) const;
   int user_to_disp_dy(double y) const;

   double disp_to_user_dx(int x) const;
   double disp_to_user_dy(int y) const;

   double _user_xmin;
   double _user_xmax;
   double _user_ymin;
   double _user_ymax;

   int _disp_xmax;
   int _disp_ymax;
   int _disp_xoff;
   int _disp_yoff;

   BrushStyle _brush;
   FontStyle _font;
   double _font_size;

   Chi_Mouse _mouse;
friend class Chi_Mouse;
};

#endif

/* FUNCTIONS ***************************************************************/

int main(int argc, char** argv)
{  return chi_main(argc, argv);
}

/*-------------------------------------------------------------------------*/

Chi_GraphicsContext::Chi_GraphicsContext()
:  _brush(EMPTY_BRUSH),
   _font_size(0),
   _font(SYS_FONT)
{  const double DEFAULT_XMIN = 0;
   const double DEFAULT_YMIN = 0;
   const double DEFAULT_XMAX = 20;
   const double DEFAULT_YMAX = 15;

   int gdriver = DETECT;
   int gmode;
   initgraph(&gdriver, &gmode, "");
   int errorcode = graphresult();
   if (errorcode != grOk)
   {  cerr << "bgi_init: Can't initialize graphics. Is the .BGI file present?"
         << endl;
      exit(1);
   }

   _disp_xmax = getmaxx();
   _disp_ymax = getmaxy();
   _disp_xoff = 0;
   _disp_yoff = 0;
   set_window_coord(DEFAULT_XMIN, DEFAULT_YMIN, DEFAULT_XMAX, DEFAULT_YMAX);
   set_font_style(SYS_FONT);
   set_pen_style(NORMAL_PEN);
   set_pen_color(DEFAULT_COLOR);
   set_brush_style(EMPTY_BRUSH);
   set_brush_color(DEFAULT_COLOR);
}

//............................................................................

void Chi_GraphicsContext::close(Bool wait)
/* PURPOSE:    Shut down Borland graphics system
   RECEIVES:   wait - TRUE if we need to wait for the user to hit a key
               so they can admire the image
*/
{  if (wait)
      while (!bioskey(0))
         /* wait */ ;
   closegraph();
}

//............................................................................

void Chi_GraphicsContext::set_window_coord(double xmin, double ymin,
      double xmax, double ymax)
/* PURPOSE:    Set the logical coordinate system
   RECEIVES:   xmin, ymin - coordinates of the top left corner
               xmax, ymax - coordinates of the bottom right corner
*/
{  _user_xmin = xmin;
   _user_xmax = xmax;
   _user_ymin = ymin;
   _user_ymax = ymax;
}

//............................................................................

int Chi_GraphicsContext::user_to_disp_x(double x) const
/* PURPOSE:    scales the x coordinate from user to display coordinates
   RECEIVES:   x - a coordinate position in user coordinates
   RETURNS:    the scaled coordinate
*/
{  return (int)((x - _user_xmin) * _disp_xmax / (_user_xmax - _user_xmin))
      - _disp_xoff;
}

//............................................................................

int Chi_GraphicsContext::user_to_disp_y(double y) const
/* PURPOSE:    scales the y coordinate from user to display coordinates
   RECEIVES:   y - a coordinate position in user coordinates
   RETURNS:    the scaled coordinate
*/
{  return (int)((y - _user_ymin) * _disp_ymax / (_user_ymax - _user_ymin))
      - _disp_yoff;
}

//............................................................................

int Chi_GraphicsContext::user_to_disp_dx(double x) const
/* PURPOSE:    scales an x distance from user to display coordinates
   RECEIVES:   x - a distance in user coordinates
   RETURNS:    the scaled distance
*/
{  return (int)(x * _disp_xmax / (_user_xmax - _user_xmin));
}

//............................................................................

int Chi_GraphicsContext::user_to_disp_dy(double y) const
/* PURPOSE:    scales an y distance from user to display coordinates
   RECEIVES:   y - a distance in user coordinates
   RETURNS:    the scaled distance
*/
{  return (int)(y * _disp_ymax / (_user_ymax - _user_ymin));
}

//............................................................................

double Chi_GraphicsContext::disp_to_user_dx(int x) const
/* PURPOSE:    scales an x distance from display to user coordinates
   RETURNS:    the scaled distance
*/
{  return (double)x * (_user_xmax - _user_xmin) / _disp_xmax;
}

//............................................................................

double Chi_GraphicsContext::disp_to_user_dy(int y) const
/* PURPOSE:    scales an y distance from user to display coordinates
   RETURNS:    the scaled distance
*/
{  return (double)y * (_user_ymax - _user_ymin) / _disp_ymax;
}

//............................................................................

void Chi_GraphicsContext::point(double x, double y)
/* PURPOSE:    plot a point
   RECEIVES:   x, y - the point coordinates
   REMARKS:    We plot points as small rectangles because
               dots made with fillellipse don't XOR in BGI
*/
{  const int POINT_RADIUS = 2;
   int disp_x = user_to_disp_x(x);
   int disp_y = user_to_disp_y(y);
   ::rectangle(disp_x - POINT_RADIUS, disp_y - POINT_RADIUS,
      disp_x + POINT_RADIUS, disp_y + POINT_RADIUS);
}

//............................................................................

void Chi_GraphicsContext::line(double xfrom, double yfrom, double xto, double yto)
/* PURPOSE:    plot a line segment
   RECEIVES:   xfrom, yfrom - starting point
               xto, yto - ending point
*/
{  int x1 = user_to_disp_x(xfrom);
   int x2 = user_to_disp_x(xto);
   int y1 = user_to_disp_y(yfrom);
   int y2 = user_to_disp_y(yto);
   ::line(x1, y1, x2, y2);
}

//............................................................................

void Chi_GraphicsContext::piesegment(double x, double y, double ra, double rb,
   double x1, double y1, double x2, double y2)
/* PURPOSE:    Plot a pie segment
   RECEIVES:   x, y - center
               ra, rb - radii in x- and y- direction
   REMARKS:    Since the Borland BGI does not support elliptical segments, we
               must roll our own, using drawpoly and fillpoly.
*/
{  double afrom;
   double ato;

   int disp_x = user_to_disp_x(x);
   int disp_y = user_to_disp_y(y);
   int disp_rx = user_to_disp_dx(ra);
   int disp_ry = user_to_disp_dy(rb);

   int disp_x1 = user_to_disp_x(x1);
   int disp_y1 = user_to_disp_y(y1);
   int disp_x2 = user_to_disp_x(x2);
   int disp_y2 = user_to_disp_y(y2);

   afrom = atan2(disp_y1 - disp_y, disp_x1 - disp_x);
   ato = atan2(disp_y2 - disp_y, disp_x2 - disp_x);

   const int NPOLY = 48;  // this number of points seems to work nicely
   int coords[2 * NPOLY + 6];

   coords[0] = disp_x;
   coords[1] = disp_y;

   double adiff = ato - afrom; // do calcs outside of loops when possible.
   if (adiff < 0) adiff += 2 * M_PI;

   for (int i = 0; i <= NPOLY; i++)
   {  double angle = afrom + adiff * i / NPOLY;
      coords[2 * i + 2] = disp_x + cos(angle) * disp_rx;
      coords[2 * i + 3] = disp_y - sin(angle) * disp_ry;
   }
   coords[NPOLY * 2 + 4] = disp_x;
   coords[NPOLY * 2 + 5] = disp_y;
   if (_brush != EMPTY_BRUSH)
      ::fillpoly(NPOLY + 3, coords);
   ::drawpoly(NPOLY + 3, coords);
}

//............................................................................

void Chi_GraphicsContext::rectangle(double xleft, double ytop, double xright, double ybot)
/* PURPOSE:    Plot a rectangle
   RECEIVES:   xleft, ytop - the top left corner
               xright, ybottom - the bottom right corner
   REMARKS:    Must also call ::bar to fill inside with pattern
               Recode to take any two corners
*/
{  int x1 = user_to_disp_x(xleft);
   int x2 = user_to_disp_x(xright);
   int y1 = user_to_disp_y(ytop);
   int y2 = user_to_disp_y(ybot);
   if (x1 > x2) chi_swap(x1, x2);
   if (y1 > y2) chi_swap(y1, y2);

   if (_brush != EMPTY_BRUSH)
      ::bar(x1, y1, x2, y2);
   ::rectangle(x1, y1, x2, y2);
}

//............................................................................

void Chi_GraphicsContext::set_clip_rectangle(double xleft, double ytop, double xright, double ybot)
/* PURPOSE:    set clipping rectangle
   RECEIVES:   xleft, ytop - the top left corner
               xright, ybottom - the bottom right corner
*/
{  int x1 = user_to_disp_x(xleft) + _disp_xoff;
   int x2 = user_to_disp_x(xright) + _disp_xoff;
   int y1 = user_to_disp_y(ytop) + _disp_yoff;
   int y2 = user_to_disp_y(ybot) + _disp_yoff;
   if (x1 > x2) chi_swap(x1, x2);
   if (y1 > y2) chi_swap(y1, y2);

   ::setviewport(x1, y1, x2, y2, TRUE);
   _disp_xoff = x1;
   _disp_yoff = y1;
}

//............................................................................

void Chi_GraphicsContext::clear()
/* PURPOSE:    Clear the display
*/
{  ::clearviewport();
}

//............................................................................

void Chi_GraphicsContext::ellipse(double x, double y, double ra, double rb)
/* PURPOSE:    Plot an ellipse
   RECEIVES:   x, y - center
               ra, rb - radii in x- and y- direction
   REMARKS:    We are breaking the ellipse into a polygon since
               the XOR rop mode is not "currently" supported for ellipses
*/
{  const int NPOLY = 48;
   int coords[2 * NPOLY + 2];
   for (int i = 0; i <= NPOLY; i++)
   {  double angle = 2 * M_PI * i / NPOLY;
      coords[2 * i] = user_to_disp_x(x + cos(angle) * ra);
      coords[2 * i + 1] = user_to_disp_y(y + sin(angle) * rb);
   }
   if (_brush != EMPTY_BRUSH)
      ::fillpoly(NPOLY + 1, coords);
   ::drawpoly(NPOLY + 1, coords);
}

//............................................................................

void Chi_GraphicsContext::arc(double x, double y, double ra, double rb,
   double x1, double y1, double x2, double y2)
/* PURPOSE:    Plot an arc
   RECEIVES:   x, y - center
               ra, rb - radii in x- and y- direction
*/
{  double afrom;
   double ato;

   int disp_x = user_to_disp_x(x);
   int disp_y = user_to_disp_y(y);
   int disp_rx = user_to_disp_dx(ra);
   int disp_ry = user_to_disp_dy(rb);

   int disp_x1 = user_to_disp_x(x1);
   int disp_y1 = user_to_disp_y(y1);
   int disp_x2 = user_to_disp_x(x2);
   int disp_y2 = user_to_disp_y(y2);

   afrom = atan2(disp_y1 - disp_y, disp_x1 - disp_x);
   ato = atan2(disp_y2 - disp_y, disp_x2 - disp_x);

   const int NPOLY = 48;  // this number of points seems to work nicely
   int coords[2 * NPOLY + 2];

   double adiff = ato - afrom; // do calcs outside of loops when possible.
   if (adiff < 0) adiff += 2 * M_PI;

   for (int i = 0; i <= NPOLY; i++)
   {  double angle = afrom + adiff * i / NPOLY;
      coords[2 * i] = disp_x + cos(angle) * disp_rx;
      coords[2 * i + 1] = disp_y - sin(angle) * disp_ry;
   }
   ::drawpoly(NPOLY + 1, coords);
}

//............................................................................

void Chi_GraphicsContext::set_brush_style(BrushStyle style)
/* PURPOSE:    set the style for the brush to color the insides of
               closed shapes (rectangles, polygons, ellipses)
   RECEIVES:   style - the brush style
*/
{  static int gc2bgi[] =
   {  EMPTY_FILL, SOLID_FILL,
      /* EMPTY_BRUSH, SOLID_BRUSH, */
      LINE_FILL, SLASH_FILL, BKSLASH_FILL, XHATCH_FILL
      /* HORIZ_BRUSH, FDIAG_BRUSH, BDIAG_BRUSH, XDIAG_BRUSH */
   };

   struct fillsettingstype settings;
   getfillsettings(&settings);
   setfillstyle(gc2bgi[style], settings.color);
   _brush = style;
}

//............................................................................

void Chi_GraphicsContext::set_pen_style(PenStyle s)
/* PURPOSE:    set the pen thickness
   RECEIVES:   t - the pen thickness
*/
{  static int gc2bgi[] =
   {   ::SOLID_LINE, ::SOLID_LINE, ::DOTTED_LINE, ::DASHED_LINE
      /* NORMAL_PEN, THICK_PEN, DOTTED_PEN, DASHED_PEN */
   };

   struct linesettingstype settings;
   ::getlinesettings(&settings);
   ::setlinestyle(gc2bgi[s], settings.upattern,
      s == THICK_PEN ? ::THICK_WIDTH : ::NORM_WIDTH);
}

//............................................................................

void Chi_GraphicsContext::set_pen_color(Color col)
/* PURPOSE:   set the color for writing operations
   RECEIVES:  col - the color (BLACK TO WHITE)
*/
{  ::setcolor(col);
}

//............................................................................

void Chi_GraphicsContext::set_brush_color(Color col)
/* PURPOSE:   set the color for fill operations
   RECEIVES:  col - the color (BLACK TO WHITE)
*/
{  struct fillsettingstype settings;
   getfillsettings(&settings);
   setfillstyle(settings.pattern, col);
}

//............................................................................

void Chi_GraphicsContext::set_font_style(FontStyle style)
/* PURPOSE:    set the font style for text
   RECEIVES:   style - the font style
*/
{  if (_font == style) return;

   static int gc2bgi[] =
      { DEFAULT_FONT, SANS_SERIF_FONT, TRIPLEX_FONT };
      /* SYS_FONT, SANS_FONT, ROMAN_FONT */
   textsettingstype settings;
   gettextsettings(&settings);
   settextstyle(gc2bgi[style], settings.direction, style == SYS_FONT ? 1 : 0);
   _font = style;
   set_font_size(_font_size);
}

//............................................................................

void Chi_GraphicsContext::set_font_size(double s)
/* PURPOSE:    set the font size for text output
   RECEIVES:   s - the font size (1.0 = normal)
   REMARKS:    Rescale to take points
*/
{  if (_font_size == s) return;
   if (_font != SYS_FONT)
   {  setusercharsize(1, 1, 1, 1);
      double ysize = disp_to_user_dy(textheight("X"));

      setusercharsize((int)(1000 * s / ysize), 1000, (int)(1000 * s / ysize), 1000);
   }
   _font_size = s;
}

//............................................................................

void Chi_GraphicsContext::text(Chi_String t, double x, double y)
/* PURPOSE:    plot text in the current font
   RECEIVES:   t - the text
               x, y - the top left corner
*/
{  int dx = user_to_disp_x(x);
   int dy = user_to_disp_y(y);
   if (_font != SYS_FONT) dy -= textheight("X") / 4;
   outtextxy(dx, dy, t.c_str());
}

//............................................................................

void Chi_GraphicsContext::text_extent(Chi_String t, double& x, double& y)
/* PURPOSE:    measure text in the current font
   RECEIVES:   t - the text
               x, y (OUT) - the size in x- and y-direction
*/
{  x = disp_to_user_dx(textwidth(t.c_str()));
   y = disp_to_user_dy(textheight(t.c_str()));
}

//............................................................................

void Chi_GraphicsContext::set_rop_mode(RopMode rop)
/* PURPOSE:    set the mode for raster operations
   RECEIVES:   rop - the mode
*/
{  static int gc2bgi[] = { COPY_PUT, XOR_PUT, AND_PUT, NOT_PUT, OR_PUT };
      /* COPY_ROP, XOR_ROP, AND_ROP, NOT_ROP, OR_ROP */
   ::setwritemode(gc2bgi[rop]);
}

//............................................................................

void Chi_GraphicsContext::mouse_events(int (*h)(const Chi_MouseEvent&, void*), void* p)
/* PURPOSE:    keep handling mouse events until callback returns zero
   RECEIVES:   h - function to call on mouse event
               p - pointer to pass to that function
*/
{  for(;;)
   {  Chi_MouseEvent e = _mouse.read(self);
      if (e.type() == Chi_MouseEvent::NOT_INSTALLED) return;
      if ((*h)(e, p) == 0) return;
   }
}

/*-------------------------------------------------------------------------*/

Chi_MouseEvent::Chi_MouseEvent(Type t, double x, double y)
/* RECEIVES:   t - event type
               x, y - mouse coordinates
*/
:  _type(t), _x(x), _y(y)
{}

/*-------------------------------------------------------------------------*/

Chi_Mouse::Chi_Mouse()
{  unsigned char* mouse_vect = (unsigned char*)_dos_getvect(MOUSE_INTR);
   _installed = mouse_vect != 0 && *mouse_vect != RETF;
}

//............................................................................

Bool Chi_Mouse::is_installed() const
/* RETURNS:    TRUE iff a mouse is installed
*/
{  return _installed;
}

//............................................................................

void Chi_Mouse::show_cursor()
/* PURPOSE:    show mouse cursor
*/
{  if (!_installed) return;
   union REGS regs;
   regs.x.ax = SHOWCURSOR;
   int86(MOUSE_INTR, &regs, &regs);
}

//............................................................................

void Chi_Mouse::hide_cursor()
/* PURPOSE:    hide mouse cursor
*/
{  if (!_installed) return;
   union REGS regs;
   regs.x.ax = HIDECURSOR;
   int86(MOUSE_INTR, &regs, &regs);
}

//............................................................................

int Chi_Mouse::get_position(int& x, int& y) const
/* PURPOSE:    get current mouse position
   RECEIVES:   x, y (OUT) - set to mouse position
   RETURNS:    mouse button flags (bit #0 = left button, #1 = middle button,
               #2 = right button)
*/
{  if (!_installed) return 0;
   union REGS regs;
   regs.x.ax = READMOUSE;
   int86(MOUSE_INTR, &regs, &regs);
   x = regs.x.cx;
   y = regs.x.dx;
   return regs.x.bx;
}

//............................................................................

Chi_MouseEvent Chi_Mouse::read(const Chi_GraphicsContext& gc)
/* PURPOSE:    get next mouse event
   RECEIVES:   gc - graphics context (to convert mouse coordinates to graphics)
   RETURNS:    the next mouse event
*/
{  if (!_installed) return Chi_MouseEvent(Chi_MouseEvent::NOT_INSTALLED, 0, 0);
   // get buttons
   // get x, y
   int old_x, old_y, old_b;
   int x, y, b;
   show_cursor();
   old_b = get_position(old_x, old_y) & 1;
   Bool changed = FALSE;
   while (!changed) // buttons and x, y not changed
   {  b = get_position(x, y) & 1;
      if (old_x != x || old_y != y || old_b != b) changed = TRUE;

   }
   Chi_MouseEvent::Type t;
   if (b == old_b) t = Chi_MouseEvent::MOUSEMOVE;
   else if (b) t = Chi_MouseEvent::LBUTTONDOWN;
   else t = Chi_MouseEvent::LBUTTONUP;
   hide_cursor();

   double dx = gc.disp_to_user_dx(x);
   double dy = gc.disp_to_user_dx(y);

   return Chi_MouseEvent(t, dx, dy);
}

/* TEST ********************************************************************/

#ifdef CHIGCBGI_TEST

int chi_main(int, char**)
{  Chi_GraphicsContext gc;
   double tx, ty;
   char hello[] = "Hgllo, World";
   gc.text_extent(hello, tx, ty);
   gc.set_rop_mode(Chi_GraphicsContext::XOR_ROP);
   for (double d = -1; d <= 1; d += 0.1)
   {  gc.rectangle(10 -tx/2*1.15 + d, 8 - ty/2*1.15 + d, 10 + tx/2*1.15 + d, 8 + ty/2*1.15 + d);
      gc.text(hello, 10 -tx/2 + d, 8 - ty/2 + d);

      gc.rectangle(10 - tx/2*1.15 + d, 8 - ty/2*1.15 + d, 10 + tx/2*1.15 + d, 8 + ty/2*1.15 + d);
      gc.text(hello, 10 - tx/2 + d, 8 - ty/2 + d);
   }
   gc.rectangle(10 - tx/2*1.15 + d, 8 - ty/2*1.15 + d, 10 + tx/2*1.15 + d, 8 + ty/2*1.15 + d);
   gc.text(hello, 10 - tx/2 + d, 8 - ty/2 + d);

   gc.rectangle(5, 5, 7, 7);
   gc.set_font_style(Chi_GraphicsContext::ROMAN_FONT);
   gc.set_font_size(2);
   gc.text(hello, 5, 5);

   gc.rectangle(3, 3, 4, 4);
   gc.set_font_style(Chi_GraphicsContext::SYS_FONT);
   gc.set_font_size(2);
   gc.text(hello, 3, 3);

   gc.close();

   return 0;
}

#endif
