/****************************************************************************
** COPYRIGHT (C):    1994 Cay S. Horstmann. All Rights Reserved.
** PROJECT:          ChiLib
** FILE:             chigcwin.cpp
** PURPOSE:          Graphics context class (Windows version)
** VERSION           1.0
** LANGUAGE:         Borland  C++ 3.1/4.0
** TARGET:           IBM PC
** PROGRAMMERS:      Cay Horstmann (CSH)
**                   Charles Miller (CBM)
** START DATE:       4-25-94  (CBM)
** UPDATE HISTORY:
9/27/95 CSH 			changed user_to_disp_dx -> user_to_disp_x in
							rectangle and set_clip_rectangle
****************************************************************************/

/* IMPORT ******************************************************************/

#define CHI_WIN

#ifndef EXPORT
#include "chisetup.h"
#endif

EXPORT
#include <windows.h>
#include <math.h>
#include <assert.h>
#include <iostream.h>
#include <string.h>
#include <dos.h>
#include <ctype.h>
EXPORT
#include "chistrng.h"
#include "chigcwin.h"

EXPORT int chi_main(int, char**);

/* CONSTANTS ***************************************************************/

const int NORM_WIDTH = 1;
const int THICK_WIDTH = 3;

const double DEFAULT_XMIN = 0;
const double DEFAULT_YMIN = 0;
const double DEFAULT_XMAX = 20;
const double DEFAULT_YMAX = 15;

/* TYPES *******************************************************************/

#ifndef EXPORT

class Chi_MouseEvent
{
public:
   enum Type { NOT_INSTALLED, LBUTTONDOWN, MOUSEMOVE, LBUTTONUP };
   Chi_MouseEvent(Type, double, double);

   Type type() const { return _type; }
   double x() const { return _x; }
   double y() const { return _y; }
private:
   Type _type;
   double _x;
   double _y;
};

/*-------------------------------------------------------------------------*/

class Chi_Bitmap
{
public:
   Chi_Bitmap();
   void init(HWND hwnd);
   void close() const;
   void repaint(PAINTSTRUCT&) const;
   void clear_bitmap(COLORREF color) const;
   HDC hdc() const;

private:
   HDC _hdc;                     // a long-lived device context
   HBITMAP _hbm, _hbm_saved;     // handles to the created bitmap objects
   int _disp_xmax;               // width of the bitmap (in bits)
   int _disp_ymax;               // height of the bitmap (in bits)
};

/*-------------------------------------------------------------------------*/

class Chi_GraphicsContext
// PURPOSE: Sends graphics primitives to a display window
{
public:
   enum Color
   {  BLACK, BLUE, GREEN, CYAN,
      RED, MAGENTA, BROWN, LIGHTGRAY,  /* dark colors */
      DARKGRAY, LIGHTBLUE, LIGHTGREEN, LIGHTCYAN,
      LIGHTRED, LIGHTMAGENTA, YELLOW, WHITE  /* light colors */,
      DEFAULT_COLOR = BLACK 
         /* default color for drawing--opposite from background */
   };
   enum PenStyle { NORMAL_PEN, THICK_PEN, DOTTED_PEN, DASHED_PEN };
   enum BrushStyle { EMPTY_BRUSH, SOLID_BRUSH, HORIZ_BRUSH, FDIAG_BRUSH,
                        BDIAG_BRUSH, XDIAG_BRUSH };
   enum FontStyle { SYS_FONT, SANS_FONT, ROMAN_FONT };
   enum RopMode { COPY_ROP, XOR_ROP, OR_ROP, AND_ROP, NOT_ROP };

   Chi_GraphicsContext();
   void close(Bool wait = TRUE);

   void point(double x, double y);
   void line(double xfrom, double yfrom, double xto, double yto);
   void arc(double x, double y, double ra, double rb, double x1, double y1,
      double x2, double y2);
   void rectangle(double xleft, double ytop, double xright, double ybot);
   void ellipse(double x, double y, double ra, double rb);
   void piesegment(double x, double y, double ra, double rb, double x1, double y1,
      double x2, double y2);

   void text(Chi_String t, double x, double y);
   void text_extent(Chi_String t, double& x, double& y);

   void set_brush_style(BrushStyle style);
   void set_brush_color(Color color);
   void set_pen_style(PenStyle style);
   void set_pen_color(Color color);
   void set_font_size(double size);
   void set_font_style(FontStyle style);
   void set_rop_mode(RopMode mode);

   void set_window_coord(double xmin, double ymin, double xmax, double ymax);
   void set_clip_rectangle(double xmin, double ymin, double xmax, double ymax);
   void clear();
   void mouse_events(int (*h)(const Chi_MouseEvent&, void*), void* p);

private:
   COLORREF win_color(Color index) const;

   int user_to_disp_x(double x) const;
   int user_to_disp_y(double y) const;
	int user_to_disp_dx(double x) const;
	int user_to_disp_dy(double y) const;
   double disp_to_user_dx(int x) const;
   double disp_to_user_dy(int y) const;

   double _user_xmin; // the user's window's logical window coords
   double _user_xmax;
   double _user_ymin;
   double _user_ymax;
   int _disp_xmax; // the physical window dimension (in pixels)
   int _disp_ymax;
   Color _brush_color;
   BrushStyle _brush_style;
   Color _pen_color;
   PenStyle _pen_style;
   FontStyle _font_style;
   int _win_rop_mode; // the Windows raster op mode
   int _def_fontheight; // the font height to make 1 unit characters

   HBRUSH _hbrush; // currently selected brush
   HPEN  _hpen;
   LOGFONT _lf; // for maintaining font status
   HFONT _hfont;

   Chi_Bitmap& _bm; // the bitmap to draw on
   HDC _hdc; // device context to the display
};

#endif

/* GLOBALS *****************************************************************/

static Chi_Bitmap mainwin_bm;
static HDC mainwin_hdc;

static int chi_argc;
static char chi_cmdline[256];
static char* chi_argv[128];

/* FUNCTIONS ***************************************************************/

#ifdef CHI_MC21
static long FAR PASCAL chiwin_proc(HWND hwnd, UINT message,
   UINT wParam, LONG lParam)
#elif defined(CHI_WC10)
long FAR PASCAL _export chiwin_proc(HWND hwnd, UINT message,
   UINT wParam, LONG lParam)
#else
static long FAR PASCAL _export chiwin_proc(HWND hwnd, UINT message,
   UINT wParam, LONG lParam)
#endif
/* PURPOSE:    Window procedure for graphics window
*/
{  static BOOL paint_flag = TRUE;   // tells WinProc to call chi_main()
   PAINTSTRUCT ps;            // the display's paint struct

   switch (message)
   {
      case WM_PAINT:
         mainwin_hdc = BeginPaint(hwnd, &ps);
         if (paint_flag)
         {  mainwin_bm.init(hwnd);
            paint_flag = FALSE;
            chi_main(chi_argc, chi_argv);
         }
         else
            mainwin_bm.repaint(ps);
         EndPaint(hwnd, &ps);
         break;
      case WM_DESTROY:
         mainwin_bm.close();
         PostQuitMessage(0);
         return 0;
   }
   return DefWindowProc(hwnd, message, wParam, lParam);
}

/*-------------------------------------------------------------------------*/

int PASCAL WinMain (HANDLE hInstance, HANDLE hPrevInstance,
                    LPSTR lpszCmdParam, int nCmdShow)
{  MSG msg;
   WNDCLASS wndclass;
   char title[sizeof(chi_cmdline)];

   char* p = chi_cmdline;

   chi_argc = 0;
   GetModuleFileName(hInstance, chi_cmdline, sizeof(chi_cmdline));
   strncat(chi_cmdline, " ", sizeof(chi_cmdline) - strlen(chi_cmdline));
   strncat(chi_cmdline, lpszCmdParam, sizeof(chi_cmdline) - strlen(chi_cmdline));
   chi_cmdline[sizeof(chi_cmdline) - 1] = 0;
   strcpy(title, chi_cmdline);

   while (*p != 0 && isspace(*p)) p++; // could be leading spaces

   while (*p != 0)
   {  chi_argv[chi_argc] = p;
      chi_argc++;
      while (*p != 0 && !isspace(*p)) p++;
      if (*p != 0) { *p = 0; p++; }
      while (*p != 0 && isspace(*p)) p++;
   }
   chi_argv[chi_argc] = 0;

   if (!hPrevInstance)
   {  wndclass.style         = CS_HREDRAW | CS_VREDRAW;
      wndclass.lpfnWndProc   = chiwin_proc;
      wndclass.cbClsExtra    = 0;
      wndclass.cbWndExtra    = 0;
      wndclass.hInstance     = hInstance;
      wndclass.hIcon         = LoadIcon(NULL, IDI_APPLICATION);
      wndclass.hCursor       = LoadCursor (NULL, IDC_ARROW);
      wndclass.hbrBackground = GetStockObject (WHITE_BRUSH);
      wndclass.lpszMenuName  = NULL;
      wndclass.lpszClassName = "CHIWIN";

      RegisterClass (&wndclass);
   }

  HWND hwnd = CreateWindow("CHIWIN", title,
      WS_OVERLAPPEDWINDOW, CW_USEDEFAULT, CW_USEDEFAULT,
      CW_USEDEFAULT, CW_USEDEFAULT, NULL, NULL, hInstance,
      NULL);

   ShowWindow(hwnd, SW_SHOWMAXIMIZED);
   UpdateWindow(hwnd);

   while (GetMessage(&msg, NULL, 0, 0))
   {  TranslateMessage(&msg);
      DispatchMessage(&msg);
   }
   return msg.wParam;
}

/*-------------------------------------------------------------------------*/

Chi_MouseEvent::Chi_MouseEvent(Type t, double x, double y)
/* RECEIVES:   t - event type
               x, y - mouse coordinates
*/
:  _type(t), _x(x), _y(y)
{}

/*-------------------------------------------------------------------------*/

Chi_GraphicsContext::Chi_GraphicsContext()
:  _user_xmin(DEFAULT_XMIN),
   _user_ymin(DEFAULT_YMIN),
   _user_xmax(DEFAULT_XMAX),
   _user_ymax(DEFAULT_YMAX),
   _brush_color(DEFAULT_COLOR),
   _pen_color(DEFAULT_COLOR),
   _font_style(SYS_FONT),
   _hfont(0),
   _hpen(0),
   _hbrush(0),
   _hdc(mainwin_hdc),
   _bm(mainwin_bm)
{  _disp_xmax = GetSystemMetrics(SM_CXFULLSCREEN);
   _disp_ymax = GetSystemMetrics(SM_CYFULLSCREEN);


   set_brush_style(EMPTY_BRUSH);
   set_pen_style(NORMAL_PEN);

   // set up the font info...
   _lf.lfHeight = 100;
   _lf.lfWidth = 0;
   _lf.lfWeight = 0;
   _lf.lfEscapement = 0;
   _lf.lfWeight = FW_NORMAL;
   _lf.lfItalic = 0;
   _lf.lfUnderline = 0;
   _lf.lfStrikeOut = 0;
   _lf.lfCharSet = ANSI_CHARSET;
   _lf.lfOutPrecision = OUT_DEFAULT_PRECIS;
   _lf.lfClipPrecision = CLIP_DEFAULT_PRECIS;
   _lf.lfQuality = DEFAULT_QUALITY,
   _lf.lfPitchAndFamily = VARIABLE_PITCH | FF_ROMAN;
   strcpy(_lf.lfFaceName, "Roman");
   HFONT font = CreateFontIndirect(&_lf);
   HDC hdc = GetDC(NULL);
   HFONT oldfont = SelectObject(hdc, font);
#ifdef CHI_WIN32
   SIZE sz;
   GetTextExtentPoint32(hdc, "X", 1, &sz);
   int ty = sz.cy;
#else
   int ty = HIWORD(GetTextExtent(hdc, "X", 1));
#endif
   _def_fontheight = 100 / disp_to_user_dy(ty);
   DeleteObject(SelectObject(hdc, oldfont));
   ReleaseDC(NULL, hdc);

   set_font_style(SYS_FONT);
   set_rop_mode(Chi_GraphicsContext::COPY_ROP);
   clear();
}


//...........................................................................

void Chi_GraphicsContext::close(Bool wait)
/* PURPOSE:    Deallocates various objects when gc is no longer needed
   RECEIVES:   wait - TRUE if we wait for the user to close the application
               FALSE to close the application immediately
*/
{  SelectObject(_hdc, GetStockObject(NULL_BRUSH));
   SelectObject(_bm.hdc(), GetStockObject(NULL_BRUSH));
   DeleteObject(_hbrush);

   SelectObject(_hdc, GetStockObject(BLACK_PEN));
   SelectObject(_bm.hdc(), GetStockObject(BLACK_PEN));
   DeleteObject(_hpen);

   SelectObject(_hdc, GetStockObject(SYSTEM_FONT));
   SelectObject(_bm.hdc(), GetStockObject(SYSTEM_FONT));
   DeleteObject(_hfont);

   if (!wait) PostQuitMessage(0);
}

//...........................................................................

void Chi_GraphicsContext::clear()
/* PURPOSE:    Clears the DC the gc is in, and the bitmap too.
*/
{
   HBRUSH brush1 = CreateSolidBrush(win_color(WHITE));
   HBRUSH brush2 = SelectObject(_hdc, brush1);
   PatBlt(_hdc, 0, 0, _disp_xmax, _disp_ymax, PATCOPY);
   SelectObject(_hdc, brush2);
   DeleteObject(brush1);
   _bm.clear_bitmap(win_color(WHITE));
}

//...........................................................................

void Chi_GraphicsContext::set_window_coord(double xmin, double ymin,
      double xmax, double ymax)
/* PURPOSE:    Set the logical coordinate system
   RECEIVES:   xmin, ymin - coordinates of the top left corner
               xmax, ymax - coordinates of the bottom right corner
*/
{  _user_xmin = xmin;
   _user_xmax = xmax;
   _user_ymin = ymin;
   _user_ymax = ymax;
}

//...........................................................................

int Chi_GraphicsContext::user_to_disp_x(double x) const
/* PURPOSE:    scales the x coordinate from user to display coordinates
   RECEIVES:   x - a coordinate position in user coordinates
   RETURNS:    the scaled coordinate
*/
{  return (int) ((x - _user_xmin) * _disp_xmax / (_user_xmax - _user_xmin));
}

//...........................................................................

int Chi_GraphicsContext::user_to_disp_y(double y) const
/* PURPOSE:    scales the y coordinate from user to display coordinates
   RECEIVES:   y - a coordinate position in user coordinates
   RETURNS:    the scaled coordinate
*/
{  return (int) ((y - _user_ymin) * _disp_ymax / (_user_ymax - _user_ymin));
}

//...........................................................................

int Chi_GraphicsContext::user_to_disp_dx(double x) const
/* PURPOSE:    scales an x distance from user to display coordinates
   RECEIVES:   x - a distance in user coordinates
   RETURNS:    the scaled distance
*/
{  return (int) (x * _disp_xmax / (_user_xmax - _user_xmin));
}

//...........................................................................

int Chi_GraphicsContext::user_to_disp_dy(double y) const
/* PURPOSE:    scales an y distance from user to display coordinates
   RECEIVES:   y - a distance in user coordinates
   RETURNS:    the scaled distance
*/
{  return (int) (y * _disp_ymax / (_user_ymax - _user_ymin));
}

//...........................................................................

double Chi_GraphicsContext::disp_to_user_dx(int x) const
/* PURPOSE:    scales an x distance from display to user coordinates
   RETURNS:    the scaled distance
*/
{  return (double)x * (_user_xmax - _user_xmin) / _disp_xmax;
}

//...........................................................................

double Chi_GraphicsContext::disp_to_user_dy(int y) const
/* PURPOSE:    scales an y distance from user to display coordinates
   RETURNS:    the scaled distance
*/
{  return (double) y * (_user_ymax - _user_ymin) / _disp_ymax;
}

//...........................................................................

void Chi_GraphicsContext::point(double x, double y)
/* PURPOSE:    plot a point
   RECEIVES:   x, y - the point coordinates
*/
{  const int POINT_RADIUS = 2;
   int disp_x = user_to_disp_x(x);
   int disp_y = user_to_disp_y(y);

   SetROP2(_hdc, _win_rop_mode);
   Ellipse(_hdc, disp_x - POINT_RADIUS, disp_y - POINT_RADIUS,
            disp_x + POINT_RADIUS, disp_y + POINT_RADIUS);

   SetROP2(_bm.hdc(), _win_rop_mode);
   Ellipse(_bm.hdc(), disp_x - POINT_RADIUS, disp_y - POINT_RADIUS,
            disp_x + POINT_RADIUS, disp_y + POINT_RADIUS);

}

//...........................................................................

void Chi_GraphicsContext::ellipse(double x, double y, double ra, double rb)
/* PURPOSE:    Plot an ellipse on device context _hdc
   RECEIVES:   x, y - center
               ra, rb - radii in x- and y- direction
*/
{  SetROP2(_hdc, _win_rop_mode);
   Ellipse(_hdc, user_to_disp_x(x - ra), user_to_disp_y(y - rb),
            user_to_disp_x(x + ra),user_to_disp_y(y + rb));

   SetROP2(_bm.hdc(), _win_rop_mode);
   Ellipse(_bm.hdc(), user_to_disp_x(x - ra), user_to_disp_y(y - rb),
            user_to_disp_x(x + ra),user_to_disp_y(y + rb));
}

//...........................................................................

void Chi_GraphicsContext::line(double xfrom, double yfrom, double xto,
                              double yto)
/* PURPOSE:    plot a line segment with current pen style and color
   RECEIVES:   _hdc - the DC on which to draw
               xfrom, yfrom - starting point
               xto, yto - ending point
*/
{
   SetROP2(_hdc, _win_rop_mode);
#ifdef CHI_WIN32
   MoveToEx(_hdc, user_to_disp_x(xfrom), user_to_disp_y(yfrom), 0);
#else
   MoveTo(_hdc, user_to_disp_x(xfrom), user_to_disp_y(yfrom));
#endif
   LineTo(_hdc,user_to_disp_x(xto), user_to_disp_y(yto));

   SetROP2(_bm.hdc(), _win_rop_mode);

#ifdef CHI_WIN32
   MoveToEx(_bm.hdc(), user_to_disp_x(xfrom), user_to_disp_y(yfrom), 0);
#else
   MoveTo(_bm.hdc(), user_to_disp_x(xfrom), user_to_disp_y(yfrom));
#endif

   LineTo(_bm.hdc(),user_to_disp_x(xto), user_to_disp_y(yto));
}

//...........................................................................

void Chi_GraphicsContext::arc(double x, double y, double ra, double rb, double x1, double y1,
      double x2, double y2)
/* PURPOSE:    Plot an arc
   RECEIVES:   x, y - center
               ra, rb - radii in x- and y- direction
*/
{  int disp_x = user_to_disp_x(x);
   int disp_y = user_to_disp_y(y);
	int disp_rx = user_to_disp_dx(ra);
	int disp_ry = user_to_disp_dy(rb);

   int disp_x1 = user_to_disp_x(x1);
   int disp_y1 = user_to_disp_y(y1);
   int disp_x2 = user_to_disp_x(x2);
   int disp_y2 = user_to_disp_y(y2);

   SetROP2(_hdc, _win_rop_mode);
   Arc(_hdc, disp_x - disp_rx, disp_y - disp_ry, disp_x + disp_rx,
         disp_y + disp_ry, disp_x1, disp_y1, disp_x2, disp_y2);

   SetROP2(_bm.hdc(), _win_rop_mode);
   Arc(_bm.hdc(), disp_x - disp_rx, disp_y - disp_ry, disp_x + disp_rx,
         disp_y + disp_ry, disp_x1, disp_y1, disp_x2, disp_y2);
}

//...........................................................................

void Chi_GraphicsContext::piesegment(double x, double y, double ra, double rb, double x1, double y1,
      double x2, double y2)
/* PURPOSE:    Plot a pie segment
   RECEIVES:   x, y - center
               ra, rb - radii in x- and y- direction
*/
{  int disp_x = user_to_disp_x(x);
   int disp_y = user_to_disp_y(y);
	int disp_rx = user_to_disp_dx(ra);
	int disp_ry = user_to_disp_dy(rb);

   int disp_x1 = user_to_disp_x(x1);
   int disp_y1 = user_to_disp_y(y1);
   int disp_x2 = user_to_disp_x(x2);
   int disp_y2 = user_to_disp_y(y2);


   SetROP2(_hdc, _win_rop_mode);
   Pie(_hdc, disp_x - disp_rx, disp_y - disp_ry, disp_x + disp_rx,
         disp_y + disp_ry, disp_x1, disp_y1, disp_x2, disp_y2);

   SetROP2(_bm.hdc(), _win_rop_mode);
   Pie(_bm.hdc(), disp_x - disp_rx, disp_y - disp_ry, disp_x + disp_rx,
         disp_y + disp_ry, disp_x1, disp_y1, disp_x2, disp_y2);
}

//...........................................................................

void Chi_GraphicsContext::set_clip_rectangle(double xleft, double ytop,
                                 double xright, double ybottom)
/* PURPOSE:    set clipping rectangle
   RECEIVES:   xleft, ytop - the top left corner
               xright, ybottom - the bottom right corner
*/
{  HRGN hrgn = CreateRectRgn(user_to_disp_x(xleft), user_to_disp_y(ytop),
		user_to_disp_x(xright), user_to_disp_y(ybottom));
   SelectClipRgn(_hdc, hrgn);
   SelectClipRgn(_bm.hdc(), hrgn);
   DeleteObject(hrgn);
}

//...........................................................................

void Chi_GraphicsContext::rectangle(double xleft, double ytop,
                                 double xright, double ybottom)
/* PURPOSE:    Plot a rectangle
   RECEIVES:   xleft, ytop - the top left corner
               xright, ybottom - the bottom right corner
   REMARKS:    Recode to take two corners (ie, Point p1, p2)
*/
{
   SetROP2(_hdc, _win_rop_mode);
   Rectangle(_hdc, user_to_disp_x(xleft), user_to_disp_y(ytop),
		user_to_disp_x(xright), user_to_disp_y(ybottom));

	SetROP2(_bm.hdc(), _win_rop_mode);
	Rectangle(_bm.hdc(), user_to_disp_x(xleft), user_to_disp_y(ytop),
		user_to_disp_x(xright), user_to_disp_y(ybottom));
}

//...........................................................................

void Chi_GraphicsContext::set_brush_style(BrushStyle style)
/* PURPOSE:    sets the style for the brush to color the insides of
               closed shapes (rectangles, polygons, ellipses)
   RECEIVES:   style - the brush style
*/
{
   static int gc2win[] =
   {  0, 0,
      /* HOLLOW_BRUSH, SOLID_BRUSH (ignored), */
      HS_HORIZONTAL, HS_FDIAGONAL, HS_BDIAGONAL, HS_DIAGCROSS
      /* HORIZ_BRUSH, FDIAG_BRUSH, BDIAG_BRUSH, XDIAG_BRUSH */
   };
   LOGBRUSH lb;
   lb.lbHatch = gc2win[style];
   lb.lbColor = win_color(_brush_color);
   if (style == EMPTY_BRUSH)
      lb.lbStyle = BS_HOLLOW;
   else if (style == SOLID_BRUSH)
      lb.lbStyle = BS_SOLID;
   else
      lb.lbStyle = BS_HATCHED;

   HBRUSH brush = CreateBrushIndirect(&lb);
   SelectObject(_hdc, brush);
   SelectObject(_bm.hdc(), brush);
   DeleteObject(_hbrush);
   _hbrush = brush;

   _brush_style = style;
}


//...........................................................................

void Chi_GraphicsContext::set_brush_color(Color color)
/* PURPOSE:    set the color of the brush that colors the insides of
               closed shapes (rectangles, polygons, ellipses)
   RECEIVES:   color - the brush color
*/
{  static int gc2win[] =
   {  0, 0,
      /* EMPTY_BRUSH, SOLID_BRUSH (ignored), */
      HS_HORIZONTAL, HS_FDIAGONAL, HS_BDIAGONAL, HS_DIAGCROSS
      /* HORIZ_BRUSH, FDIAG_BRUSH, BDIAG_BRUSH, XDIAG_BRUSH */
   };
   LOGBRUSH lb;
   lb.lbHatch = gc2win[_brush_style];
   lb.lbColor = win_color(color);
   if (_brush_style == EMPTY_BRUSH)
      lb.lbStyle = BS_HOLLOW;
   else if (_brush_style == SOLID_BRUSH)
      lb.lbStyle = BS_SOLID;
   else
      lb.lbStyle = BS_HATCHED;

   HBRUSH brush = CreateBrushIndirect(&lb);
   SelectObject(_hdc, brush);
   SelectObject(_bm.hdc(), brush);
   DeleteObject(_hbrush);
   _hbrush = brush;

   _brush_color = color;
}

//...........................................................................

void Chi_GraphicsContext::set_pen_style(PenStyle style)
/* PURPOSE:    set the style (dotted, dashed, solid) of the pen that draws
               shapes (lines, points, polygons)
   RECEIVES:   style - the pen style
*/
{  static int gc2win[]= { PS_SOLID, PS_SOLID, PS_DOT, PS_DASH };
               // NORMAL_PEN, THICK_PEN, DOTTED_PEN, DASHED_PEN

   HPEN pen = CreatePen(gc2win[style], style == THICK_PEN ? THICK_WIDTH : NORM_WIDTH,
      win_color(_pen_color));
   SelectObject(_hdc, pen);
   SelectObject(_bm.hdc(), pen);
   DeleteObject(_hpen);
   _hpen = pen;

   _pen_style = style;
}

//...........................................................................

void Chi_GraphicsContext::set_pen_color(Color color)
/* PURPOSE:    set the color of the pen that draws all shapes
   RECEIVES:   color- the pen color
*/
{  static int gc2win[]= { PS_SOLID, PS_SOLID, PS_DOT, PS_DASH };
               // NORMAL_PEN, THICK_PEN, DOTTED_PEN, DASHED_PEN
   COLORREF wcolor = win_color(color);
   HPEN pen = CreatePen(gc2win[_pen_style], _pen_style == THICK_PEN ? THICK_WIDTH : NORM_WIDTH,
      wcolor);
   SelectObject(_hdc, pen);
   SelectObject(_bm.hdc(), pen);
   DeleteObject(_hpen);
   _hpen = pen;
   SetTextColor(_hdc, wcolor);
   SetTextColor(_bm.hdc(), wcolor);

   _pen_color = color;
}

//...........................................................................

void Chi_GraphicsContext::set_font_size(double s)
/* PURPOSE:    set the font size for text output
   RECEIVES:   s - the font size (1.0 = normal, yheight 1 unit)
*/
{  if (_font_style != SYS_FONT)
   {  _lf.lfHeight = s * _def_fontheight;
      HFONT font = CreateFontIndirect(&_lf);
      SelectObject(_hdc, font);
      SelectObject(_bm.hdc(), font);
      DeleteObject(_hfont);
      _hfont = font;
   }
}

//...........................................................................

void Chi_GraphicsContext::set_font_style(FontStyle style)
/* PURPOSE:    set the font style for text output
   RECEIVES:   style - the style with which to write text.
   REMARKS:    Font size is reset to "normal."
*/
{  HFONT font;
   if (style == SYS_FONT)
      font = GetStockObject(SYSTEM_FONT);
   else
   {  _lf.lfHeight = _def_fontheight;
      switch (style)
      {  case SANS_FONT:
            strcpy(_lf.lfFaceName, "MS Sans Serif");
            break;
         case ROMAN_FONT:
            strcpy(_lf.lfFaceName, "Roman");
            break;
      }
      font = CreateFontIndirect(&_lf);
   }

   SelectObject(_hdc, font);
   SelectObject(_bm.hdc(), font);
   DeleteObject(_hfont);

   if (style == SYS_FONT)
      _hfont = 0;
   else
      _hfont = font;
   _font_style = style;
}

//...........................................................................

void Chi_GraphicsContext::text(Chi_String s, double x, double y)
/* PURPOSE:    plot text in the current font
   RECEIVES:   t - the text
               x, y - top left coordinate
*/
{  const char* t = s.c_str();
   SetROP2(_hdc, _win_rop_mode);
   SetBkMode(_hdc, TRANSPARENT);
   TextOut(_hdc, user_to_disp_x(x), user_to_disp_y(y), t, lstrlen(t));
   SetROP2(_bm.hdc(), _win_rop_mode);
   SetBkMode(_bm.hdc(), TRANSPARENT);
   TextOut(_bm.hdc(), user_to_disp_x(x), user_to_disp_y(y), t, lstrlen(t));
}

//...........................................................................

void Chi_GraphicsContext::text_extent(Chi_String s, double& x,
                                    double& y)
/* PURPOSE:    measure text in the current font
   RECEIVES:   t - the text
               x, y (OUT) - the size in x- and y-direction
*/
{  const char* t = s.c_str();

#ifdef CHI_WIN32
   SIZE sz;
   GetTextExtentPoint32(_hdc, t, s.length(), &sz);
   x = disp_to_user_dx(sz.cx);
   y = disp_to_user_dy(sz.cy);
#else
   DWORD d = GetTextExtent(_hdc, t, s.length());
   x = disp_to_user_dx(LOWORD(d));
   y = disp_to_user_dy(HIWORD(d));
#endif
}


//...........................................................................

void Chi_GraphicsContext::set_rop_mode(RopMode rop)
/* PURPOSE:    set the mode for raster operations
   RECEIVES:   rop - the new rop mode
*/
{  static int gc2win[] =
      {  R2_COPYPEN,  R2_NOTXORPEN, R2_MERGEPEN, R2_MASKPEN, R2_NOTCOPYPEN };
      // COPY_ROP,   XOR_ROP,    OR_ROP,     AND_ROP,    NOT_ROP
   _win_rop_mode = gc2win[rop];
}

//...........................................................................

void Chi_GraphicsContext::mouse_events(int (*h)(const Chi_MouseEvent&, void*), void* p)
{  MSG msg;

   while (GetMessage(&msg, (HWND) NULL, 0, 0))
   {  TranslateMessage(&msg);
      DispatchMessage(&msg);
      switch(msg.message)
      {  case WM_LBUTTONDOWN:
            if ((*h)(Chi_MouseEvent(Chi_MouseEvent::LBUTTONDOWN,
               disp_to_user_dx(LOWORD(msg.lParam)),
               disp_to_user_dy(HIWORD(msg.lParam))), p) == 0) return;
            break;

         case WM_LBUTTONUP:
            if ((*h)(Chi_MouseEvent(Chi_MouseEvent::LBUTTONUP,
               disp_to_user_dx(LOWORD(msg.lParam)), 
               disp_to_user_dy(HIWORD(msg.lParam))), p) == 0) return;
            break;

         case WM_MOUSEMOVE:
            if ((*h)(Chi_MouseEvent(Chi_MouseEvent::MOUSEMOVE,
               disp_to_user_dx(LOWORD(msg.lParam)),
               disp_to_user_dy(HIWORD(msg.lParam))), p) == 0) return;
            break;
      }
   }
}

//...........................................................................

COLORREF Chi_GraphicsContext::win_color(Color col) const
/* PURPOSE:    Returns the Windows equivalent color values of Color values
   RECEIVES:   col - A graphics context color to translate
   RETURNS:    the Windows equivalent color
*/
{  static COLORREF gc2win[] = {
      RGB(0, 0, 0), RGB(0, 0, 128), RGB(0, 128, 0), RGB(0, 128, 128),
         // black, blue, green, cyan
      RGB(128, 0, 0), RGB(128, 0, 128), RGB(128, 64, 0), RGB(192, 192, 192),
         // red, magenta, brown, lt gray
      RGB(128, 128, 128), RGB(0, 0, 255), RGB(0, 255, 0), RGB(0, 255, 255),
         // dk gray, lt blue, lt green, lt cyan
      RGB(255, 0, 0), RGB(255, 0, 255), RGB(255, 255, 0), RGB(255, 255, 255)
         // lt red, lt magenta, yellow, white
   };
   return gc2win[col];
}

/*-------------------------------------------------------------------------*/

Chi_Bitmap::Chi_Bitmap()
{  _disp_xmax = GetSystemMetrics(SM_CXFULLSCREEN);
   _disp_ymax = GetSystemMetrics(SM_CYFULLSCREEN);
}

//............................................................................
void Chi_Bitmap::init(HWND hwnd)
/* PURPOSE:    Initializes the bitmap window structures.
   RECEIVES:   Handle to the window in which it is created.
   RETURNS:    A device context specific to this bitmap.
   REMARKS:    We need to know a HDC of the main window in order to assure
                  the bitmap has the proper number of color planes, bits
                  per pixel, etc.
*/
{  HDC hdc = GetDC(hwnd);
   _hdc = CreateCompatibleDC(hdc);
   _hbm = CreateCompatibleBitmap(hdc, _disp_xmax, _disp_ymax);
   ReleaseDC(hwnd, hdc);
   if (!_hbm)
   {  MessageBox(hwnd, (LPSTR)"Memory problems in Bitmap Creation",
         (LPSTR)"ChiLib Error", MB_ICONHAND);
      SendMessage(hwnd, WM_DESTROY, 0, 0L);
      return;
   }
   _hbm_saved = SelectObject(_hdc, _hbm);
   clear_bitmap(Chi_GraphicsContext::WHITE); // clear to white
}


//............................................................................

void Chi_Bitmap::close() const
/* PURPOSE:    Deallocates resources for this bitmap
*/
{  SelectObject(_hdc, _hbm_saved);
   DeleteDC(_hdc);
   DeleteObject(_hbm);
}


//............................................................................

void Chi_Bitmap::repaint(PAINTSTRUCT& ps) const
/* PURPOSE:    Repaints with the saved bitmap--use when resizing.
   RECEIVES:   A pointer to the PAINTSTRUCT in WinProc
*/
{
   BitBlt(ps.hdc, ps.rcPaint.left, ps.rcPaint.top,
      ps.rcPaint.right, ps.rcPaint.bottom, _hdc,
      ps.rcPaint.left, ps.rcPaint.top, SRCCOPY);
}


//............................................................................

void Chi_Bitmap::clear_bitmap(COLORREF color) const
/* PURPOSE:    Clears the bitmap
   RECEIVES:   color - the color to fill the bitmap
*/
{
   HBRUSH brush = CreateSolidBrush(color);
   HBRUSH saved_brush = SelectObject(_hdc, brush);
   PatBlt(_hdc, 0, 0, _disp_xmax, _disp_ymax, PATCOPY);
   SelectObject(_hdc, saved_brush);
   DeleteObject(brush);
   // note our greed in not releasing a DC
}

//...........................................................................

HDC Chi_Bitmap::hdc() const
/* RETURNS:    the device context handle associated with this bitmap
*/
{  return _hdc;
}


/* TEST ********************************************************************/

#ifdef CHIGCWIN_TEST

int chi_main(int, char**)
{  Chi_GraphicsContext gc;

   gc.point(3, 3);
   gc.ellipse(12,5,1,4);

   gc.set_brush_style(Chi_GraphicsContext::SOLID_BRUSH);
   gc.set_brush_color(Chi_GraphicsContext::LIGHTRED);
   gc.rectangle(8, 9, 10, 11);
   gc.set_brush_color(Chi_GraphicsContext::GREEN);
   gc.ellipse(4, 8, 5, 1);
   gc.set_brush_color(Chi_GraphicsContext::MAGENTA);
   gc.set_brush_style(Chi_GraphicsContext::XDIAG_BRUSH);
   gc.piesegment(5, 5, 1, 1, 6, 4, 6, 6);
   gc.text("Hello, Windows GraphicsContext!", 3, 6);
   gc.set_font_style(Chi_GraphicsContext::ROMAN_FONT);
   gc.set_font_size(1.5);
   gc.text("WINDOWS GRAPHICS TEST", 2, 2);
   gc.close(); // always close before returning
   return 0;
}

#endif


