/****************************************************************************
** COPYRIGHT (C):    1993 Cay S. Horstmann. All Rights Reserved.
** PROJECT:          ChiLib
** FILE:             chilist.cpp
** PURPOSE:          Linked list template
** VERSION           1.0
** LANGUAGE:         Borland C++ 3.1
** TARGET:           IBM PC
** PROGRAMMER:       Cay Horstmann (CSH)
** START DATE:       8-16-93 (CSH)
** UPDATE HISTORY:
**
****************************************************************************/

/* IMPORT ******************************************************************/

#ifndef EXPORT
#include "chisetup.h"
#endif

EXPORT
#include "chierror.h"

#include "chilist.h"


/* TYPES *******************************************************************/

#ifndef EXPORT

template<class T> class Chi_Link
/* PURPOSE: A link in a linked list or queue
   REMARKS: The data fields are public, hence users of this class
            should not propagate Chi_Link pointers
*/
{
public:
   Chi_Link(const T&, Chi_Link<T>*);

   T _info;
   Chi_Link<T>* _next;

   static const T& defval();
};

/*-------------------------------------------------------------------------*/



#ifdef CHI_BC50
template<class T> class Chi_ListIterator;
#endif

template<class T> class Chi_List
/* PURPOSE:    A linked list
*/
{
public:
   Chi_List();
   void insert_head(T n);
   void insert_tail(T n);
   void insert(T n);
      // insert before current position
   T remove(); // remove current position
   T& current(); // info of current element
   void next(); // advance current position
   void reset();
   void empty();

   const T& current() const; // info of current element
   Bool at_end() const;
   const T& head() const;
   const T& tail() const;
   int length() const;

   Chi_List(const Chi_List<T>&);
   const Chi_List<T>& operator=(const Chi_List<T>&);
   ~Chi_List();

private:
   void free();
   void copy(const Chi_List<T>&);

   Chi_Link<T>* cursor() const { return _pre ? _pre->_next : _head; }

   Chi_Link<T>* _head;
   Chi_Link<T>* _tail;
   Chi_Link<T>* _pre; // predecessor of cursor
   int _len;

   friend class Chi_ListIterator<T>;
};

/*-------------------------------------------------------------------------*/

template<class T> class Chi_ListIterator
/* REMARKS:    1. This iterator is for inspection of a list only. You can
               only modify a list through its own cursor
               2. Modification of a list invalidates all iterators attached
               to it.
*/
{
public:
   Chi_ListIterator(const Chi_List<T>&);

   void next(); // advance current position
   void reset();

   const T& current() const; // info of current element
   Bool at_end() const;

private:
   Chi_Link<T>* cursor() const { return _pre ? _pre->_next : _list._head; }

   const Chi_List<T>& _list;
   Chi_Link<T>* _pre;
};

//***************************************************************************

template<class T> const T& Chi_Link<T>::defval()
/* PURPOSE:    make a default object of type T
   REMARKS:    We use a function because Borland 3.1 doesn't handle static
               data members in templates correctly
*/
{  static T t;
   return t;
}

//............................................................................

template<class T> Chi_Link<T>::Chi_Link(const T& i, Chi_Link<T>* n) :
/* RECEIVES:   i - the information to store in the link
               n - pointer to next link
*/
   _info(i),
   _next(n)
{}

/*-------------------------------------------------------------------------*/

template<class T> Chi_List<T>::Chi_List() :
   _head(0),
   _tail(0),
   _pre(0),
   _len(0)
{}

//............................................................................

template<class T> Chi_List<T>::~Chi_List()
{  free();
}

//............................................................................

template<class T> Chi_List<T>::Chi_List(const Chi_List<T>& b)
{  copy(b);
}

//............................................................................

template<class T> const Chi_List<T>& Chi_List<T>::operator=(const Chi_List<T>& b)
{  if (this != &b)
   {  free();
      copy(b);
   }
   return *this;
}

//............................................................................

template<class T> void Chi_List<T>::free()
{  Chi_Link<T>* p = _head;
   while(p != 0)
   {  Chi_Link<T>* q = p;
      p = p->_next;
      delete q;
   }
}

//............................................................................

template<class T> void Chi_List<T>::copy(const Chi_List<T>& b)
{  _len = b._len;
   _pre = _head = 0;

   Chi_Link<T>* pa = 0;
   for (Chi_Link<T>* pb = b._head; pb != 0; pb = pb->_next)
   {  Chi_Link<T>* pnew = new Chi_Link<T>(pb->_info, 0);
      CHI_ASSERT_MEM(pnew);
      if (pa)
         pa = pa->_next = pnew;
      else
         pa = _head = pnew;
      if (b._pre == pb) _pre = pa;
   }
   _tail = pa;
}

//............................................................................

template<class T> void Chi_List<T>::empty()
/* PURPOSE:    empty the list
*/
{  free();
   _head = 0;
   _tail = 0;
   _pre = 0;
   _len = 0;
}

//............................................................................

template<class T> void Chi_List<T>::reset()
/* PURPOSE:    reset the iterator
*/
{  _pre = 0;
}

//............................................................................

template<class T> Bool Chi_List<T>::at_end() const
/* RETURNS:    TRUE iff the iterator is at the end of the list
*/
{  return cursor() == 0;
}

//............................................................................

template<class T> void Chi_List<T>::next()
/* PURPOSE:    move the iterator to the next position
*/
{  if (!_pre) _pre = _head;
   else if (_pre->_next) _pre = _pre->_next;
   // do not advance past the tail of the list
}

//............................................................................

template<class T> T& Chi_List<T>::current()
/* RETURNS:    the value stored at the iterator position
*/
{  Chi_Link<T>* cur = cursor();
   CHI_ASSERT_PRECOND(cur);
   return cur->_info;
}

//............................................................................

template<class T> const T& Chi_List<T>::current() const
/* RETURNS:    the value stored at the iterator position
*/
{  Chi_Link<T>* cur = cursor();
   return cur ? cur->_info : Chi_Link<T>::defval();
}

//............................................................................

template<class T> const T& Chi_List<T>::head() const
/* RETURNS:    the value stored at the head of the list
*/
{  return _head ? _head->_info : Chi_Link<T>::defval();
}

//............................................................................

template<class T> const T& Chi_List<T>::tail() const
/* RETURNS:    the value stored at the tail of the list
*/
{  return _tail ? _tail->_info : Chi_Link<T>::defval();
}

//............................................................................

template<class T> void Chi_List<T>::insert(T n)
/* PURPOSE:    insert before the iterator position
   RECEIVES:   n - the value to insert
*/
{  Chi_Link<T>* p = new Chi_Link<T>(n, cursor());
   CHI_ASSERT_MEM(p);

   if (_pre)
   {  _pre->_next = p;
      if (_pre == _tail) _tail = p;
   }
   else
   {  if (_head == 0) _tail = p;
      _head = p;
   }

   _pre = p;
   _len++;
};

//............................................................................

template<class T> void Chi_List<T>::insert_head(T n)
/* PURPOSE:    insert before the head of the list
   RECEIVES:   n - the value to insert
*/
{  Chi_Link<T>* p = new Chi_Link<T>(n, _head);
   CHI_ASSERT_MEM(p);
   _head = p;
   _len++;
}

//............................................................................

template<class T> void Chi_List<T>::insert_tail(T n)
/* PURPOSE:    insert after the tail of the list
   RECEIVES:   n - the value to insert
*/
{  Chi_Link<T>* p = new Chi_Link<T>(n, 0);
   CHI_ASSERT_MEM(p);
   if (_head == 0) _head = _tail = p;
   else
   {  _tail->_next = p;
      _tail = p;
   }
   _len++;
}

//............................................................................

template<class T> T Chi_List<T>::remove()
/* PURPOSE:    remove the element under the cursor
   RETURNS:    the removed element
*/
{  Chi_Link<T>* cur = cursor();
   if (!cur) return Chi_Link<T>::defval();
   if (_tail == cur) _tail = _pre;
   if (_pre)
      _pre->_next = cur->_next;
   else
      _head = cur->_next;
   T r = cur->_info;
   delete cur;
   _len--;
   return r;
}

//............................................................................

template<class T> int Chi_List<T>::length() const
/* RETURNS:    the number of elements in the list
*/
{  return _len;
}

/*-------------------------------------------------------------------------*/

template<class T> Chi_ListIterator<T>::Chi_ListIterator(const Chi_List<T>& l)
/* RECEIVES:   l - the list to iterate over
*/
:  _list(l),
   _pre(0)
{}

//............................................................................

template<class T> void Chi_ListIterator<T>::reset()
/* PURPOSE:    reset the iterator
*/
{  _pre = 0;
}

//............................................................................

template<class T> Bool Chi_ListIterator<T>::at_end() const
/* RETURNS:    TRUE iff the iterator is at the end of the list
*/
{  return cursor() == 0;
}

//............................................................................

template<class T> void Chi_ListIterator<T>::next()
/* PURPOSE:    move the iterator to the next position
*/
{  if (!_pre) _pre = _list._head;
   else if (_pre->_next) _pre = _pre->_next;
   // do not advance past the tail of the list
}

//............................................................................

template<class T> const T& Chi_ListIterator<T>::current() const
/* RETURNS:    the value stored at the iterator position
*/
{  Chi_Link<T>* cur = cursor();
   return cur ? cur->_info : Chi_Link<T>::defval();
}

#endif

/* TEST ********************************************************************/

#ifdef CHILIST_TEST

#include <iostream.h>

void pr(const Chi_List<int>& list)
{  Chi_ListIterator<int> it(list);
   while (!it.at_end())
   {  cout << it.current() << " ";
      it.next();
   }
   cout << endl;
}

void pr2(Chi_List<int> list)
// to test copy constructor, destructor
{  Chi_ListIterator<int> it(list);
   while (!it.at_end())
   {  cout << it.current() << " ";
      it.next();
   }
   cout << endl;
}

int main()
{  Chi_List<int> a;
   for (int i = 1; i <= 10; i++) a.insert(i);
   pr2(a);
   a.reset();
   while (!a.at_end())
   {  a.remove();
      a.next();
   }
   pr(a);
   return 0;
}

#endif

/***************************************************************************/