/****************************************************************************
** COPYRIGHT (C):    1993 Cay S. Horstmann. All Rights Reserved.
** PROJECT:          ChiLib
** FILE:             chimap.cpp
** PURPOSE:          Map template
** VERSION           1.0
** LANGUAGE:         Borland C++ 3.1
** TARGET:           IBM PC
** PROGRAMMER:       Cay Horstmann (CSH)
** START DATE:       7-29-94 (CSH)
** UPDATE HISTORY:
11/22/98 CSH
Made copy constructor and operator= inline for g++
**
****************************************************************************/

/* IMPORT ******************************************************************/

#ifndef EXPORT
#include "chisetup.h"
#endif

EXPORT
#include "chiarray.h"

#include "chimap.h"

/* TYPES *******************************************************************/

#ifndef EXPORT

template<class K, class T> class Chi_Map;

template<class K, class T> class Chi_MapLink
/* PURPOSE: A link in a map implemented as a hash table
*/
{
private:
   Chi_MapLink(const K&, const T&, Chi_MapLink<K, T>*);

   K _key;
   T _value;
   Chi_MapLink<K, T>* _next;

friend class Chi_Map<K, T>;
};

/*-------------------------------------------------------------------------*/

template<class K, class T>
class Chi_Map
/* PURPOSE:    A map (or dictionary) mapping objects of the key type K to
               the type T
*/
{
public:

   Chi_Map(int (*hash)(const K&), int (*comp)(const K&, const K&), int size = 0)
   /* RECEIVES:   hash - the hash function for keys
                  comp - the comparison function to match keys
                  size - the desired size of the underlying hash table
   */
   :  _hash(hash),
      _comp(comp),
      _size(size == 0 ? DEFAULT_SIZE : size),
      _buckets(0, size == 0 ? DEFAULT_SIZE : size),
      _length(0)
   {}

//............................................................................



   void set(const K& key, const T& t);
   Bool remove(const K& key);
   T& operator[](const K& key);

   T get(const K& key) const;
   Bool find(const K& key) const;
   Bool find(const K& key, T& t) const;
   const T& operator[](const K& key) const;

   void empty() { free(); }
   int length() const { return _length; }

   Chi_Map(const Chi_Map<K, T>& b) { copy(b); }
   const Chi_Map<K, T>& operator=(const Chi_Map<K, T>& b)
   {  if (this != &b)
      {  free();
         copy(b);
      }
      return *this;
   }
   
   ~Chi_Map() { free(); }

private:
   void free();
   void copy(const Chi_Map<K, T>&);

   T* find_link(const K& k) const;

   enum { DEFAULT_SIZE = 103 };

   Chi_PtrArray< Chi_MapLink<K, T> > _buckets;

   int (*_hash)(const K&);
   int (*_comp)(const K&, const K&);
   int _size;
   int _length;
};

//***************************************************************************

template<class K, class T>
Chi_MapLink<K, T>::Chi_MapLink(const K& k, const T& i, Chi_MapLink<K, T>* n)
/* RECEIVES:   k - the key
               v - the associated value
               n - the next link
*/
:  _key(k),
   _value(i),
   _next(n)
{}

/*-------------------------------------------------------------------------*/

template<class K, class T> void Chi_Map<K, T>::free()
{  for (int i = _buckets.low(); i <= _buckets.high(); i++)
   {  Chi_MapLink<K, T>* p = _buckets[i];
      while(p != 0)
      {  Chi_MapLink<K, T>* q = p;
         p = p->_next;
         delete q;
      }
      _buckets[i] = 0;
   }
   _length = 0;
}

//............................................................................

template<class K, class T> void Chi_Map<K, T>::copy(const Chi_Map<K, T>& b)
{  _comp = b._comp;
   _hash = b._hash;
   _size = b._size;
   _length = b._length;
   int low = b._buckets.low();
   int high = b._buckets.high();
   _buckets.grow(low, high);
   _buckets.shrink(low, high);

   for (int i = low; i <= high; i++)
   {  Chi_MapLink<K, T>* pa = 0;
      _buckets[i] = 0;
      for (Chi_MapLink<K, T>* pb = b._buckets[i]; pb != 0; pb = pb->_next)
      {  Chi_MapLink<K, T>* pnew = new Chi_MapLink<K, T>(pb->_key, pb->_value, 0);
         CHI_ASSERT_MEM(pnew);
         if (pa)
            pa = pa->_next = pnew;
         else
            pa = _buckets[i] = pnew;
      }
   }
}

//............................................................................

template <class K, class T>
void Chi_Map<K, T>::set(const K& k, const T& t)
/* PURPOSE:    set the value corresponding to a key
   RECEIVES:   k - the key
               v - the corresponding value
*/
{  T* n = find_link(k);
   if (n) *n = t;
   else
   {  int h = (*_hash)(k) % _size;
      if (h < 0) h += _size;
      Chi_MapLink<K, T> *n = new Chi_MapLink<K, T>(k, t, _buckets[h]);
      CHI_ASSERT_MEM(n);
      _buckets[h] = n;
      _length++;
   }
}

//............................................................................

template <class K, class T>
Bool Chi_Map<K, T>::remove(const K& k)
/* PURPOSE:    remove a key and its corresponding value
   RETURNS:    TRUE if the key existed in the map
*/
{  int h = (*_hash)(k) % _size;
   if (h < 0) h += _size;
   Chi_MapLink<K, T>* l = _buckets[h];
   Chi_MapLink<K, T>* pre = 0;
   while (l != 0)
   {  if ((*_comp)(l->_key, k) == 0)
      {  if (pre) pre->_next = l->_next;
         else _buckets[h]->_next = l->_next;
         delete l;
         return TRUE;
      }
      else
      {  pre = l;
         l = l->_next;
      }
   }
   return FALSE;
}

//............................................................................

template <class K, class T>
Bool Chi_Map<K, T>::find(const K& k) const
/* PURPOSE:    find a key
   RETURNS:    TRUE if the key exists in the map
*/
{  T* n = find_link(k);
   return n != 0;
}

//............................................................................

template <class K, class T>
T Chi_Map<K, T>::get(const K& k) const
/* PURPOSE:    find the value associated with a key
   RECEIVES:   k - the key
   RETURNS:    the associated value
*/
{  T* n = find_link(k);
   if (n != 0) return *n;
   else
   {  static T t;
      return t;
   }
}

//............................................................................

template <class K, class T>
const T& Chi_Map<K, T>::operator[](const K& k) const
/* PURPOSE:    find the value associated with a key
   RECEIVES:   k - the key
   RETURNS:    the associated value
   REMARKS:    only use this with a key that is known to exist
*/
{  T* n = find_link(k);
   CHI_ASSERT_PRECOND(n != 0);
   return *n;
}

//............................................................................

template <class K, class T>
T& Chi_Map<K, T>::operator[](const K& k)
/* PURPOSE:    find the value associated with a key
   RECEIVES:   k - the key
   RETURNS:    the associated value
   REMARKS:    only use this with a key that is known to exist
*/
{  T* n = find_link(k);
   CHI_ASSERT_PRECOND(n != 0);
   return *n;
}

//............................................................................

template <class K, class T>
Bool Chi_Map<K, T>::find(const K& k, T& t) const
/* PURPOSE:    find the value associated with a key
   RECEIVES:   k - the key
               t (OUT) - the associated value
   RETURNS:    TRUE if the key exists in the map
*/
{  T* n = find_link(k);
   if (n != 0) t = *n;
   return n != 0;
}

//............................................................................

template <class K, class T>
T* Chi_Map<K, T>::find_link(const K& k) const
/* PURPOSE:    find the value associated with a key
   RECEIVES:   k - the key
   RETURNS:    a pointer to the associated value, or 0 if the key doesn't
               exist in the map
*/
{  int h = (*_hash)(k) % _size;
   if (h < 0) h += _size;
   Chi_MapLink<K, T>* l = _buckets[h];
   while (l != 0)
   {  if ((*_comp)(l->_key, k) == 0)
         return &l->_value;
      else
         l = l->_next;
   }
   return 0;
}

#endif

/* TEST ********************************************************************/

#ifdef CHIMAP_TEST
#include "chistrng.h"

char* names[] = { "Tom", "Dick", "Harry" };

int main()
{  Chi_Map<Chi_String, int> map(Chi_String::hash, Chi_String::compare);
   int i;
   for (i = 0; i < sizeof(names)/sizeof(names[0]); i++)
      map.set(names[i], i);
   for (i = 0; i < sizeof(names)/sizeof(names[0]); i++)
      cout << names[i] << " " << map.get(names[i]) << endl;
   Chi_Map<Chi_String, int> map2(Chi_String::hash, Chi_String::compare);
   for (i = 0; i < sizeof(names)/sizeof(names[0]); i++)
      map2.set(names[i], -i);
   map2 = map;
   for (i = 0; i < sizeof(names)/sizeof(names[0]); i++)
      map2[names[i]]++;
   for (i = 0; i < sizeof(names)/sizeof(names[0]); i++)
      cout << names[i] << " " << map2[names[i]] << endl;
   return 0;
}

#endif

/***************************************************************************/
