/****************************************************************************
** COPYRIGHT (C):    1994 Cay S. Horstmann. All Rights Reserved.
** PROJECT:          ChiLib
** FILE:             chishape.cpp
** PURPOSE:          Shapes for graphic display
** VERSION           1.0
** LANGUAGE:         Borland  C++ 3.1/4.0
** TARGET:           IBM PC
** PROGRAMMER:       Cay Horstmann (CSH)
** START DATE:       2-19-94  (CSH)
** UPDATE HISTORY:
**
****************************************************************************/

/* IMPORT ******************************************************************/

#ifndef EXPORT
#include "chisetup.h"
#endif

EXPORT
#include <iostream.h>
#include <math.h>

EXPORT
#include "chiarray.h"
EXPORT
#include "chigc.h"
EXPORT
#include "chistrng.h"
#include "chierror.h"

#include "chishape.h"

/* TYPES *******************************************************************/

#ifndef EXPORT

EXPORT class Chi_Point;
EXPORT class Chi_Segment;

EXPORT class Chi_Shape
/* PURPOSE:    A shape that can be plotted and moved around
*/
{
public:
   virtual void scale(Chi_Point center, double s) = 0;
   virtual void move(double x, double y) = 0;
   virtual void read(istream& is) = 0;

   virtual void plot(Chi_GraphicsContext& gc) const = 0;
   virtual void print(ostream& os) const = 0;
   virtual Chi_Shape* clone() const;

   virtual ~Chi_Shape() {}
};

/*-------------------------------------------------------------------------*/

EXPORT class Chi_Point : public Chi_Shape
/* PURPOSE:    A single point
*/
{
public:
   Chi_Point(double xx =0, double yy =0);

   virtual void scale(Chi_Point center, double s);
   virtual void move(double x, double y);
   virtual void read(istream& is);

   virtual void plot(Chi_GraphicsContext& gc) const;
   virtual void print(ostream& os) const;
   virtual Chi_Shape* clone() const;

   double x() const;
   double y() const;

   double distance(Chi_Point b) const;
   double angle(Chi_Point b) const;

   ~Chi_Point();
private:
   double _x;
   double _y;
};

/*-------------------------------------------------------------------------*/

EXPORT class Chi_Rectangle : public Chi_Shape
/* PURPOSE:    A rectangle (with sides parallel to the coordinate axes)
*/
{
public:
   Chi_Rectangle();
   Chi_Rectangle(Chi_Point, Chi_Point);

   virtual void scale(Chi_Point center, double s);
   virtual void move(double x, double y);
   virtual void read(istream& is);

   virtual void plot(Chi_GraphicsContext& gc) const;
   virtual void print(ostream& os) const;
   virtual Chi_Shape* clone() const;

   Chi_Point left_top() const;
   Chi_Point left_bottom() const;
   Chi_Point right_top() const;
   Chi_Point right_bottom() const;
   Chi_Point center() const;

   double ytop() const;
   double ybottom() const;
   double xleft() const;
   double xright() const;
   double xsize() const;
   double ysize() const;

   Bool is_inside(Chi_Point p) const;
   Chi_Point boundary_point(Chi_Point p) const;

private:
   Chi_Point _corner1;
   Chi_Point _corner2;
};

/*-------------------------------------------------------------------------*/

class Chi_FilledRect : public Chi_Rectangle
/* PURPOSE:    A rectangle that is filled with a pattern
*/
{
public:
   Chi_FilledRect();
   Chi_FilledRect(Chi_Point, Chi_Point, Chi_GraphicsContext::BrushStyle);
   virtual void plot(Chi_GraphicsContext& gc) const;
private:
   Chi_GraphicsContext::BrushStyle _pattern;
};

/*-------------------------------------------------------------------------*/

EXPORT class Chi_Segment : public Chi_Shape
/* PURPOSE:    A line segment
*/
{
public:
   Chi_Segment();
   Chi_Segment(Chi_Point f, Chi_Point t);

   virtual void scale(Chi_Point center, double s);
   virtual void move(double x, double y);
   virtual void read(istream& is);

   virtual void plot(Chi_GraphicsContext& gc) const;
   virtual void print(ostream& os) const;
   virtual Chi_Shape* clone() const;

   Chi_Point center() const;
   Bool intersect(Chi_Segment s, Chi_Point& p) const;
   Chi_Point closest(Chi_Point p) const;
   double distance(Chi_Point p) const;
   void reflect(double& x, double& y) const;

   Chi_Point from() const;
   Chi_Point to() const;

private:
   Chi_Point _from;
   Chi_Point _to;
};

/*-------------------------------------------------------------------------*/

EXPORT class Chi_Ellipse : public Chi_Shape
/* PURPOSE:    An ellipse (with axes parallel to the coordinate axes)
*/
{
public:
   Chi_Ellipse();
   Chi_Ellipse(Chi_Point, double, double);

   virtual void scale(Chi_Point center, double s);
   virtual void move(double x, double y);
   virtual void read(istream& is);

   virtual void plot(Chi_GraphicsContext& gc) const;
   virtual void print(ostream& os) const;
   virtual Chi_Shape* clone() const;

   Chi_Point center() const;
   double xradius() const;
   double yradius() const;
   Bool is_inside(Chi_Point p) const;
   Chi_Point boundary_point(Chi_Point p) const;

private:
   Chi_Point _center;
   double _xradius;
   double _yradius;
};

/*-------------------------------------------------------------------------*/

EXPORT class Chi_Polygon : public Chi_Shape
/* PURPOSE:    A closed (not necessarily convex) polygon
*/
{
public:
   Chi_Polygon();
   Chi_Polygon(int);

   void set_vertex(int, Chi_Point);

   virtual void scale(Chi_Point center, double s);
   virtual void move(double x, double y);
   virtual void read(istream& is);

   virtual void plot(Chi_GraphicsContext& gc) const;
   virtual void print(ostream& os) const;
   virtual Chi_Shape* clone() const;
   
   Chi_Point vertex(int) const;
   Chi_Point center() const;

private:
   Chi_Array<Chi_Point> _vert;
};

/*-------------------------------------------------------------------------*/

EXPORT class Chi_Text : public Chi_Shape
/* PURPOSE:    Text on the screen
*/
{
public:
   Chi_Text() {}
   Chi_Text(Chi_Point, Chi_String);

   virtual void scale(Chi_Point center, double s);
   virtual void move(double x, double y);
   virtual void read(istream& is);

   virtual void plot(Chi_GraphicsContext& gc) const;
   virtual void print(ostream& os) const;
   virtual Chi_Shape* clone() const;

   Chi_Rectangle extent(Chi_GraphicsContext& gc) const;

private:
   Chi_Point _start;
   Chi_String _text;
};

/*-------------------------------------------------------------------------*/

EXPORT class Chi_ScalableText : public Chi_Text
/* PURPOSE:    Text on the screen that can change size
*/
{
public:
   Chi_ScalableText();
   Chi_ScalableText(Chi_Point, Chi_String, double size = 1.0);

   virtual void scale(Chi_Point center, double s);
   virtual void read(istream& is);

   virtual void plot(Chi_GraphicsContext& gc) const;
   virtual void print(ostream& os) const;
   virtual Chi_Shape* clone() const;

   Chi_Rectangle extent(Chi_GraphicsContext& gc) const;

private:
   double _size;
};

#endif

/* FUNCTIONS ***************************************************************/

inline Bool approx_eq(double x, double y)
/* PURPOSE:    tests whether two floating point numbers are approximately equal
   RECEIVES:   x, y - floating point numbers
   RETURNS:    TRUE iff |x-y| approx. 0
*/
{  const double EPS = 0.000001;
   return fabs(x - y) < EPS;
}

//............................................................................

inline Bool approx_le(double x, double y)
/* PURPOSE:    compares two floating point numbers, allowing for roundoff
   RECEIVES:   x, y - floating point numbers
   RETURNS:    TRUE iff x < y + roundoff
*/
{  const double EPS = 0.000001;
   return x - y < EPS;
}

/*-------------------------------------------------------------------------*/

Chi_Shape* Chi_Shape::clone() const
/* REMARKS:    This function is not pure to make it easier to redefine
               shapes in the early chapters of the book
*/
{  return 0;
}

/*-------------------------------------------------------------------------*/

Chi_Point::Chi_Point(double xx, double yy)
/* RECEIVES:   xx, yy - coordinates
*/
   : _x(xx), _y(yy)
{}

//............................................................................

Chi_Point::~Chi_Point() {}

//............................................................................

void Chi_Point::plot(Chi_GraphicsContext& gc) const
/* PURPOSE:    plot this shape
   RECEIVES:   gc - graphics context
*/
{  gc.point(_x, _y);
}

//............................................................................

void Chi_Point::print(ostream& os) const
/* PURPOSE:    print this shape to a stream
   RECEIVES:   os - an output streeam
*/
{  os << "(" << _x << "," << _y << ")";
}

//............................................................................

Chi_Shape* Chi_Point::clone() const
/* RETURNS:    a new copy of self
   REMARKS:    the caller must delete the returned object
*/
{  return new Chi_Point(self);
}

//............................................................................

void Chi_Point::read(istream& is)
/* PURPOSE:    read this shape from a stream
   RECEIVES:   is - an input stream
   REMARKS:    the old contents of self is overwritten
*/
{  char ch;
   is >> ch;
   if (is.fail()) return;
   if (ch != '(') { is.clear(ios::failbit); return; }
   is >> _x;
   is >> ch;
   if (is.fail()) return;
   if (ch != ',') { is.clear(ios::failbit); return; }
   is >> _y;
   is >> ch;
   if (is.fail()) return;
   if (ch != ')') { is.clear(ios::failbit); return; }
}

//............................................................................

void Chi_Point::scale(Chi_Point center, double s)
/* PURPOSE:    Scale this point by a factor from a center of projection
   RECEIVES:   center - the center of projection
               s - the scale factor
   REMARKS:    the point is moved along the line joining it with center
               if 0 < s < 1, it is moved closer to the center
*/
{  _x = center._x + (_x - center._x) * s;
   _y = center._y + (_y - center._y) * s;
}

//............................................................................

void Chi_Point::move(double x, double y)
/* PURPOSE:    Move this point by a vector
   RECEIVES:   x, y - the amount to move in x- and y-direction
*/
{  _x += x;
   _y += y;
}

//............................................................................

double Chi_Point::distance(Chi_Point b) const
/* RETURNS:    the Euclidean distance between self and b
*/
{  double dx = _x - b._x;
   double dy = _y - b._y;
   return sqrt(dx * dx + dy * dy);
}

//............................................................................

double Chi_Point::angle(Chi_Point b) const
/* RETURNS:    the angle between the x axis and the line joining self and b
*/
{  return atan2(b._y - _y, b._x - _x);
}

//............................................................................

#ifndef EXPORT

EXPORT inline double Chi_Point::x() const
/* RETURNS:    the x-coordinate
*/
{  return _x;
}

//............................................................................

EXPORT inline double Chi_Point::y() const
/* RETURNS:    the y-coordinate
*/
{  return _y;
}

#endif

/*-------------------------------------------------------------------------*/

Chi_Rectangle::Chi_Rectangle() {}

//............................................................................

Chi_Rectangle::Chi_Rectangle(Chi_Point pp, Chi_Point qq)
/* RECEIVES:   pp, qq - two corner points
*/
   : _corner1(pp), _corner2(qq)
{}

//............................................................................

void Chi_Rectangle::plot(Chi_GraphicsContext& gc) const
/* PURPOSE:    plot this shape
   RECEIVES:   gc - graphics context
*/
{  gc.rectangle(_corner1.x(), _corner1.y(), _corner2.x(), _corner2.y());
}

//............................................................................

void Chi_Rectangle::print(ostream& os) const
/* PURPOSE:    print this shape to a stream
   RECEIVES:   os - an output streeam
*/
{  _corner1.print(os);
   _corner2.print(os);
}

//............................................................................

Chi_Shape* Chi_Rectangle::clone() const
/* RETURNS:    a new copy of self
   REMARKS:    the caller must delete the returned object
*/
{  return new Chi_Rectangle(self);
}

//............................................................................

void Chi_Rectangle::read(istream& is)
/* PURPOSE:    read this shape from a stream
   RECEIVES:   is - an input stream
   REMARKS:    the old contents of self is overwritten
*/
{  _corner1.read(is);
   _corner2.read(is);
}

//............................................................................

void Chi_Rectangle::scale(Chi_Point center, double s)
/* PURPOSE:    Scale this point by a factor from a center of projection
   RECEIVES:   center - the center of projection
               s - the scale factor
*/
{  _corner1.scale(center, s);
   _corner2.scale(center, s);
}

//............................................................................

void Chi_Rectangle::move(double x, double y)
/* PURPOSE:    Move this rectangle by a vector
   RECEIVES:   x, y - the amount to move in x- and y-direction
*/
{  _corner1.move(x, y);
   _corner2.move(x, y);
}

//............................................................................

Bool Chi_Rectangle::is_inside(Chi_Point p) const
/* PURPOSE:    test whether a point falls inside a rectangle
   RECEIVES:   p - a point
   RETURNS:    TRUE iff p is inside this rectangle
*/
{  return xleft() <= p.x() && p.x() <= xright() &&
      ytop() <= p.y() && p.y() <= ybottom();
}

//............................................................................

Chi_Point Chi_Rectangle::left_top() const
/* RETURNS:    the left top corner point
*/
{  return Chi_Point(chi_min(_corner1.x(), _corner2.x()), chi_min(_corner1.y(), _corner2.y()));
}

//............................................................................

Chi_Point Chi_Rectangle::left_bottom() const
/* RETURNS:    the left bottom corner point
*/
{  return Chi_Point(chi_min(_corner1.x(), _corner2.x()), chi_max(_corner1.y(), _corner2.y()));
}

//............................................................................

Chi_Point Chi_Rectangle::right_top() const
/* RETURNS:    the right top corner point
*/
{  return Chi_Point(chi_max(_corner1.x(), _corner2.x()), chi_min(_corner1.y(), _corner2.y()));
}

//............................................................................

Chi_Point Chi_Rectangle::right_bottom() const
/* RETURNS:    the right bottom corner point
*/
{  return Chi_Point(chi_max(_corner1.x(), _corner2.x()), chi_max(_corner1.y(), _corner2.y()));
}

//............................................................................

double Chi_Rectangle::ytop() const
/* RETURNS:    the y-coordinate of the top of this rectangle
*/
{  return chi_min(_corner1.y(), _corner2.y());
}

//............................................................................

double Chi_Rectangle::ybottom() const
/* RETURNS:    the y-coordinate of the bottom of this rectangle
*/
{  return chi_max(_corner1.y(), _corner2.y());
}

//............................................................................

double Chi_Rectangle::xleft() const
/* RETURNS:    the x-coordinate of the left of this rectangle
*/
{  return chi_min(_corner1.x(), _corner2.x());
}

//............................................................................

double Chi_Rectangle::xright() const
/* RETURNS:    the x-coordinate of the right of this rectangle
*/
{  return chi_max(_corner1.x(), _corner2.x());
}

//............................................................................

double Chi_Rectangle::xsize() const
/* RETURNS:    the horizontal size
*/
{  return fabs(_corner2.x() - _corner1.x());
}

//............................................................................

double Chi_Rectangle::ysize() const
/* RETURNS:    the vertical size
*/
{  return fabs(_corner2.y() - _corner1.y());
}

//............................................................................

Chi_Point Chi_Rectangle::center() const
/* RETURNS: the center (i.e. center of gravity) of this shape
*/
{  return Chi_Point((_corner1.x() + _corner2.x()) / 2.0, (_corner1.y()+ _corner2.y()) / 2.0);
}

//............................................................................

Chi_Point Chi_Rectangle::boundary_point(Chi_Point p) const
/* PURPOSE:    Compute a point on the boundary of the shape
   RECEIVES:   p - a point distinct from the center
   RETURNS:    the intersection of the line joining the center and p with
               the boundary
   REMARKS:    The computation is easiest to understand if the rectangle
               is the unit square. Let p = (x, y). Then the boundary point
               b is
                  (x / |y|, sign(y)) if |x| <= |y|
                  (sign(x), y / |x|) if |x| >= |y|
               For a general rectangle, first subtract the center from
               the coordinates and then divide by xsize/2, ysize/2.
*/
{  Chi_Point c = center();
   double x = p.x() - c.x();
   double y = p.y() - c.y();
   if (approx_eq(x, 0) && approx_eq(y, 0)) return c;
   double xs2 = xsize() / 2;
   double ys2 = ysize() / 2;
   double xys = x * ys2;
   double yxs = y * xs2;

   if (fabs(xys) <= fabs(yxs)) // meets the top or bottom
   {  double bx = xys / y;
      if (y > 0) return Chi_Point(c.x() + bx, c.y() + ys2);
      else return Chi_Point(c.x() - bx, c.y() - ys2);
   }
   else // meets the left or right
   {  double by = yxs / x;
      if (x > 0) return Chi_Point(c.x() + xs2, c.y() + by);
      else return Chi_Point(c.x() - xs2, c.y() - by);
   }
}

/*-------------------------------------------------------------------------*/

Chi_FilledRect::Chi_FilledRect()
: _pattern(Chi_GraphicsContext::SOLID_BRUSH)
{}

//............................................................................

Chi_FilledRect::Chi_FilledRect(Chi_Point p, Chi_Point q, Chi_GraphicsContext::BrushStyle b)
/* RECEIVES:   p, q - corner points
               b - fill pattern
*/
: _pattern(b), Chi_Rectangle(p, q)
{}

//............................................................................

void Chi_FilledRect::plot(Chi_GraphicsContext& gc) const
/* PURPOSE:    plot this shape
   RECEIVES:   gc - graphics context
*/
{  gc.set_brush_style(_pattern);
   Chi_Rectangle::plot(gc);
   gc.set_brush_style(Chi_GraphicsContext::EMPTY_BRUSH);
}

/*-------------------------------------------------------------------------*/

Chi_Segment::Chi_Segment() {}

//............................................................................

Chi_Segment::Chi_Segment(Chi_Point f, Chi_Point t)
/* RECEIVES:   f, t - the end points
*/
:  _from(f),
   _to(t)
{}

//............................................................................

Chi_Point Chi_Segment::from() const
/* RETURNS:    the first end point
*/
{  return _from;
}

//............................................................................

Chi_Point Chi_Segment::to() const
/* RETURNS:    the other end point
*/
{  return _to;
}

//............................................................................

void Chi_Segment::plot(Chi_GraphicsContext& gc) const
/* PURPOSE:    plot this shape
   RECEIVES:   gc - graphics context
*/
{  gc.line(_from.x(), _from.y(), _to.x(), _to.y());
}

//............................................................................

void Chi_Segment::print(ostream& os) const
/* PURPOSE:    print this shape to a stream
   RECEIVES:   os - an output streeam
*/
{  _from.print(os);
   _to.print(os);
}

//............................................................................

Chi_Shape* Chi_Segment::clone() const
/* RETURNS:    a new copy of self
   REMARKS:    the caller must delete the returned object
*/
{  return new Chi_Segment(self);
}

//............................................................................

void Chi_Segment::read(istream& is)
/* PURPOSE:    read this shape from a stream
   RECEIVES:   is - an input stream
   REMARKS:    the old contents of self is overwritten
*/
{  _from.read(is);
   _to.read(is);
}

//............................................................................

void Chi_Segment::scale(Chi_Point center, double s)
/* PURPOSE:    Scale this point by a factor from a center of projection
   RECEIVES:   center - the center of projection
               s - the scale factor
*/
{  _from.scale(center, s);
   _to.scale(center, s);
}

//............................................................................

void Chi_Segment::move(double x, double y)
/* PURPOSE:    Move this rectangle by a vector
   RECEIVES:   x, y - the amount to move in x- and y-direction
*/
{  _from.move(x, y);
   _to.move(x, y);
}

//............................................................................

Chi_Point Chi_Segment::center() const
/* RETURNS: the center (i.e. center of gravity) of this shape
*/
{  return Chi_Point((_from.x() + _to.x()) / 2.0, (_from.y()+ _to.y()) / 2.0);
}

//............................................................................

Bool Chi_Segment::intersect(Chi_Segment seg, Chi_Point& p) const
/* PURPOSE:    Computes whether two segments intersect
   RECEIVES:   seg - another segment
               p (OUT) - the point of intersection
   RETURNS:    TRUE iff this segment intersects b
   REMARKS:    The line joining a and b is
                  s * a + (1 - s) * b.
               The line joining c and d is
                  t * c + (1 - t) * d.
               We solve for (s,t) in
                  s * a + (1 - s) * b = t * c + (1 - t) * d
               by solving the system of two linear equations. Then
               we check whether the solution fulfills 0 <= s <= 1 and
               0 <= t <= 1.
*/
{  Chi_Point a = _from;
   Chi_Point b = _to;
   Chi_Point c = seg._from;
   Chi_Point d = seg._to;
   double msx, mtx, msy, mty; // the matrix to invert
   double rx, ry; // the right hand side of the system
   msx = a.x() - b.x();
   mtx = -c.x() + d.x();
   msy = a.y() - b.y();
   mty = -c.y() + d.y();
   rx = -b.x() + d.x();
   ry = -b.y() + d.y();
   double det = msx * mty - msy * mtx;
   if (approx_eq(det, 0)) return FALSE; // the lines are parallel
   double s, t; // the solutions of the system
   s = (mty * rx - mtx * ry) / det;
   t = (-msy * rx + msx * ry) / det;
   if (!approx_le(0, s) || !approx_le(s, 1)
      || !approx_le(0, t) || !approx_le(t, 1))
      return FALSE; // intersection out of range
   p = Chi_Point(t * c.x() + (1 - t) * d.x(), t * c.y() + (1 - t) * d.y());
   return TRUE;
}

//............................................................................

double Chi_Segment::distance(Chi_Point p) const
/* PURPOSE:    compute the distance between a point and a segment
   RECEIVES:   p - a point
   RETURNS:    the distance between p and this segment
*/
{  return p.distance(closest(p));
}

//............................................................................

Chi_Point Chi_Segment::closest(Chi_Point p) const
/* PURPOSE:    compute the closest point to a given point on a line segment
   RECEIVES:   p - the point
   RETURNS:    the closest point on the line segment
   REMARKS:    First make a the origin, i.e. subtract a from both p and b.
               If that makes b = 0, then ab is just a point, and the
               result is a.
               The projection of p onto b is b<p,b>/<b,b>. (<.,.> is scalar
               product.)
               This lies on the line segment if 0 <= <p,b>/<b,b> <= 1.
               If < 0, the closest point is a, if > 1, it is b.
*/
{  Chi_Point a = _from;
   Chi_Point b = _to;
   p.move(-a.x(), -a.y());
   Chi_Point r = b;
   b.move(-a.x(), -a.y());
   double blen = b.x() * b.x() + b.y() * b.y();
   if (approx_eq(blen, 0)) return a;
   double proj = (b.x() * p.x() + b.y() * p.y())/blen;
   if (approx_le(1, proj)) return r;
   if (approx_le(proj, 0)) return a;
   a.move(b.x() * proj, b.y() * proj);
   return a;
}

/*-------------------------------------------------------------------------*/

void Chi_Segment::reflect(double& x, double& y) const
/* PURPOSE:    Reflect a vector along a line
   RECEIVES:   a, b - two points on the reflection line
               x, y (IN OUT) - the coordinates of a vector
   REMARKS:    Let c be the vector joining _from and _to
               Let v = (x,y). Then v can be decomposed into the projection
               p onto the reflection line and the perpendicular part q.
                  v = p + q,
                  p = c <c,v>/<c,c>
               where <.,.> is scalar product.
               The reflected vector is then
                  r = 2 * p - v
*/
{  double cx = _from.x() - _to.x();
   double cy = _from.y() - _to.y();
   double clen = (cx * cx + cy * cy);
   if (approx_eq(clen, 0)) return; // ab is a point--don't reflect
   double proj = (cx * x + cy * y)/clen;
   x = 2 * cx * proj - x;
   y = 2 * cy * proj - y;
}

/*-------------------------------------------------------------------------*/

Chi_Ellipse::Chi_Ellipse() {}

//............................................................................

Chi_Ellipse::Chi_Ellipse(Chi_Point c, double a, double b)
/* RECEIVES:   c - the center
               a, b - the radii of the axes parallel to the x- and y-axis
*/
:  _center(c),
   _xradius(a),
   _yradius(b)
{}

//............................................................................

double Chi_Ellipse::xradius() const
{  return _xradius;
}

//............................................................................

double Chi_Ellipse::yradius() const
{  return _yradius;
}

//............................................................................

void Chi_Ellipse::plot(Chi_GraphicsContext& gc) const
/* PURPOSE:    plot this shape
   RECEIVES:   gc - graphics context
*/
{  gc.ellipse(_center.x(), _center.y(), _xradius, _yradius);
}

//............................................................................

void Chi_Ellipse::print(ostream& os) const
/* PURPOSE:    print this shape to a stream
   RECEIVES:   os - an output streeam
*/
{  _center.print(os);
   os << " " << _xradius << " " << _yradius;
}

//............................................................................

Chi_Shape* Chi_Ellipse::clone() const
/* RETURNS:    a new copy of self
   REMARKS:    the caller must delete the returned object
*/
{  return new Chi_Ellipse(self);
}

//............................................................................

void Chi_Ellipse::read(istream& is)
/* PURPOSE:    read this shape from a stream
   RECEIVES:   is - an input stream
   REMARKS:    the old contents of self is overwritten
*/
{  _center.read(is);
   is >> _xradius >> _yradius;
}

//............................................................................

void Chi_Ellipse::scale(Chi_Point center, double s)
/* PURPOSE:    Scale this point by a factor from a center of projection
   RECEIVES:   center - the center of projection
               s - the scale factor
*/
{  _center.scale(center, s);
   _xradius *= s;
   _yradius *= s;
}

//............................................................................

void Chi_Ellipse::move(double x, double y)
/* PURPOSE:    Move this rectangle by a vector
   RECEIVES:   x, y - the amount to move in x- and y-direction
*/
{  _center.move(x, y);
}

//............................................................................

Chi_Point Chi_Ellipse::center() const
/* RETURNS: the center (i.e. center of gravity) of this shape
*/
{  return _center;
}

//............................................................................

Bool Chi_Ellipse::is_inside(Chi_Point p) const
/* PURPOSE:    test whether a point is inside the ellipse
   RECEIVES:   p - a point
   RETURNS:    TRUE iff p is in the inside (or on the boundary)
*/
{  double x = (p.x() - _center.x())/_xradius;
   double y = (p.y() - _center.y())/_yradius;
   return approx_le(x * x + y * y, 1);
}

//............................................................................

Chi_Point Chi_Ellipse::boundary_point(Chi_Point p) const
/* PURPOSE:    Compute a point on the boundary of the shape
   RECEIVES:   p - a point distinct from the center
   RETURNS:    the intersection of the line joining the center and p with
               the boundary
*/
{  double a = _center.angle(p);
   Chi_Point r = _center;
   r.move(_xradius * cos(a), _yradius * sin(a));
   return r;
}

/*-------------------------------------------------------------------------*/

Chi_Polygon::Chi_Polygon()
/* REMARKS:    Makes a useless null polygon
*/
{}

//............................................................................

Chi_Polygon::Chi_Polygon(int n)
/* RECEIVES:   n - the number of vertices
*/
:  _vert(1, n)
{}

//............................................................................

Chi_Point Chi_Polygon::vertex(int i) const
/* RECEIVES:   i - an integer
   RETURNS:    the ith vertex
*/
{  return _vert.get(i);
}

//............................................................................

void Chi_Polygon::set_vertex(int i , Chi_Point p)
/* PURPOSE:    set a vertex of the polygon
   RECEIVES:   i - an integer between 1 and the number of vertices
               p - the point to store
*/
{  _vert[i] = p;
}

//............................................................................

void Chi_Polygon::plot(Chi_GraphicsContext& gc) const
/* PURPOSE:    plot this shape
   RECEIVES:   gc - graphics context
*/
{  for (int i = _vert.low(); i <= _vert.high(); i++)
   {  int j = i + 1;
      if (j > _vert.high()) j = _vert.low();
      Chi_Point p = _vert.get(i);
      Chi_Point q = _vert.get(j);
      gc.line(p.x(), p.y(), q.x(), q.y());
   }
}

//............................................................................

void Chi_Polygon::print(ostream& os) const
/* PURPOSE:    print this shape to a stream
   RECEIVES:   os - an output streeam
*/
{  os << _vert.length() << " ";
   for (int i = _vert.low(); i <= _vert.high(); i++)
   {  _vert[i].print(os);
   }
}

//............................................................................

Chi_Shape* Chi_Polygon::clone() const
/* RETURNS:    a new copy of self
   REMARKS:    the caller must delete the returned object
*/
{  return new Chi_Polygon(self);
}

//............................................................................

void Chi_Polygon::read(istream& is)
/* PURPOSE:    read this shape from a stream
   RECEIVES:   is - an input stream
   REMARKS:    the old contents of self is overwritten
*/
{  int n;
   is >> n;
   _vert.empty();
   _vert.grow(1, n);
   for (int i = 1; i <= n; i++)
   {  _vert[i].read(is);
   }
}

//............................................................................

void Chi_Polygon::scale(Chi_Point center, double s)
/* PURPOSE:    Scale this point by a factor from a center of projection
   RECEIVES:   center - the center of projection
               s - the scale factor
*/
{  for (int i = _vert.low(); i <= _vert.high(); i++)
      _vert[i].scale(center, s);
}

//............................................................................

void Chi_Polygon::move(double x, double y)
/* PURPOSE:    Move this polygon by a vector
   RECEIVES:   x, y - the amount to move in x- and y-direction
*/
{  for (int i = _vert.low(); i <= _vert.high(); i++)
      _vert[i].move(x, y);
}

//............................................................................

Chi_Point Chi_Polygon::center() const
/* RETURNS: the center (i.e. center of gravity) of this shape
*/
{  CHI_ASSERT(_vert.length() > 0);
   Chi_Point p;
   for (int i = _vert.low(); i <= _vert.high(); i++)
      p = Chi_Point(p.x() + _vert[i].x(), p.y() + _vert[i].y());
   return Chi_Point(p.x() / _vert.length(), p.y() + _vert.length());
}

/*-------------------------------------------------------------------------*/

Chi_Text::Chi_Text(Chi_Point s, Chi_String t)
/* RECEIVES:   s - upper left corner
               t - text string
*/
:  _start(s), _text(t)
{}

//............................................................................

void Chi_Text::plot(Chi_GraphicsContext& gc) const
/* PURPOSE:    plot this shape
   RECEIVES:   gc - graphics context
*/
{  gc.text(_text, _start.x(), _start.y());
}

//............................................................................

Chi_Rectangle Chi_Text::extent(Chi_GraphicsContext& gc) const
/* PURPOSE:    compute the size of a text item
   RECEIVES:   the graphics context on which the text will be rendered
               (it is necessary to determine the character sizes)
   RETURNS:    the rectangle enclosing the text
*/
{  double x, y;
   gc.text_extent(_text, x, y);
   Chi_Point end = _start;
   end.move(x, y);
   return Chi_Rectangle(_start, end);
}

//............................................................................

void Chi_Text::scale(Chi_Point center, double s)
/* PURPOSE:    Scale this shape by a factor from a center of projection
   RECEIVES:   center - the center of projection
               s - the scale factor
*/
{  _start.scale(center, s);
}

//............................................................................

void Chi_Text::move(double x, double y)
/* PURPOSE:    Move this polygon by a vector
   RECEIVES:   x, y - the amount to move in x- and y-direction
*/
{  _start.move(x, y);
}

//............................................................................

void Chi_Text::read(istream& is)
/* PURPOSE:    read this shape from a stream
   RECEIVES:   is - an input stream
   REMARKS:    the old contents of self is overwritten
*/
{  _start.read(is);
   _text.read_delimited(is);
}

//............................................................................

void Chi_Text::print(ostream& os) const
/* PURPOSE:    print this shape to a stream
   RECEIVES:   os - an output streeam
*/
{  _start.print(os);
   os << " ";
   _text.write_delimited(os);
}

//............................................................................

Chi_Shape* Chi_Text::clone() const
/* RETURNS:    a new copy of self
   REMARKS:    the caller must delete the returned object
*/
{  return new Chi_Text(self);
}

//****************************************************************************

Chi_ScalableText::Chi_ScalableText()
:  _size(1.0)
{}

//............................................................................

Chi_ScalableText::Chi_ScalableText(Chi_Point ll, Chi_String t, double size) :
   Chi_Text(ll, t), _size(size)
/* RECEIVES:   s - upper left corner
               t - text string
               size - vertical height
*/
{}

//............................................................................

void Chi_ScalableText::plot(Chi_GraphicsContext& gc) const
/* PURPOSE:    plot this shape
   RECEIVES:   gc - graphics context
*/
{  gc.set_font_style(Chi_GraphicsContext::SANS_FONT);
   gc.set_font_size(_size);
   Chi_Text::plot(gc);
   gc.set_font_style(Chi_GraphicsContext::SYS_FONT);
}

//............................................................................

Chi_Rectangle Chi_ScalableText::extent(Chi_GraphicsContext& gc) const
/* PURPOSE:    compute the size of a text item
   RECEIVES:   the graphics context on which the text will be rendered
               (it is necessary to determine the character sizes)
   RETURNS:    the rectangle enclosing the text
*/
{  gc.set_font_style(Chi_GraphicsContext::SANS_FONT);
   gc.set_font_size(_size);
   Chi_Rectangle r = Chi_Text::extent(gc);
   gc.set_font_style(Chi_GraphicsContext::SYS_FONT);
   return r;
}

//............................................................................

void Chi_ScalableText::scale(Chi_Point center, double s)
/* PURPOSE:    Scale this shape by a factor from a center of projection
   RECEIVES:   center - the center of projection
               s - the scale factor
*/
{  Chi_Text::scale(center, s);
   _size = _size * s;
}

//............................................................................

void Chi_ScalableText::read(istream& is)
/* PURPOSE:    read this shape from a stream
   RECEIVES:   is - an input stream
   REMARKS:    the old contents of self is overwritten
*/
{  Chi_Text::read(is);
   is >> _size;
}

//............................................................................

void Chi_ScalableText::print(ostream& os) const
/* PURPOSE:    print this shape to a stream
   RECEIVES:   os - an output streeam
*/
{  Chi_Text::print(os);
   os << " " << _size;
}

//............................................................................

Chi_Shape* Chi_ScalableText::clone() const
/* RETURNS:    a new copy of self
   REMARKS:    the caller must delete the returned object
*/
{  return new Chi_ScalableText(self);
}

/* TEST ********************************************************************//***************************************************************************//***************************************************************************/

#ifdef CHISHAPE_TEST

int chi_main(int, char**)
{  Chi_PtrArray<Chi_Shape> fig;
   Chi_GraphicsContext gc;
   fig.append(new Chi_Point(1, 1));
#ifndef CHI_BC45
   fig.append(new Chi_Rectangle(Chi_Point(0.1, 2.2), Chi_Point(1.9, 3.8)));
   fig.append(new Chi_FilledRect(Chi_Point(0.2, 4.1), Chi_Point(1.8, 5.9),
      Chi_GraphicsContext::XDIAG_BRUSH));
   fig.append(new Chi_Segment(Chi_Point(0.1, 6.2), Chi_Point(1.9, 7.8)));
#else
   // code above causes internal error in Borland C++ 4.5
   Chi_Point p2 = Chi_Point(1.9, 3.8);
   fig.append(new Chi_Rectangle(Chi_Point(0.1, 2.2), p2));
   p2 = Chi_Point(1.8, 5.9);
   fig.append(new Chi_FilledRect(Chi_Point(0.2, 4.1), p2,
      Chi_GraphicsContext::XDIAG_BRUSH));
   p2 = Chi_Point(1.9, 7.8);
   fig.append(new Chi_Segment(Chi_Point(0.1, 6.2), p2));
#endif
   fig.append(new Chi_Ellipse(Chi_Point(1, 9), 0.8, 0.6));
   Chi_Polygon* p = new Chi_Polygon(3);
   p->set_vertex(1, Chi_Point(0.1, 10.1));
   p->set_vertex(2, Chi_Point(1.9, 11));
   p->set_vertex(3, Chi_Point(1, 11.9));
   fig.append(p);
   fig.append(new Chi_ScalableText(Chi_Point(0.1, 12.1), "Hi"));
   fig.append(new Chi_Text(Chi_Point(0.1, 14.1), "Hi"));
   for (int i = 1; i < 10; i++)
   {  for (int j = fig.low(); j <= fig.high(); j++)
      {  fig[j]->plot(gc);
         fig[j]->move(2, 0);
         fig[j]->scale(Chi_Point(2 * i - 0.5, 8), 0.9);
      }
   }
   gc.close();
   return 0;
}

#endif

/***************************************************************************//***************************************************************************//***************************************************************************/
