#ifndef CHISHAPE_H
#define CHISHAPE_H

/****************************************************************************
** COPYRIGHT (C):    1994 Cay S. Horstmann. All Rights Reserved.
** PROJECT:          ChiLib
** FILE:             chishape.h
** PURPOSE:          Shapes for graphic display
** VERSION           1.0
** LANGUAGE:         Borland  C++ 3.1/4.0
** TARGET:           IBM PC
** PROGRAMMER:       Cay Horstmann (CSH)
** START DATE:       2-19-94  (CSH)
** UPDATE HISTORY:
**
****************************************************************************/

#line 17 "chishape.cpp"

#include "chisetup.h"

#line 21 "chishape.cpp"

#include <iostream.h>

#line 25 "chishape.cpp"

#ifndef CHIARRAY_H
#include "chiarray.h"
#endif

#line 27 "chishape.cpp"

#ifndef CHIGC_H
#include "chigc.h"
#endif

#line 29 "chishape.cpp"

#ifndef CHISTRNG_H
#include "chistrng.h"
#endif

#line 37 "chishape.cpp"


 class Chi_Point;
 class Chi_Segment;

 class Chi_Shape
/* PURPOSE:    A shape that can be plotted and moved around
*/
{
public:
   virtual void scale(Chi_Point center, double s) = 0;
   virtual void move(double x, double y) = 0;
   virtual void read(istream& is) = 0;

   virtual void plot(Chi_GraphicsContext& gc) const = 0;
   virtual void print(ostream& os) const = 0;
   virtual Chi_Shape* clone() const;

   virtual ~Chi_Shape() {}
};

/*-------------------------------------------------------------------------*/

 class Chi_Point : public Chi_Shape
/* PURPOSE:    A single point
*/
{
public:
   Chi_Point(double xx =0, double yy =0);

   virtual void scale(Chi_Point center, double s);
   virtual void move(double x, double y);
   virtual void read(istream& is);

   virtual void plot(Chi_GraphicsContext& gc) const;
   virtual void print(ostream& os) const;
   virtual Chi_Shape* clone() const;

   double x() const;
   double y() const;

   double distance(Chi_Point b) const;
   double angle(Chi_Point b) const;

   ~Chi_Point();
private:
   double _x;
   double _y;
};

/*-------------------------------------------------------------------------*/

 class Chi_Rectangle : public Chi_Shape
/* PURPOSE:    A rectangle (with sides parallel to the coordinate axes)
*/
{
public:
   Chi_Rectangle();
   Chi_Rectangle(Chi_Point, Chi_Point);

   virtual void scale(Chi_Point center, double s);
   virtual void move(double x, double y);
   virtual void read(istream& is);

   virtual void plot(Chi_GraphicsContext& gc) const;
   virtual void print(ostream& os) const;
   virtual Chi_Shape* clone() const;

   Chi_Point left_top() const;
   Chi_Point left_bottom() const;
   Chi_Point right_top() const;
   Chi_Point right_bottom() const;
   Chi_Point center() const;

   double ytop() const;
   double ybottom() const;
   double xleft() const;
   double xright() const;
   double xsize() const;
   double ysize() const;

   Bool is_inside(Chi_Point p) const;
   Chi_Point boundary_point(Chi_Point p) const;

private:
   Chi_Point _corner1;
   Chi_Point _corner2;
};

/*-------------------------------------------------------------------------*/

class Chi_FilledRect : public Chi_Rectangle
/* PURPOSE:    A rectangle that is filled with a pattern
*/
{
public:
   Chi_FilledRect();
   Chi_FilledRect(Chi_Point, Chi_Point, Chi_GraphicsContext::BrushStyle);
   virtual void plot(Chi_GraphicsContext& gc) const;
private:
   Chi_GraphicsContext::BrushStyle _pattern;
};

/*-------------------------------------------------------------------------*/

 class Chi_Segment : public Chi_Shape
/* PURPOSE:    A line segment
*/
{
public:
   Chi_Segment();
   Chi_Segment(Chi_Point f, Chi_Point t);

   virtual void scale(Chi_Point center, double s);
   virtual void move(double x, double y);
   virtual void read(istream& is);

   virtual void plot(Chi_GraphicsContext& gc) const;
   virtual void print(ostream& os) const;
   virtual Chi_Shape* clone() const;

   Chi_Point center() const;
   Bool intersect(Chi_Segment s, Chi_Point& p) const;
   Chi_Point closest(Chi_Point p) const;
   double distance(Chi_Point p) const;
   void reflect(double& x, double& y) const;

   Chi_Point from() const;
   Chi_Point to() const;

private:
   Chi_Point _from;
   Chi_Point _to;
};

/*-------------------------------------------------------------------------*/

 class Chi_Ellipse : public Chi_Shape
/* PURPOSE:    An ellipse (with axes parallel to the coordinate axes)
*/
{
public:
   Chi_Ellipse();
   Chi_Ellipse(Chi_Point, double, double);

   virtual void scale(Chi_Point center, double s);
   virtual void move(double x, double y);
   virtual void read(istream& is);

   virtual void plot(Chi_GraphicsContext& gc) const;
   virtual void print(ostream& os) const;
   virtual Chi_Shape* clone() const;

   Chi_Point center() const;
   double xradius() const;
   double yradius() const;
   Bool is_inside(Chi_Point p) const;
   Chi_Point boundary_point(Chi_Point p) const;

private:
   Chi_Point _center;
   double _xradius;
   double _yradius;
};

/*-------------------------------------------------------------------------*/

 class Chi_Polygon : public Chi_Shape
/* PURPOSE:    A closed (not necessarily convex) polygon
*/
{
public:
   Chi_Polygon();
   Chi_Polygon(int);

   void set_vertex(int, Chi_Point);

   virtual void scale(Chi_Point center, double s);
   virtual void move(double x, double y);
   virtual void read(istream& is);

   virtual void plot(Chi_GraphicsContext& gc) const;
   virtual void print(ostream& os) const;
   virtual Chi_Shape* clone() const;
   
   Chi_Point vertex(int) const;
   Chi_Point center() const;

private:
   Chi_Array<Chi_Point> _vert;
};

/*-------------------------------------------------------------------------*/

 class Chi_Text : public Chi_Shape
/* PURPOSE:    Text on the screen
*/
{
public:
   Chi_Text() {}
   Chi_Text(Chi_Point, Chi_String);

   virtual void scale(Chi_Point center, double s);
   virtual void move(double x, double y);
   virtual void read(istream& is);

   virtual void plot(Chi_GraphicsContext& gc) const;
   virtual void print(ostream& os) const;
   virtual Chi_Shape* clone() const;

   Chi_Rectangle extent(Chi_GraphicsContext& gc) const;

private:
   Chi_Point _start;
   Chi_String _text;
};

/*-------------------------------------------------------------------------*/

 class Chi_ScalableText : public Chi_Text
/* PURPOSE:    Text on the screen that can change size
*/
{
public:
   Chi_ScalableText();
   Chi_ScalableText(Chi_Point, Chi_String, double size = 1.0);

   virtual void scale(Chi_Point center, double s);
   virtual void read(istream& is);

   virtual void plot(Chi_GraphicsContext& gc) const;
   virtual void print(ostream& os) const;
   virtual Chi_Shape* clone() const;

   Chi_Rectangle extent(Chi_GraphicsContext& gc) const;

private:
   double _size;
};


#line 413 "chishape.cpp"


 inline double Chi_Point::x() const
/* RETURNS:    the x-coordinate
*/
{  return _x;
}

//............................................................................

 inline double Chi_Point::y() const
/* RETURNS:    the y-coordinate
*/
{  return _y;
}


#endif
