#ifndef CHISTRNG_H
#define CHISTRNG_H

/****************************************************************************
** COPYRIGHT (C):    1992...1994 Cay S. Horstmann. All Rights Reserved.
** PROJECT:          ChiLib
** FILE:             chistrng.h
** PURPOSE:          a simple "value semantics" string class
** VERSION           1.3
** PROGRAMMER:       Cay Horstmann (CSH)
** START DATE:       6-10-92 (CSH)
** UPDATE HISTORY:

4-15-93 CSH
Cleaned up and added functions receiving const char[]

6-17-93 CSH
Changed logic to remove one level of indirection

8-23-93 CSH
Added field functions, read_line, operator const char*

9-22-93 CSH remove_field
Now returns the field without trailing spaces to exactly match the return
value of field(). (Trailing spaces are still removed.)

10-7-93 CSH set, []
Added call to unique to disassociate from shared strings before modification

1-11-94 CSH
Changed names to conform to proposed ANSI standard
Changed (from, to) range specs to (from, n) to conform to proposed ANSI
standard

4-28-94 CSH
Switched to chierror for error handling

10-5-95 CSH
Removed non-portable functions gcvt, ltoa and stricmp

****************************************************************************/

#line 184 "chistrng.cpp"

#include "chisetup.h"

#line 189 "chistrng.cpp"

#include <string.h>

#line 194 "chistrng.cpp"

#include <limits.h>

#line 197 "chistrng.cpp"

#include <iostream.h>

#line 207 "chistrng.cpp"


class Chi_String
/* PURPOSE:       A string class
*/
{
public:
   enum {NPOS = INT_MAX}; // maximum number of positions in string

   Chi_String();                       // constructs an empty string
   Chi_String(const char s[]);
   Chi_String(char ch);

   char& operator[](int i);
   void set(int i, char ch);

   void insert(int at, const Chi_String& s);
   void insert(int at, const char s[]);
   void insert(int at, char ch);
   void insert_field(int, const Chi_String&);

   void append(const Chi_String& s);
   void append(const char s[]);
   void append(char ch);
   const Chi_String& operator+=(const Chi_String& s);
   const Chi_String& operator+=(const char s[]);
   const Chi_String& operator+=(char ch);

   Chi_String remove(int from, int n = NPOS);
   char remove_char(int at);
   Chi_String remove_field(int);

   void replace(int from, int n, const Chi_String& s);
   void replace(int from, int n, const char s[]);
   void replace(int from, int n, char ch);
   Chi_String replace_field(int, const Chi_String&);

   Chi_String& to_lower();
   Chi_String& to_upper();
   void from_long(long n);
   void from_double(double d);

   void grow(int n);
   void shrink(int n);
   void empty();

   int read_line(istream& is);
   void read_delimited(istream& is);

   int length() const;
   Bool is_empty() const;

   char operator[](int i) const;
   char get(int i) const;
   Chi_String substr(int from, int n = NPOS) const;
   Chi_String field(int) const;

   char* c_array(char buf[], size_t buflen, int from = 0, int n = NPOS) const;
   const char* c_str() const;
   char* new_c_array() const;

#ifdef CHI_SC70
   int compare_to(const Chi_String&) const;
   int compare_insens_to(const Chi_String&) const;
   int compare_to(const char s[]) const;
   int compare_to(const Chi_String&, int n) const;
   int compare_to(const char s[], int n) const;
   int compare_insens_to(const char s[]) const;
   // Symantec 7.0 can't handle name conflict with static function
#else
   int compare(const Chi_String&) const;
   int compare_insens(const Chi_String&) const;
   int compare(const char s[]) const;
   int compare(const Chi_String&, int n) const;
   int compare(const char s[], int n) const;
   int compare_insens(const char s[]) const;
#endif

   int find(char c, int from = 0) const;
   int find(const Chi_String& s, int from = 0) const;
   int find(const char s[], int from = 0) const;

   int rfind(char c, int from = NPOS) const;

   int find_first_of(const Chi_String& s, int from = 0) const;
   int find_first_of(const char s[], int from = 0) const;
   int find_first_not_of(const Chi_String& s, int from = 0) const;
   int find_first_not_of(const char s[], int from = 0) const;
   int find_last_of(const Chi_String& s, int from = 0) const;
   int find_last_of(const char s[], int from = 0) const;

   long to_long() const;
   double to_double() const;

   void write_delimited(ostream& os) const;

   Chi_String operator+(const Chi_String&) const;
   Chi_String operator+(const char b[]) const;
   Chi_String operator+(char) const;

   static int compare(const Chi_String& a, const Chi_String& b);
   static int compare_insens(const Chi_String& a, const Chi_String& b);
   static int hash(const Chi_String& a);

   friend Chi_String operator+(const char a[], const Chi_String& b);
   friend Chi_String operator+(char ch, const Chi_String& b);

   Chi_String(const Chi_String&);
   const Chi_String& operator=(const Chi_String&);
   ~Chi_String();

private:
   Chi_String(size_t len, const char s[], size_t slen);
   Chi_String(int);  /* left undefined--to avoid accidental conversion
                        int -> char -> Chi_String
                     */
   void detach();
   void unique();

   void field_range(int f, int& from, int& len) const;

   char* _str;
};

/*-------------------------------------------------------------------------*/

class Chi_StringHeader
/* PURPOSE:    The header block with reference count and available space
   REMARKS:    An instance of this structure is BELOW each string
               that is allocated on the collectible heap
*/
{
public:
   int unique() { return _rc == 1; }
   void incref() { ++_rc; }
   int decref() { return --_rc; }

   size_t length() const { return _length; }

   static char* alloc(const char[]);
   static char* alloc(int, const char[], int);
   static void free(char*);

   static Chi_StringHeader* header(const char* p) { return (Chi_StringHeader*)p - 1; }

private:
   char* start() const { return (char*)(this+1); }

   unsigned _rc;                       // reference count
   size_t _length;

friend class Chi_String;
};

#line 380 "chistrng.cpp"
extern Chi_String operator+(const char a[], const Chi_String& b);

#line 393 "chistrng.cpp"
extern Chi_String operator+(char ch, const Chi_String& b);

#line 405 "chistrng.cpp"
extern ostream& operator<<(ostream& os, const Chi_String& s);

#line 416 "chistrng.cpp"
extern istream& operator>>(istream& is, Chi_String& s);

#line 439 "chistrng.cpp"


inline Chi_String::Chi_String()
/* REMARKS:    Constructs an empty string
*/
:  _str(0)
{}

/*...........................................................................*/

inline Chi_String::Chi_String(const char s[])
/* RECEIVES:   s - C character array
*/
:  _str(Chi_StringHeader::alloc(s))
{}

/*...........................................................................*/

inline Chi_String::Chi_String(char ch)
/* RECEIVES:   ch - single character
*/
:  _str(Chi_StringHeader::alloc(1, &ch, 1))
{}

/*...........................................................................*/

inline Chi_String::Chi_String(size_t tlen, const char s[], size_t slen)
/* RECEIVES:      tlen - total length of the string
                  s - character string to initialize first bytes
                  slen - length of s
   REMARKS:       (1) The string gets a 0 at position len
                  (2) This constructor is private, and its interface is made
                      slightly obscure on purpose to avoid accidental invocation
                      by class users. Originally we had a Chi_String(size_t),
                      and under some circumstances the compiler complained about
                      ambiguities with String(char)
*/
:  _str(Chi_StringHeader::alloc(tlen, s, slen))
{}

/*.........................................................................*/

inline Chi_String::Chi_String(const Chi_String& b)
:  _str(b._str)
{  if (_str) Chi_StringHeader::header(_str)->incref(); }

/*...........................................................................*/

inline void Chi_String::empty()
/* PURPOSE:    Empties the string
*/
{  detach();
   _str = 0;
}

/*.........................................................................*/

inline Chi_String::~Chi_String()
{  detach();
}

/*...........................................................................*/

inline int Chi_String::length() const
/* RETURNS:       the length of the string
*/
{  return _str ? Chi_StringHeader::header(_str)->_length : 0;
}

/*...........................................................................*/

inline Bool Chi_String::is_empty() const
/* RETURNS:       TRUE iff the string is empty
*/
{  return length() == 0;
}

/*...........................................................................*/

inline char Chi_String::get(int i) const
/* RETURNS:       the ith character in a string (0 = first character)
   REMARKS:       returns 0 if index is out of bounds
*/
{  return 0 <= i && i < length() ? _str[i] : 0;
}

//............................................................................

inline const char* Chi_String::c_str() const
/* PURPOSE:    get an internal pointer to the content of the string
   RETURNS:    the character array of the internal storage
   REMARKS:    the returned pointer is short-lived and will become
               invalid if self is modified or destroyed
*/
{  return _str ? _str : "";
}

//............................................................................

inline int operator==(const Chi_String& a, const Chi_String& b)
{ return strcmp(a.c_str(), b.c_str()) == 0; }
inline int operator!=(const Chi_String& a, const Chi_String& b)
{ return strcmp(a.c_str(), b.c_str()) != 0; }
inline int operator<(const Chi_String& a, const Chi_String& b)
{ return strcmp(a.c_str(), b.c_str()) < 0; }
inline int operator<=(const Chi_String& a, const Chi_String& b)
{ return strcmp(a.c_str(), b.c_str()) <= 0; }
inline int operator>(const Chi_String& a, const Chi_String& b)
{ return strcmp(a.c_str(), b.c_str()) > 0; }
inline int operator>=(const Chi_String& a, const Chi_String& b)
{ return strcmp(a.c_str(), b.c_str()) >= 0; }

inline int operator==(const Chi_String& a, const char b[])
{ return strcmp(a.c_str(), b ? b : "") == 0; }
inline int operator!=(const Chi_String& a, const char b[])
{ return strcmp(a.c_str(), b ? b : "") != 0; }
inline int operator<(const Chi_String& a, const char b[])
{ return strcmp(a.c_str(), b ? b : "") < 0; }
inline int operator<=(const Chi_String& a, const char b[])
{ return strcmp(a.c_str(), b ? b : "") <= 0; }
inline int operator>(const Chi_String& a, const char b[])
{ return strcmp(a.c_str(), b ? b : "") > 0; }
inline int operator>=(const Chi_String& a, const char b[])
{ return strcmp(a.c_str(), b ? b : "") >= 0; }

inline int operator==(const char a[], const Chi_String& b)
{ return strcmp(a ? a : "", b.c_str()) == 0; }
inline int operator!=(const char a[], const Chi_String& b)
{ return strcmp(a ? a : "", b.c_str()) != 0; }
inline int operator<(const char a[], const Chi_String& b)
{ return strcmp(a ? a : "", b.c_str()) > 0; }
inline int operator<=(const char a[], const Chi_String& b)
{ return strcmp(a ? a : "", b.c_str()) >= 0; }
inline int operator>(const char a[], const Chi_String& b)
{ return strcmp(a ? a : "", b.c_str()) < 0; }
inline int operator>=(const char a[], const Chi_String& b)
{ return strcmp(a ? a : "", b.c_str()) <= 0; }


#line 744 "chistrng.cpp"


inline const Chi_String& Chi_String::operator+=(const Chi_String& b)
/* PURPOSE:    append a string to self
   RECEIVES:   b - the string to append
   RETURNS:    self
*/
{  append(b);
   return self;
}

/*...........................................................................*/

inline const Chi_String& Chi_String::operator+=(const char b[])
/* PURPOSE:    append a string to self
   RECEIVES:   b - the string to append
   RETURNS:    self
*/ 
{  append(b);
   return self;
}

/*...........................................................................*/

inline const Chi_String& Chi_String::operator+=(char ch)
/* PURPOSE:    append a character to self
   RECEIVES:   ch - the character to append
   RETURNS:    self
   REMARKS:    this is slow since the entire string must be copied.
               Do not use this function to build up a string. Instead,
               use grow() to desired size and set() when possible
*/
{  append(ch);
   return self;
}


#endif
