/****************************************************************************
** COPYRIGHT (C):    1994 Cay S. Horstmann. All Rights Reserved.
** PROJECT:          Mastering Object-Oriented Design
** FILE:             bank2.cpp
** PURPOSE:          A bank simulation
** VERSION           1.0
** LANGUAGE:         Borland C++ 3.0
** TARGET:           IBM PC
** PROGRAMMER:       Cay Horstmann (CSH)
** START DATE:       9-18-94 (CSH)
** UPDATE HISTORY:
****************************************************************************/

/* IMPORT ******************************************************************/

#include "simfrmwk.h"
#include "chiarray.h"
#include "chiqueue.h"

/* TYPES *******************************************************************/

class Arrival : public Event
/* PURPOSE:    Customer arrives at bank
*/
{
public:
   Arrival(Time);
   virtual void process(Simulation&);
};

/*-------------------------------------------------------------------------*/

class Departure : public Event
/* PURPOSE:    Customer departs from teller
*/
{
public:
   Departure(Time, int teller);
   virtual void process(Simulation&);
private:
   int _teller;
};

/*-------------------------------------------------------------------------*/

class Customer
/* PURPOSE:    A bank customer
*/
{
public:
   Customer(Time t);
   Time arrival_time() const;
private:
   Time _arrival_time;
};

/*-------------------------------------------------------------------------*/

class BankStatistics
/* PURPOSE:    Gather statistics to compute averages
*/
{
public:
   BankStatistics();
   void add(Time t);
   void print() const;
   double average_time() const;
private:
   int _ncust;
   Time _total_time;
};

/*-------------------------------------------------------------------------*/

class Bank : public Simulation
/* PURPOSE:    Simulation of customer traffic in a bank
*/
{
public:
   Bank();
   void add(Customer*);
   void add_to_teller(int teller, Customer*);
   Customer* remove(int teller);
   void print() const;
   virtual void initialize();
   virtual void update();
   virtual void terminate();
private:
   enum { NTELLER = 5 };
   Chi_PtrArray<Customer> _teller;
   Chi_Queue<Customer*> _cust_queue;
   BankStatistics _stat;
};

/* GLOBALS *****************************************************************/

const Time INTERARRIVAL = 1.0; // 1 minute
const Time AVG_PROCTIME = 5.0; // 5 minutes

/* FUNCTIONS ***************************************************************/

int main()
{  Bank bank;
   bank.run(9 * 60, 17 * 60);
   return 0;
}

/*-------------------------------------------------------------------------*/

Arrival::Arrival(Time t)
/* RECEIVES:   t - the time at which the arrival will occur
*/
: Event(t)
{}

/*.........................................................................*/

void Arrival::process(Simulation& s)
/* PURPOSE:    process an arrival event
   RECEIVES:   the simulation generating the event
*/
{  Bank& bank = (Bank&)s;
   Customer* c = new Customer(bank.now());
   bank.add(c);
   Time t = expdist(INTERARRIVAL);
   bank.schedule(new Arrival(bank.now() + t));
}

/*-------------------------------------------------------------------------*/

Departure::Departure(Time t, int teller)
/* RECEIVES:   t - the time at which the departure will occur
               teller - teller position
*/
: Event(t), _teller(teller)
{}

/*.........................................................................*/

void Departure::process(Simulation& s)
/* PURPOSE:    process a departure event
   RECEIVES:   the simulation generating the event
*/
{  Bank& bank = (Bank&)s;
   Customer* c = bank.remove(_teller);
   delete c;
}

/*-------------------------------------------------------------------------*/

Customer::Customer(Time t)
/* RECEIVES:   the time at which the customer entered the bank
*/
: _arrival_time(t)
{}

/*.........................................................................*/

Time Customer::arrival_time() const
/* RETURNS:    the time at which the customer entered the bank
*/
{  return _arrival_time;
}

/*-------------------------------------------------------------------------*/

BankStatistics::BankStatistics()
: _ncust(0), _total_time(0)
{}

/*.........................................................................*/

void BankStatistics::add(Time t)
/* PURPOSE:    add another customer who is leaving the bank to the statistics
   RECEIVES:   t - the time the customer left
*/
{  _ncust++;
   _total_time += t;
}

/*.........................................................................*/

double BankStatistics::average_time() const
/* PURPOSE:    returns the average time the customers spent in the bank
*/
{  if (_ncust == 0) return 0;
   else return _total_time/_ncust;
}

/*.........................................................................*/

void BankStatistics::print() const
/* PURPOSE:    print a summary of the gathered statistics
*/
{  cout << _ncust
      << " customers. Average time "
      << average_time() << " minutes." << endl;
}

/*-------------------------------------------------------------------------*/

Bank::Bank()
: _teller(1, NTELLER)
{}

/*.........................................................................*/

void Bank::initialize()
/* PURPOSE:    initialize bank simulation
*/
{  schedule(new Arrival(now()));
}

/*.........................................................................*/

void Bank::update()
/* PURPOSE:    action after every event processing
*/
{  print();
}

/*.........................................................................*/

void Bank::terminate()
/* PURPOSE:    action after simulation end
*/
{  _stat.print();
}

/*.........................................................................*/

void Bank::add(Customer* c)
/* PURPOSE:    Add a customer to the bank
   RECEIVES:   c - the customer
*/
{  for (int i = 1; i <= NTELLER; i++)
      if (_teller[i] == 0)
      {  add_to_teller(i, c);
         return;
      }
   _cust_queue.insert(c);
}

/*.........................................................................*/

void Bank::add_to_teller(int i, Customer* c)
/* PURPOSE:    add customer to an empty teller
   RECEIVES:   i - teller position
               c - the customer to add
*/
{  CHI_ASSERT_PRECOND(_teller[i] == 0);
   _teller[i] = c;
   Time t = expdist(AVG_PROCTIME);
   schedule(new Departure(now() + t, i));
}

/*.........................................................................*/

Customer* Bank::remove(int i)
/* PURPOSE:    remove customer from teller
   RECEIVES:   i - teller position
*/
{  Customer* c = _teller[i];
   _teller[i] = 0;
   _stat.add(now() - c->arrival_time());
   if (_cust_queue.length() > 0)
      add_to_teller(i, _cust_queue.remove());
   return c;
}

/*.........................................................................*/

void Bank::print() const
/* PURPOSE:    Print teller and queue
*/
{  for (int i = 1; i <= NTELLER; i++)
      cout << (_teller[i] == 0 ? '.' : 'C');
   cout << '<';
   int q = _cust_queue.length();
   for (int j = 1; j <= q; j++) cout << 'C';
   cout << endl;
}

/***************************************************************************/
