/****************************************************************************
** COPYRIGHT (C):    1994 Cay S. Horstmann. All Rights Reserved.
** PROJECT:          Mastering Object-Oriented Design
** FILE:             pstream.cpp
** PURPOSE:          A simple persistent stream implementation
** VERSION           1.0
** LANGUAGE:         Borland C++ 3.1
** TARGET:           IBM PC
** PROGRAMMERS:      Cay S. Horstmann (CSH)
** START DATE:       September 14, 1994
** UPDATE HISTORY:
****************************************************************************/

/* IMPORT ******************************************************************/

#ifndef EXPORT
#include "chisetup.h"
#endif

EXPORT
#include <fstream.h>
EXPORT
#include "chistrng.h"
EXPORT
#include "chiarray.h"
EXPORT
#include "chimap.h"

#include "pstream.h"

/* TYPES *******************************************************************/

#ifndef EXPORT

class Persistent;

class PersistentStream : public fstream
/* PURPOSE:    a stream for saving and loading persistent objects
*/
{
public:
   PersistentStream();
   void load(Chi_String);
   void save(Chi_String);
   void close();
   Bool is_saving() const;

   void archive(Persistent*& x);
   Persistent* find(int) const;
   void add(int, Persistent*);
   int find(const Persistent* p) const;
   int add(const Persistent* p);

private:
   enum Mode { CLOSED, LOAD, SAVE };
   Chi_PtrArray<Persistent> _toptr;
   Chi_Map<const Persistent*, int> _fromptr;
   int _maxnum;
   Mode _mode;
};

/*-------------------------------------------------------------------------*/

class Persistent
/* PURPOSE:    the base class for persistent objects
*/
{
public:
   virtual void archive(PersistentStream&) = 0;
   virtual Chi_String name() const = 0;
   virtual Persistent* clone() const = 0;

   static Persistent* add_type(Chi_String, Persistent*);
   static Persistent* clone(Chi_String);
private:
   static Chi_Map<Chi_String, Persistent*> _typemap;
};

#endif

/* GLOBALS *****************************************************************/

Chi_Map<Chi_String, Persistent*> Persistent::_typemap(Chi_String::hash, Chi_String::compare);
   // maps type names to model objects

/* FUNCTIONS ***************************************************************/

int hash_ptr(Persistent const* const& p)
/* PURPOSE:    hash a pointer
   RECEIVES:   p - a (constant reference to) a persistent pointer
   RETURNS:    a hash value for p
*/
{  return (((int)p) >> 16) ^ (((int)p) & 0xFFFF);
}

/*.........................................................................*/

int comp_ptr(Persistent const* const& a, Persistent const* const& b)
/* PURPOSE:    compare two pointers
   RECEIVES:   a, b - (constant references to) persistent pointers
   RETURNS:    0 if the pointers are identical, != 0 otherwise
*/
{  return !(a == b);
}

/*.........................................................................*/

void archive(int& i, PersistentStream& ps)
/* PURPOSE:    archive an integer on a stream
   RECEIVES:   i - an integer
               ps - a persistent stream
*/
{  if(ps.is_saving())
      ps << i << endl;
   else
      ps >> i;
}

/*.........................................................................*/

void archive(Chi_String& s, PersistentStream& ps)
/* PURPOSE:    archive a string on a stream
   RECEIVES:   s - a string
               ps - a persistent stream
*/
{  if(ps.is_saving())
      ps << s << endl; // actually use delimited
   else
      ps >> s;
}

/*.........................................................................*/

void archive(Persistent& p, PersistentStream& ps)
/* PURPOSE:    archive a persistent object on a stream
   RECEIVES:   p - a persistent object
               ps - a persistent stream
*/
{  p.archive(ps);
}

/*.........................................................................*/

#ifndef EXPORT

template<class X>
void archive(X*& p, PersistentStream& ps)
/* PURPOSE:    archive a persistent pointer on a stream
   RECEIVES:   p - a pointer to an object of a type derived from Persistent
               ps - a persistent stream
*/
{  ps.archive((Persistent*&)p);
}

/*.........................................................................*/

template<class X>
void archive(Chi_Array<X>& a, PersistentStream& ps)
/* PURPOSE:    archive an array of archivable items
   RECEIVES:   a - an array of object of a type that can be archived
               ps - a persistent stream
*/
{  if (ps.is_saving())
      ps << a.low() << " " << a.high() << endl;
   else
   {  int l, h;
      ps >> l >> h;
      a.empty();
      a.grow(l, h);
   }
   for (int i = a.low(); i <= a.high(); i++)
   {  archive( a[i], ps);
   }
}

/*.........................................................................*/

template<class X>
void archive(Chi_PtrArray<X>& a, PersistentStream& ps)
/* PURPOSE:    archive an array of archivable items
   RECEIVES:   a - an array of object of a type that can be archived
               ps - a persistent stream
*/
{  if (ps.is_saving())
      ps << a.low() << " " << a.high() << endl;
   else
   {  int l, h;
      ps >> l >> h;
      a.empty();
      a.grow(l, h);
   }
   for (int i = a.low(); i <= a.high(); i++)
   {  archive( a[i], ps);
   }
}

/*.........................................................................*/

template<class X>
void archive(Chi_NumArray<X>& a, PersistentStream& ps)
/* PURPOSE:    archive an array of archivable items
   RECEIVES:   a - an array of object of a type that can be archived
               ps - a persistent stream
*/
{  if (ps.is_saving())
      ps << a.low() << " " << a.high() << endl;
   else
   {  int l, h;
      ps >> l >> h;
      a.empty();
      a.grow(l, h);
   }
   for (int i = a.low(); i <= a.high(); i++)
   {  archive( a[i], ps);
   }
}

/*.........................................................................*/

#define DEFINE_PERSISTENT(X) \
   Chi_String name() const { return #X; } \
   Persistent* clone() const { return new X(*this); } \
   static Persistent* _pobj;

/*.........................................................................*/

#define IMPLEMENT_PERSISTENT(X) \
   Persistent* X::_pobj = Persistent::add_type(#X, new X);

#endif

/*-------------------------------------------------------------------------*/

PersistentStream::PersistentStream()
:  _fromptr(hash_ptr, comp_ptr)
{  _mode = CLOSED;
}

/*.........................................................................*/

void PersistentStream::load(Chi_String s)
/* PURPOSE:    open a persistent stream for loading
*/
{  fstream::open(s.c_str(), ios::in);
   _mode = LOAD;
   _toptr.empty();
}

/*.........................................................................*/

Bool PersistentStream::is_saving() const
/* RETURNS:    TRUE iff the stream is open for saving
*/
{  return _mode == SAVE;
}

/*.........................................................................*/

void PersistentStream::close()
/* PURPOSE:    close the stream
*/
{  fstream::close();
   _mode = CLOSED;
}

/*.........................................................................*/

Persistent* PersistentStream::find(int n) const
/* PURPOSE:    find a previously archived pointer
   RECEIVES:   n - a serial number
   RETURNS:    the archived pointer with serial number n
*/
{  return _toptr.get(n);
}

/*.........................................................................*/

void PersistentStream::add(int n, Persistent* p)
/* PURPOSE:    remember a pointer with its serial number
   RECEIVES:   n - serial number
               p - pointer to persistent object
*/
{  _toptr.set(n, p);
}

/*.........................................................................*/

void PersistentStream::save(Chi_String s)
/* PURPOSE:    open a persistent stream for saving
*/
{  fstream::open(s.c_str(), ios::out);
   _mode = SAVE;
   _fromptr.empty();
   _maxnum = 0;
}

/*.........................................................................*/

int PersistentStream::find(const Persistent* p) const
/* PURPOSE:    find the serial number of a pointer
   RECEIVES:   p - pointer to persistent object
   RETURNS:    the pointer's serial number
*/
{  int l;
   if (_fromptr.find(p, l)) return l;
   return 0;
}

/*.........................................................................*/

int PersistentStream::add(const Persistent* p)
/* PURPOSE:    add a new pointer
   RECEIVES:   p - pointer to persistent object
   RETURNS:    the pointer's serial number
*/
{  _maxnum++;
   _fromptr.set(p, _maxnum);
   return _maxnum;
}

/*.........................................................................*/

void PersistentStream::archive(Persistent*& x)
/* PURPOSE:    archive a persistent pointer
   RECEIVES:   x - a reference to a pointer to a parsistent object
*/
{  if (is_saving())
   {  if (x == 0) { self << "0" << endl; return; }
      int n = find(x);
      if (n) self << n << endl;
      else
      {  n = add(x);
         self << n << " " << x->name() << endl; // write type name
         x->archive(self);
      }
   }
   else
   {  int n;
      self >> n;
      if (n == 0) { x = 0; return; }
      x = find(n);
      if (x == 0) // new entry
      {  Chi_String name;
         self >> name;
         x = Persistent::clone(name);
         if (x == 0)
         {  clear(ios::failbit);
            return;
         }
         add(n, x);
         x->archive(self);
      }
   }
}

/*-------------------------------------------------------------------------*/

Persistent* Persistent::add_type(Chi_String s, Persistent* p)
/* PURPOSE:    add a new type to the type map
   RECEIVES:   s - type name
               p - model object
*/
{  _typemap.set(s, p);
   return p;
}

/*.........................................................................*/

Persistent* Persistent::clone(Chi_String s)
/* PURPOSE:    make a new object of a given type
   RECEIVES:   s - type name
   RETURNS:    an object of type s allocated on the heap
*/
{  Persistent* p;
   if (_typemap.find(s, p))
      return p->clone();
   else
      return 0;
}

/* TEST ********************************************************************/

#ifdef PSTREAM_TEST

class Person : public Persistent
{
public:
   Person();
   Person(Chi_String, int);

   virtual void archive(PersistentStream&);

   DEFINE_PERSISTENT(Person)

private:
   Chi_String _name;
   int _age;
};

/*.........................................................................*/

IMPLEMENT_PERSISTENT(Person)

Person::Person()
:  _age(0)
{}

Person::Person(Chi_String s, int a)
:  _name(s), _age(a) {}

void Person::archive(PersistentStream& ps)
{  ::archive(_name, ps);
   ::archive(_age, ps);
}

/*-------------------------------------------------------------------------*/

class Vehicle : public Persistent
{
public:
   Vehicle(Person* owner = 0);
   void drive(Person*);

   virtual void archive(PersistentStream&);

   DEFINE_PERSISTENT(Vehicle)
   // name, clone

private:
   Person* _owner;
   Person* _driver;
};

/*.........................................................................*/

IMPLEMENT_PERSISTENT(Vehicle)

Vehicle::Vehicle(Person* p) : _owner(p), _driver(0) {}
void Vehicle::drive(Person* p) { _driver = p; }

void Vehicle::archive(PersistentStream& ps)
{  ::archive(_owner, ps);
   ::archive(_driver, ps);
}

/*-------------------------------------------------------------------------*/

class Sequence : public Persistent
{
public:
   Sequence();
   Sequence(int n);
   virtual void archive(PersistentStream&);

   DEFINE_PERSISTENT(Sequence)

private:
#ifdef CHI_WC10
   // Watcom 10.0 won't instantiate all functions needed for Chi_NumArray<int>
   Chi_Array<int> _a;
#else
   Chi_NumArray<int> _a;
#endif
};

/*.........................................................................*/

Sequence::Sequence() {}

Sequence::Sequence(int n) : _a(1, n)
{  for (int i = 1; i <= n; i++) _a[i] = i*i;
}

void Sequence::archive(PersistentStream& ps)
{  ::archive(_a, ps);
}

IMPLEMENT_PERSISTENT(Sequence)

/*-------------------------------------------------------------------------*/

class Bank : public Persistent
{
public:
   Bank();
   void add(Person*);
   virtual void archive(PersistentStream&);

   DEFINE_PERSISTENT(Bank)

private:
   Chi_PtrArray<Person> _a;
};

/*.........................................................................*/

Bank::Bank() {}
void Bank::add(Person* p) { _a.append(p); }

void Bank::archive(PersistentStream& ps)
{  ::archive(_a, ps);
}

IMPLEMENT_PERSISTENT(Bank)

/*-------------------------------------------------------------------------*/

int main()
{  Person* joe = new Person("Joe", 41);
   Person* carl = new Person("Carl", 17);
   Vehicle car(joe);
   Vehicle truck(joe);
   Sequence s(5);
   Bank bank;

   car.drive(carl);
   truck.drive(joe);
   bank.add(carl);
   bank.add(joe);

   PersistentStream ps;

   ps.save("TEST.DAT");
   car.archive(ps);
   truck.archive(ps);
   s.archive(ps);
   bank.archive(ps);
   ps.close();

   delete joe;
   delete carl;

   ps.load("TEST.DAT");
   car.archive(ps);
   truck.archive(ps);
   s.archive(ps);
   bank.archive(ps);
   ps.close();

   ps.save("TEST.DAT");
   archive(car, ps);
   archive(truck, ps);
   archive(s, ps);
   archive(bank, ps);
   ps.close();

   return 0;
}

#endif

/***************************************************************************/
