/****************************************************************************
** COPYRIGHT (C):    1994 Cay S. Horstmann. All Rights Reserved.
** PROJECT:          Mastering Object-Oriented Design
** FILE:             shapes4.cpp
** PURPOSE:          Multiple inheritance
** VERSION           1.0
** LANGUAGE:         Borland C++ 3.1
** TARGET:           IBM PC
** PROGRAMMERS:      Cay S. Horstmann (CSH)
** START DATE:       September 14, 1994
** UPDATE HISTORY:
****************************************************************************/

/* IMPORT ******************************************************************/

#include "chiarray.h"
#include "chistrng.h"
#include "chigc.h"
#include "math.h"

/* TYPES *******************************************************************/

class Chi_Point;

class Chi_Shape
/* PURPOSE:    A shape that can be plotted and moved around
*/
{
public:
   Chi_Shape();
   Chi_Shape(Chi_GraphicsContext::Color);

   void set_color(Chi_GraphicsContext& gc) const; 

   virtual void plot(Chi_GraphicsContext& gc) const = 0;
   virtual ~Chi_Shape() {}
private:
   Chi_GraphicsContext::Color _color;
};

/*-------------------------------------------------------------------------*/

class Chi_Point : public Chi_Shape
/* PURPOSE:    A single point
*/
{
public:
   Chi_Point(double xx =0, double yy =0);

   virtual void plot(Chi_GraphicsContext& gc) const;

   double x() const;
   double y() const;

private:
   double _x;
   double _y;
};

/*-------------------------------------------------------------------------*/

class Chi_Polygon : public Chi_Shape
/* PURPOSE:    A closed (not necessarily convex) polygon
*/
{
public:
   Chi_Polygon();
   Chi_Polygon(int);

   Chi_Point vertex(int) const;
   void set_vertex(int, Chi_Point);

   virtual void plot(Chi_GraphicsContext& gc) const;

private:
   Chi_Array<Chi_Point> _vert;
};

/*-------------------------------------------------------------------------*/

class Chi_Text : public Chi_Shape, public Chi_String
/* PURPOSE:    A displayed string
*/
{
public:
   Chi_Text(Chi_Point, Chi_String, Chi_GraphicsContext::Color);

   void plot(Chi_GraphicsContext& gc) const;
private:
   Chi_Point _start;
};

/*-------------------------------------------------------------------------*/

class Chi_TrafficSign : public Chi_Text, public Chi_Polygon
/* PURPOSE:    A traffic sign shape--a polygon with text inside
*/
{
public:
   Chi_TrafficSign(Chi_Point c, int nvert, Chi_String t);
   void plot(Chi_GraphicsContext& gc) const;
};

/* FUNCTIONS ***************************************************************/

int chi_main(int, char**)
{  Chi_GraphicsContext gc;
   Chi_PtrArray<Chi_Shape> a;
   a.append(new Chi_Point(1, 1));
   a.append(new Chi_Text(Chi_Point(5, 5), "Hello", Chi_GraphicsContext::BLUE));
   Chi_TrafficSign* sign = new Chi_TrafficSign(Chi_Point(9, 9), 8, "Stop");
   sign->to_upper(); // it is a text :-)
   a.append((Chi_Polygon*)sign); // but it isn't a shape :-(
   for (int j = a.low(); j <= a.high(); j++)
      a[j]->plot(gc);
   gc.close();
   return 0;
}

/*-------------------------------------------------------------------------*/

Chi_Shape::Chi_Shape() : _color(Chi_GraphicsContext::DEFAULT_COLOR) {}

//............................................................................

Chi_Shape::Chi_Shape(Chi_GraphicsContext::Color c) : _color(c) {}

//............................................................................

void Chi_Shape::set_color(Chi_GraphicsContext& gc) const
/* PURPOSE:    Select the shape color into the graphics context
*/
{  gc.set_pen_color(_color);
}

/*-------------------------------------------------------------------------*/

Chi_Point::Chi_Point(double xx, double yy)
/* RECEIVES:   xx, yy - coordinates
*/
   : _x(xx), _y(yy)
{}

//............................................................................

void Chi_Point::plot(Chi_GraphicsContext& gc) const
/* PURPOSE:    plot this shape
   RECEIVES:   gc - graphics context
*/
{  set_color(gc);
   gc.point(_x, _y);
}

//............................................................................

inline double Chi_Point::x() const
/* RETURNS:    the x-coordinate
*/
{  return _x;
}

//............................................................................

inline double Chi_Point::y() const
/* RETURNS:    the y-coordinate
*/
{  return _y;
}

/*-------------------------------------------------------------------------*/

Chi_Text::Chi_Text(Chi_Point p, Chi_String t, Chi_GraphicsContext::Color c)
/* RECEIVES:   p - center of text
               t - text contents
               c - text color
   REMARKS:    This text class is different from the one in ChiLib--it
               plots the text with center p, not p as upper left corner.
*/
:  Chi_String(t),
   Chi_Shape(c),
   _start(p)
{}

//............................................................................

void Chi_Text::plot(Chi_GraphicsContext& gc) const
/* PURPOSE:    plot this shape
   RECEIVES:   gc - graphics context
*/
{  set_color(gc);
   double x, y;
   gc.text_extent(self, x, y);
   gc.text(self, _start.x() - x / 2, _start.y() - y / 2);
}

/*-------------------------------------------------------------------------*/

Chi_Polygon::Chi_Polygon()
/* REMARKS:    Makes a useless null polygon
*/
{}

//............................................................................

Chi_Polygon::Chi_Polygon(int n)
/* RECEIVES:   n - the number of vertices
*/
:  _vert(1, n)
{}

//............................................................................

Chi_Point Chi_Polygon::vertex(int i) const
/* RECEIVES:   i - an integer
   RETURNS:    the ith vertex
*/
{  return _vert.get(i);
}

//............................................................................

void Chi_Polygon::set_vertex(int i , Chi_Point p)
/* PURPOSE:    set a vertex of the polygon
   RECEIVES:   i - an integer between 1 and the number of vertices
               p - the point to store
*/
{  _vert[i] = p;
}

//............................................................................

void Chi_Polygon::plot(Chi_GraphicsContext& gc) const
/* PURPOSE:    plot this shape
   RECEIVES:   gc - graphics context
*/
{  set_color(gc);
   for (int i = _vert.low(); i <= _vert.high(); i++)
   {  int j = i + 1;
      if (j > _vert.high()) j = _vert.low();
      Chi_Point p = _vert.get(i);
      Chi_Point q = _vert.get(j);
      gc.line(p.x(), p.y(), q.x(), q.y());
   }
}

/*-------------------------------------------------------------------------*/

Chi_TrafficSign::Chi_TrafficSign(Chi_Point p, int nvert, Chi_String t)
:  Chi_Text(p, t, Chi_GraphicsContext::RED),
   Chi_Polygon(nvert)
/* RECEIVES:   p - center of traffic sign
               nvert - number of verices
               t - text of traffic sign
*/
{  // nvert calls to set_vertex
   const double RAD = 1;
   for (int i = 1; i <= nvert; i++)
   {  double angle = (1 + 2 * i) * M_PI / nvert;
      set_vertex(i, Chi_Point(p.x() + RAD * cos(angle), p.y() + RAD * sin(angle)));
   }
}

//............................................................................

void Chi_TrafficSign::plot( Chi_GraphicsContext& gc ) const
/* PURPOSE:    plot this shape
   RECEIVES:   gc - graphics context
*/
{  Chi_Polygon::plot(gc);
   Chi_Text::plot(gc);
}

/***************************************************************************/
